******************************************************************************
*       PROGRAM AUTHOR:         RICHARD E. CHANDLER
*                               (email: richard@stats.ucl.ac.uk)
*       AFFILIATION:            DEPARTMENT OF STATISTICAL SCIENCE
*                                UNIVERSITY COLLEGE LONDON
*                                 GOWER STREET
*                                  LONDON WC1E 6BT
*                                   UK
*	Copyright  UCL 2002-2010
******************************************************************************
* This file contains a collection of FORTRAN routines used in both fitting
* programs relating to Generalised Linear Modelling of daily rainfall. For
* more details of these programs, see the headers to the files fit_logi.f
* and fit_gamm.f.
*
* REVISION HISTORY: See manual.
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE MXITST(MAXIT)
*
*       Sets maximum number of iterations. Needs to be relegated to a 
*       subroutine in order to enable defaults. The rather clumsy way 
*       of doing things, which is a product of FORTRAN's wonderful way 
*       of treating blanks in numeric input as zeroes, is: read the 
*       number as a character (so blank is distinguished from zero); if 
*       blank, set to default, otherwise read number from the character
*       string. Variables: MAXIT, the numeric value and TMPSTR, the 
*       character storage.
*      
      INTEGER MAXIT
      CHARACTER TMPSTR*4

 40   WRITE(*,1)
      READ(*,'(A4)',ERR=40) TMPSTR
      IF (TMPSTR.EQ.'    ') THEN
       MAXIT = -1
      ELSE
       READ(TMPSTR,'(I4)',ERR=40) MAXIT
      ENDIF
      RETURN

 1    FORMAT(//5X,'Input maximum number of iterations for ',
     +'iterative weighted',/5X,
     +'least squares (default unlimited):')
      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE UPDATE(FILNO,RAIN,PREVIOUS,THRESH,THRTYP,NSITES,
     +                            YY,MM,DD,MXLAG,MXN,SCODES,FILOPN)
*
*       This updates the values held in the RAIN and PREVIOUS arrays.
*       If thresholding has been requested (THRTYP > 0) it's done here. 
*
      INTEGER FILNO,MXN,YY,MM,DD,MXLAG,NSITES,FILOPN,THRTYP
      DOUBLE PRECISION RAIN(MXN),PREVIOUS(MXN,10),THRESH
      CHARACTER SCODES(MXN)*4
******************************************************************************
*       Additional INTEGERS
*       ^^^^^^^^^^^^^^^^^^^
*       DAYS_IN_MONTH   - needed to calculate DP
******************************************************************************
      INTEGER DAYS_IN_MONTH
      INTEGER I,J

*
*       First update the PREVIOUS values. 
*
      DO 10 I=1,NSITES
       DO 11 J=MXLAG,2,-1
        PREVIOUS(I,J) = PREVIOUS(I,J-1)
 11    CONTINUE
       PREVIOUS(I,1) = RAIN(I)
       RAIN(I) = -1.0D0
 10   CONTINUE

*
*       Increment date unless this is the first time through, in which
*       case we don't know what the first date is going to be
*
      IF (YY.NE.0) THEN
       DD = DD + 1
       IF (DD.GT.DAYS_IN_MONTH(MM,YY)) THEN
	 DD = 1
	MM = MM + 1
	IF (MM.GT.12) THEN
	 MM = 1
	 YY = YY + 1
	ENDIF
       ENDIF
      ENDIF
*
*       Read data for today
*    
      CALL DATRD(FILNO,RAIN,SCODES,NSITES,YY,MM,DD,MXN,FILOPN)

*
*       Threshold if requested
*
      IF (THRTYP.GT.0) THEN
       DO 20 I = 1,NSITES
        IF (RAIN(I).LT.0.0D0) GOTO 20
        IF (RAIN(I).LT.THRESH) THEN
         RAIN(I) = 0.0D0
        ELSEIF (THRTYP.EQ.1) THEN
         RAIN(I) = RAIN(I) - THRESH
        ENDIF
 20    CONTINUE
      ENDIF
      
      RETURN

      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE DATRD(FILNO,RAIN,SCODES,NSITES,YY,MM,DD,MXN,FILOPN)
*
*       Reads in rows of data from file connected to unit FILNO and allocates
*       them to array RAIN if they correspond to date DD/MM/YY. Also updates
*       DATINIT if necessary, and counts the number of lines before the first
*       data value if 
*
      INTEGER FILNO,YY,MM,DD,MXN,NSITES,FILOPN
      DOUBLE PRECISION RAIN(MXN)
      CHARACTER*4 SCODES(MXN)
******************************************************************************
*       Additional INTEGERS
*       ^^^^^^^^^^^^^^^^^^^
*       DF      - Day, as read from input file
*       MF      - Month, ditto
*       YF      - Year, ditto
*       IOERR   - Input file error flag - will be < 0 when EOF is attained
******************************************************************************
      INTEGER DF,MF,YF,IOERR,I
******************************************************************************
*       Additional DOUBLEs
*       ^^^^^^^^^^^^^^^^^^
*       RAINF   - rainfall value read from file
******************************************************************************
      DOUBLE PRECISION RAINF
******************************************************************************
*       Additional CHARACTERs
*       ^^^^^^^^^^^^^^^^^^^^^
*       CODEF   - site code as defined in input file
******************************************************************************
      CHARACTER CODEF*4

      YF=0
      MF=0
      DF=0
*
*       Read a line of the input file. 
*
 19   READ(FILNO,'(I4,I2,I2,A4,F6.2)',ERR=91,END=25,IOSTAT=IOERR)
     +                                      YF,MF,DF,CODEF,RAINF
*
*       If we've got the right day, update RAIN according to the site code. 
*       We will have the right day EITHER if it's the first time of
*       reading the file OR if YF,MF and DF correspond to YY,MM and DD
*
      IF ( ((10000*(YF-YY))+(100*(MF-MM))+DF-DD.EQ.0)
     +                                  .OR.(YY.EQ.0) ) THEN
       DO 20 I=1,NSITES
	IF (CODEF.EQ.SCODES(I)) THEN
	 RAIN(I) = RAINF
	ENDIF
 20    CONTINUE
*
*       If this is the first time through, set YY,MM and DD correctly
*
       IF (YY.EQ.0) THEN
	YY = YF
	MM = MF
	DD = DF
       ENDIF
*
*       If we overshoot, rewind the input file 1 line and go back
*
      ELSEIF ((10000*(YF-YY))+(100*(MF-MM))+DF-DD.GT.0) THEN
       BACKSPACE (UNIT=FILNO)
       RETURN
*
*       If the input file is the wronf order, stop with an error 
*       message
*
      ELSEIF ((10000*(YF-YY))+(100*(MF-MM))+DF-DD.LT.0) THEN
       GOTO 92
      ENDIF

      GOTO 19
*
*       End of file
*       
 25   IF (IOERR.LT.0) THEN
       CLOSE(FILNO)
       FILOPN = 0
      ENDIF

      RETURN

*
*       Error trapping
*
 91   WRITE(*,1) DF,MF,YF
      STOP
 92   WRITE(*,2) DF,MF,YF,CODEF
      STOP

 1    FORMAT(//'****ERROR**** while reading input file.',
     +' The last successful access',/,'was for data on ',
     +I2,'/',I2,'/',I4,'.')
 2    FORMAT(//'****ERROR**** input file out of order.',
     +' Offending date is',/,I2,'/',I2,'/',I4,', at site ',A4,'.')

      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE IWLS(MODEL,NP,P,COVCODE,TWO,THREE,NLSTAT,SPMOD,
     +			FOUIDX,LEGIDX,PWTIDX,MXP,MXN,MAXIT,MXLAG,
     +			SCRFN1,SCRFN2,OUTFNO,NOBS,NSITES,NATTR,
     +			SITXFM,GLBCOD,LOGL,BETA,SEBETA,THETA,SETHETA,
     +			NUIPAR,RHO,GLBVAL,SITINF,TITLE,XLABELS,
     +			NLTXT,SPATXT,SPPTXT,CORTXT,ATTRTXT,GLBTXT,
     +                  SCODES,FILNM,FILNO)
******************************************************************************
*	To fit a GLM, possibly involving nonlinear transformations of
*	covariates, using Iterative Weighted Least Squares. Parameter
*	vectors, standard errors and a log-likelihood are returned.
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
*////////////	NB THIS ROUTINE CONTAINS QUITE A FEW SECTIONS WHICH  ///////
*////////////	ARE MODEL-SPECIFIC. THESE SECTIONS ARE ENCLOSED BY   ///////
*////////////   LINES OF HASHES LIKE THIS, FOR EASY RECOGNITION IF   ///////
*////////////   YOU SUBSEQUENTLY WANT TO AMEND THEM OR ADD EXTRA     ///////
*////////////   MODELS.                                              ///////
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
******************************************************************************
*	PARAMETERS
*	~~~~~~~~~~
*	MXP2	- Maximum number of parameters in model. Used for
*		  dimensioning internal arrays. Should be set 
*		  equal to MXP (which should be a PARAMETER in the
*		  calling program)
*	MXN2	- Maximum number of sites. Should be equal to MXN
******************************************************************************
      INTEGER MXP2,MXN2
      PARAMETER (MXP2=80,MXN2=200)
******************************************************************************
*	ARGUMENTS
*	~~~~~~~~~
*	MODEL	- INTEGER, input. Defines the distribution being fitted.
*		  Values:
*			1  -	Logistic regression for binary responses
*			11 -	Gamma model with log link
*	NP	- INTEGER, input. Table indicating numbers of parameters
*		  in each category of predictor.
*	P	- INTEGER, input. Number of parameters being estimated.
*	COVCODE	- INTEGER, input. Table defining covariates.
*	TWO 	- INTEGER, input. Table defining 2-way interactions.
*	THREE 	- INTEGER, input. Table defining 3-way interactions.
*	NLSTAT	- INTEGER, input. Table defining status of any nonlinear
*		  transformations.
*	SPMOD	- INTEGER, input. Defines spatial model structure selected.
*	FOUIDX	- INTEGER, input. Table pointing to boundaries of
*		  orthogonal series representation of site effects
*		  (Fourier representations).
*	LEGIDX	- INTEGER, input. Table pointing to boundaries of
*		  orthogonal series representation of site effects
*		  (Legendre polynomial representations).
*       PWTIDX  - INTEGER, input. Table pointing to locations of 
*                 parameters/covariates relating to weighted averages of
*                 previous days' values.
*	MXP	- INTEGER, input. Max no of parameters in model.
*	MXN	- INTEGER, input. Max no of sites we'll deal with.
*	MAXIT	- INTEGER, input. Maximum number of iterations required.
*	MXLAG	- INTEGER, input. Flag tp indicate how many previous
*		  days' rainfall must be available in order for a 
*		  case to be included.
*	SCRFN1	- INTEGER, input. Number of scratch file being used
*		  to store data.
*	SCRFN2	- INTEGER, input. Number of scratch file being used
*		  for residual stuff.
*	OUTFNO	- INTEGER, input. Number of output results file.
*	NOBS	- INTEGER, input. Number of observations.
*	NSITES	- INTEGER, input. Number of sites used.
*	NATTR	- INTEGER, input. Number of site attributes.
*       SITXFM  - INTEGER, input. Selects nonlinear transformations 
*		  of site attributes.
*	GLBCOD  - INTEGER, input. Codes defining global quantities.
*	LOGL	- DOUBLE, output. Final log-likelihood.
*	BETA	- DOUBLE, input/output. Parameter estimates.
*	SEBETA	- DOUBLE, output. Standard errors of BETAs.
*	THETA	- DOUBLE, input/output. Estimates of parameters in
*		  nonlinear transformations of predictors.
*	SETHETA	- DOUBLE, output. Standard errors of THETAs.
*	NUIPAR	- DOUBLE, output. Dispersion parameter, where needed.
*	RHO	- DOUBLE, input/output. Parameters in spatial 
*		  dependence structures.
*	GLBVAL  - DOUB:E, input. Values of global quantities.
*	SITINF- DOUBLE, input. Site attributes.
*	TITLE	- CHARACTER, input. Title text for model.
*	XLABELS	- CHARACTER, input. Text for covariates.
*	NLTXT	- CHARACTER, input. Text for nonlinear parameters.
*	SPATXT	- CHARACTER, input. Description of spatial dependence
*		  structure.
*	SPPTXT	- CHARACTER, input. Text for spatial dependence 
*		  parameters.
*	CORTXT	- CHARACTER, input. Text describing quantities to which
*		  spatial correlations refer.
*       ATTRTXT - CHARACTER, input. Text for site attributes
*       SCODES  - CHARACTER, input. 4-character site codes
*       FILNM   - CHARACTER, input. Name of gauge data file.
*       FILNO   - INTEGER, input. Channel number associated with FILNM
******************************************************************************
      INTEGER MXN,MXP,P,NOBS,NSITES,MODEL,MAXIT
      INTEGER NP(10),COVCODE(MXP),NLSTAT(MXP,3),SPMOD,MXLAG
      INTEGER TWO(MXP,2),THREE(MXP,3),FOUIDX(MXP),LEGIDX(MXP)
      INTEGER PWTIDX(MXP,0:3),NATTR,SITXFM(MXP,2),GLBCOD(MXP)
      INTEGER SCRFN1,SCRFN2,OUTFNO,FILNO
      DOUBLE PRECISION BETA(0:MXP),SEBETA(0:MXP),LOGL
      DOUBLE PRECISION THETA(MXP,3),SETHETA(MXP,3)
      DOUBLE PRECISION RHO(MXP),GLBVAL(MXP),NUIPAR
      DOUBLE PRECISION SITINF(MXN,MXP,0:3)
      CHARACTER XLABELS(0:MXP)*70,NLTXT(MXP,3)*70
      CHARACTER*70 TITLE,GLBTXT(MXP)
      CHARACTER*70 SPATXT(0:MXP),SPPTXT(MXP,4),ATTRTXT(MXP)
      CHARACTER*70 CORTXT(MXP)
      CHARACTER SCODES(MXN)*4,FILNM*12
******************************************************************************
*	Extra INTEGERs - those marked # are just dummies that are required
*	~~~~~~~~~~~~~~	 by other routines such as COVSET.
*       IFAIL   - for error trapping
*	ITER	- current iteration number
*	SITE	- number of current site
*	YY	- Year for current case
*	MM	- Month for current case
*	DD	- Day for current case
*       OY,OM,OD- Year, month and day for previous case
*	CASEID	- array of identifiers for each case.
*	RECNO	- Record number in file SCRFN1 - used when updating 
*		  covariate attributes
* 	MISSFL	- Flags for sites which have missing covariate data
*       CNVRGE  - convergence status: 1=initial value, 2=iterating,
*                 3=converged
*       IDMAXU  - Parameter number with largest score
*       RECALC  - indicator for whether or not we have to calculate nonlinear
*                 functions at the beginning of an iteration. Codes:
*                 0 Calculate all covariates
*                 1 Calculate only covariates associated with nonlinear 
*                   parameters
*                 2 We've converged!
*       RESWRT  - indicator for whether residual information has been 
*                 written to scratch file. 
*	XRC	- indicator for whether we have to recompute X'WX (i.e. if
*		  either X matrix or weights have changed).
*       PRVRST  - Indicator for whether we need to reset the PREVIOUS
*                 array while calculating nonlinear predictors
*	OVRSHT	- Used to keep track of number of times we correct for
*		  `overshooting' in an IWLS iteration. If this happens
*		  so that the log-likelihood decreases substantially,
*		  we backtrack up to 10 times. Rounding errors and changes
*		  to data values in EM part of algorithm may mean that we
*		  never get exactly back to where we were.
*	TRWRN	- Indicates if we've issued a warning message about trace 
*		  values on this iteration
*	FORCE	- Used to force COVSET to recalculate seasonal and yearly
*		  contributions for the first case in the dataset on each
*		  iteration - to be on the safe side.
*       THRTYP  - Type of thresholding to be used
*       Y0,M0,D0- Year, month and day read from original daily data file
*                 when recalculating the PREVIOUS array
*       FILOPN  - Records whether gauge data file is open
*       UFILNO  - Number of scratch file for daily scores.
*       NDAYS   - Number of days of data (= number of rows in UFILNO)
*       Q       - Number of zeroes in BETA0 (any such zeroes are
*                 assumed to reflect a previously fitted model with
*                 the corresponding covariates omitted).
*       DELTA   - Stores indices of zero elements in starting value
*	I,J,K	- Counters
******************************************************************************
      INTEGER IFAIL,ITER,SITE,YY,MM,DD,OY,OM,OD,CASEID(4),RECNO
      INTEGER MISSFL(MXN2),CNVRGE,RECALC,RESWRT,XRC,PRVRST,FILOPN
      INTEGER IDMAXU,OVRSHT,TRWRN,FORCE,THRTYP,Y0,M0,D0,UFILNO,NDAYS
      INTEGER DELTA(MXP2+1)
      INTEGER I,J,K,Q
******************************************************************************
*	Extra DOUBLEs - those marked as # are just dummies which are
*	~~~~~~~~~~~~~   required by other routines such as COVSET
*	OLDLOGL	- Log-likelihood from previous iteration
*       LLINVL  - `Invalid' log-likelihood
*       LLINC   - Increase in Log L on last iteration
*       DVNCE   - Deviance
*       LLTOL   - Value of new logL which is deemed significantly
*                 worse than the old one
*	TOL	- Small number used for determining convergence
*	XWX,XWZ	- The arrays X'WX and X'Wz in IWLS (see McCullagh &
*		  Nelder, 1989, for example)
*       XWSWX   - The filling the the sandwich, for the sandwich estimator
*                 of variance in the presence of spatial dependence.
*       SCORE   - Score vector
*       SUMSCR  - Sum of scores for current day
*       SCSTD   - Standard deviations of scores
*       MAXU    - Largest component of score vector
*	BETADJ	- Adjustment to BETA from this iteration
*       BETA0   - Starting value of BETA
*       LLR0    - Log likelihood ratio corresponding to BETA0
*       LLRADJ  - Adjusted log likelihood ratio
*       LAMBDA  - Multiple of step size to use
*       LLTAB   - Table of log-likelihoods for different values of LAMBDA
*	MU	- Fitted value for current case
*	SIGSQ	- Used to calculate dispersion parameters in some models.
*	ETA	- Linear predictor for the current case
*	Z	- Adjusted dependent variate for the current case
*       U       - Contribution to score vector from current case
*	W	- Weight for the current case
*       PHI     - Dispersion parameter (for computing covariance matrix
*                 of parameter estimates)
*       NUI0    - Nuisance parameter at starting point
*	Y	- Observed value for the current case
*	X	- Covariates for the current case
*#	RAIN	- Vector of rainfall amounts for today
*#	PREVIOUS- Rainfalls over past days. 
*	COVS	- Covariate values for current case.
*       SUMLNY  - Sum of Ln(Y)s - used in calculating full likelihood.
*       THRESH  - Threshold for dealing with small positive values
*	TRACE   - Ditto, treating them as `trace' values
*       MLNUI   - Maximum likelihood estimate of dispersion parameter
*       MIND    - Cholesky square root of `independence' information
*       MADJ    - Cholesky square root of `adjusted' information
*       HADJ    - Hessian of adjusted log-likelihood
*       DHD     - The matrix Delta %*% (HADJ^-1) %*% Delta', where
*                 Delta is a qxp matrix with a single 1 in each row,
*                 in the position held in DELTA. 
*       C       - Transformation matrix for adjusted LRT
*       NUMER   - Numerator in scaling factor to adjust for the fact that
*                 the transformed starting value doesn't necessarily
*                 maximise the restricted adjusted likelihood.
*       DENOM   - Denominator, similarly
*	TMP	- Temporary storage
******************************************************************************
      DOUBLE PRECISION OLDLOGL,DVNCE,TOL,LLTOL,LLINVL,LLINC
      DOUBLE PRECISION XWX(0:MXP2,0:MXP2),XWZ(0:MXP2)
      DOUBLE PRECISION XWSWX(0:MXP2,0:MXP2),SCORE(0:MXP2),MAXU
      DOUBLE PRECISION MU,SIGSQ,ETA,Z,W,U,NUI0,SUMSCR(0:MXP2)
      DOUBLE PRECISION PHI,SCSTD(0:MXP2),Y,X(0:MXP2),LLR0,LLRADJ
      DOUBLE PRECISION BETA0(0:MXP2),BETADJ(0:MXP2),LAMBDA(3),LLTAB(3)
      DOUBLE PRECISION RAIN(MXN2),PREVIOUS(MXN2,10)
      DOUBLE PRECISION THRESH,TRACE
      DOUBLE PRECISION MIND(0:MXP2,0:MXP2),MADJ(0:MXP2,0:MXP2)
      DOUBLE PRECISION HADJ(0:MXP2,0:MXP2),DHD(MXP2+1,MXP2+1)
      DOUBLE PRECISION C(0:MXP2,0:MXP2),COVS(MXN2,MXP2)
      DOUBLE PRECISION SUMLNY,MLNUI,NUMER,DENOM,TMP
******************************************************************************
*     Check dimensioning ...
*
      IF ( (MXN.NE.MXN2).OR.(MXP.NE.MXP2) ) THEN
       CALL PRGERR('  IWLS  ',' fit_both.f ',1)
      ENDIF
*
*	Initialise ... (NB set X(0) to 1 as it will always represent a
*	constant term in a model).
*
      TOL = 1.0D-4
      LLINVL = -1.0D32
      IF (NOBS.GT.100000) TOL=TOL*DBLE(NOBS)/1.0D5
      ITER = 0
      LOGL = LLINVL
      LLINC = -LOGL
      SUMLNY = 0.0D0
      X(0) = 1.0D0
      CNVRGE = 2
      OVRSHT = 0
      TRWRN = 0
      FILOPN = 0
      IF (MODEL.EQ.1) NUIPAR = -1.0D0
      MLNUI = NUIPAR
      CALL TAUDEF(GLBVAL,GLBCOD,NP(8),THRESH,TRACE,THRTYP,MXP)
*
*     If there's spatial dependence structure in the model, we'll
*     want to write daily sums of scores to a scratch file as we
*     go along. Start looking for a free file handle at 85.
*
      UFILNO = 85
      IF (SPMOD.NE.0) CALL SCROPN('SEQUENTIAL',
     +                              'UNFORMATTED',0,UFILNO)
      NDAYS = 0
*
*     Store the initial value of BETA, for e.g. Wald tests later
*     on.
*
      DO 20 J=0,P
       BETA0(J) = BETA(J)
 20   CONTINUE
*
*     We'll need to recalculate the PREVIOUS array if there are
*     any nonlinearities relating to previous days' rainfalls
*
      PRVRST = 0
      DO 30 J=NP(6)+1,NP(7)
       I = COVCODE(J)/1000
       IF ((I.GT.NP(3)).AND.(I.LE.NP(4))) THEN
        PRVRST = 1
        GOTO 31
       ENDIF
 30   CONTINUE
*
*	By the time we get here, all covariates are set so we only
*	need to recalculate things if there are nonlinear effects
*
 31   RECALC = 1
*
*	... and start iterating. We only update the
*	old log-likelihood when it was produced by an IWLS step
*	rather than a `corrected overshoot'. We also set
*       a threshold for judging whether we're going to 
*       overshoot.
*
 100  IF (OVRSHT.EQ.0) THEN
       OLDLOGL = LOGL
       LLTOL = OLDLOGL
      ENDIF
      LOGL = 0.0D0
      NDAYS = 0
      IF (SPMOD.NE.0) REWIND(UFILNO)
      READ(SCRFN1,REC=1) SITE,OY,OM,OD
      IF (TRACE.GT.0.0D0) SUMLNY = 0.0D0
*
*	Initialise X'WX and X'Wz for this iteration. For some models
*	we can speed things up by only calculating X'WX once. 
*	Exceptions are when the weights depend nontrivially on the
*	fitted values (eg. in logistic regression, and when there
*	are nonlinear parameters to estimate so the design matrix 
*	itself changes from iteration to iteration. The XRC flag 
*	indicates which case we're dealing with.
*
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
      IF ( (ITER.EQ.0).OR.(NP(7).GT.NP(6)).OR.(MODEL.EQ.1) ) THEN
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
       XRC = 1
      ELSE
       XRC = 0
      ENDIF
      MAXU = 0.0D0
      IDMAXU = 0
      DO 105 I=0,P
       XWZ(I) = 0.0D0
       SCORE(I) = 0.0D0
       SUMSCR(I) = 0.0D0
       IF (XRC.EQ.1) THEN
	DO 106 J=0,P
	 XWX(I,J) = 0.0D0
 106   CONTINUE
       ENDIF
 105  CONTINUE
*
*	Spit out current parameterisation (unless we're only here to 
*       write residual stuff to a scratch file)
*
      IF (RECALC.NE.2) CALL CHKMDL(BETA,SEBETA,THETA,SETHETA,GLBVAL,
     +     NUIPAR,RHO,COVCODE,TWO,THREE,GLBCOD,NP,NLSTAT,PWTIDX,SPMOD,
     +     TITLE,XLABELS,NLTXT,SPATXT,SPPTXT,CORTXT,ATTRTXT,GLBTXT,
     +     CNVRGE,MODEL,MXP)
*
*     Decide whether we're *going* to write residual stuff to scratch 
*     file
*
      IF ((DABS(LLINC).LT.1.0D2*TOL).OR.(RECALC.EQ.2).OR.
     +    ((MAXIT.GE.0).AND.(ITER.GE.MAXIT))) THEN
       RESWRT = 1
      ELSE
       RESWRT = 0
      ENDIF
*
*	Now start on likelihood etc. corresponding to current
*	parameter vector.
*
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
      IF (MODEL.NE.1) SIGSQ = 0.0D0
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
      TRWRN = 0
      DO 150 I = 1,NOBS
*
*       Compute linear predictor, fitted values etc. 
*
       CALL FITCAL(BETA,NP,P,MXP,MODEL,SCRFN1,I,ITER,TRWRN,TRACE,
     +                          MLNUI,LLINVL,X,Y,SITE,YY,MM,DD,
     +                          LOGL,MU,ETA,IFAIL)
       IF (IFAIL.EQ.1) GOTO 170
*
*	Now finish off contribution to adjusted dependent variate
*	and scores
*
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
       IF (MODEL.EQ.1) THEN
*
*	Logistic ... 
*
        W = MU * (1.0D0 - MU)
        U = (Y-MU)/W
       ELSEIF (MODEL.EQ.11) THEN
*
*	Gamma. 
*
        U = (Y - MU)/MU
        SIGSQ = SIGSQ + (U*U)
        W = 1.0D0
*
*     Need sum of log Y's for estimation of dispersion parameter
*     and calculation of deviance - the Ys may have changed if there 
*     are trace values present.
*
        IF ((ITER.EQ.0).OR.(TRACE.GT.0.0D0)) SUMLNY = SUMLNY + DLOG(Y)
       ENDIF
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
*
*       There's a bit to add on to Z if we're estimating nonlinear parameters
*       (ETA is the linear predictor, but the vector of first derivatives
*       has contributions from the derivatives of the nonlinear params
*       as well).
*
       Z = U + ETA
       IF (NP(7).GT.NP(6)) THEN
	DO 160 J=NP(6)+1,NP(7)
	 Z = Z + (BETA(J)*X(J))
 160    CONTINUE
       ENDIF
*
*	Calculate this case's contribution to X'Wz, and to
*	X'WX if this is necessary. (i.e. if the weights have
*	changed on this iteration - which happens for the logistic
*	model and if there are nonlinear transformations of 
*	predictors). Also update score vector for `independence'
*       model (it will already have been done if a spatial dependence
*       structure has been defined).
*
       Z = W*Z
       U = W*U
       DO 165 J=0,P
        XWZ(J) = XWZ(J) + (X(J)*Z)
        SCORE(J) = SCORE(J) + (X(J)*U)
        IF (XRC.EQ.1) THEN
 	 DO 166 K=J,P
	  XWX(J,K) = XWX(J,K) + ( X(J)*W*X(K) )
 166     CONTINUE
        ENDIF
 165   CONTINUE
*
*	Write prediction stuff to scratch file for use later on. We only 
*       do this if it looks like we're about to converge (or if we've 
*       already converged but this hasn't been done yet), so as not to 
*       waste a load of time with disk access.
*
       IF (RESWRT.EQ.1) THEN
        WRITE(SCRFN2,REC=I) SITE,YY,MM,DD,Y,MU,ETA
*
*     If there's spatial dependence structure in the model, write
*     daily sums of scores to scratch file
*
        IF (SPMOD.NE.0) THEN
         IF ((DD.NE.OD).OR.(MM.NE.OM).OR.(YY.NE.OY)) THEN
          WRITE(UFILNO) (SUMSCR(J),J=0,P)
          DO 157 J=0,P
           SUMSCR(J) = U*X(J)
 157      CONTINUE
          OD = DD
          OM = MM
          OY = YY
          NDAYS = NDAYS + 1
         ELSE
          DO 158 J=0,P
           SUMSCR(J) = SUMSCR(J) + (U*X(J))
 158      CONTINUE
          IF (I.EQ.NOBS) THEN
           WRITE(UFILNO) (SUMSCR(J),J=0,P)
           NDAYS = NDAYS + 1
          ENDIF
         ENDIF
        ENDIF
       ENDIF
 150  CONTINUE
*
*     Now calculate deviance and finish off likelihood calculations
*     for models that need it
*
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
      IF (MODEL.EQ.1) THEN
*
*     For logistic with cell sizes all equal to 1, log-likelihood
*     for saturated model is zero so deviance is just -2 log L
* 
       DVNCE = -(2.0D0*LOGL)
      ELSEIF (MODEL.EQ.11) THEN
*
*     For Gamma model, calculate ML estimate of dispersion parameter
*     for reporting likelihood, and moment estimate for subsequent
*     use (messy but necessary).
*
       DVNCE = -2.0D0*(LOGL + SUMLNY + DBLE(NOBS))
       NUIPAR = DBLE(NOBS - P - 1)/SIGSQ
       IF(.NOT.(NUIPAR.GT.-1.0D11)) NUIPAR = 0.0D0
       MLNUI = NUIPAR
       CALL MLNUES(MLNUI,SUMLNY,NOBS,LOGL)
      ENDIF
      IF (ITER.EQ.0) LLR0 = LOGL
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
*
*       Set RECALC flag to 2 if we've converged and, if we haven't yet
*       written the correct residual information to scratch file, go
*       back and do so (without changing any parameter values).
*
      LLINC = LOGL-OLDLOGL
      IF ( ((DABS(LLINC).LE.TOL).AND.(OVRSHT.EQ.0)).OR.
     +     ((ITER.EQ.MAXIT).AND.(LOGL.GE.LLTOL)).OR.
     +     ((ITER.GT.MAXIT).AND.(MAXIT.NE.-1)) ) THEN
       RECALC = 2
       IF (RESWRT.EQ.0) GOTO 100
      ENDIF
*
*	Now finish off X'WX and XWz. The aim is to end up with the Cholesky
*	decomposition of X'WX in XWX. If we didn't have to recalculate
*	XWX on this iteration, it will already contain the decomposition
*	from last time around, so nothing needs doing. Otherwise, just
*       the upper triangle has been calculated - fill in the rest here. 
*
      IF (XRC.EQ.1) THEN
       DO 183 I=0,P
        SCSTD(I) = DSQRT(XWX(I,I))
        IF (SPMOD.NE.0) XWSWX(I,I) = XWX(I,I)
        DO 184 J=0,I-1
         XWX(I,J) = XWX(J,I)
 184     CONTINUE
 183   CONTINUE
       CALL CHOLDEC(XWX,P+1,MXP+1,IFAIL)
       IF (IFAIL.NE.0) THEN
        WRITE(*,10) IFAIL
        STOP
       ENDIF
      ENDIF
*
*     Convergence check: calculate largest standardised element 
*     of score vector
*
      DO 187 I=0,P
       TMP=SCORE(I)/SCSTD(I)
       IF (DABS(TMP).GT.DABS(MAXU)) THEN
        MAXU = TMP
        IDMAXU = I+1
       ENDIF
 187  CONTINUE

*
*	Output summary for this iteration, and dispersion parameter
*	if necessary. We check for NaNs etc. in log-likelihoods, and set
*	them to something huge and negative.
*
 170  IF (.NOT.(LOGL.GT.LLINVL)) LOGL = LLINVL
      WRITE(*,2) ITER,LOGL,DVNCE,MAXU,IDMAXU,NOBS,P+1,NOBS-(P+1)
      WRITE(OUTFNO,6) ITER,LOGL
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
      IF (MODEL.EQ.11) WRITE(*,4) NUIPAR,MLNUI
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
*
*       Now another iteration if we haven't converged. If the last
*	iteration didn't take the log-likelihood below LLTOL, we proceed
*	in the usual way, by solving (X'WX)b = X'Wz for b. In this 
*	case we calculate X'Wz, then throw it, with the Cholesky factor 
*	of X'W'X, at a standard linear equation solver. Otherwise, we 
*       backtrack till we find an increase in log-likelihood (may slow
*	things down, but it guarantees convergence when looking for
*	nonlinear transformations).
*
      IF (RECALC.NE.2) THEN
       ITER = ITER + 1
       WRITE(*,'(//)')
       IF ( (ITER.LE.MAXIT).OR.(MAXIT.EQ.-1) ) THEN
        WRITE(*,'(''Beginning iteration '',I3)') ITER
       ELSE
        WRITE(*,'(''Recalculating best estimates found ...'')')
       ENDIF
       IF ( (LOGL.GT.LLTOL).OR.(ITER.LE.1).OR.
     +                         (OVRSHT.EQ.10) ) THEN
        OVRSHT = 0
        LAMBDA(1) = 0.0D0
        LAMBDA(2) = 1.0D0
        LAMBDA(3) = 1.0D0
        LLTAB(1) = LOGL
*
*       Equation solver. Store the adjustments that have been made.
*
        DO 200 I=0,P
         BETADJ(I) = BETA(I)
 200    CONTINUE
        CALL CDSOLV(XWX,BETA,XWZ,P+1,0,MXP+1)
        DO 205 I=0,P
         BETADJ(I) = BETA(I) - BETADJ(I)
 205    CONTINUE
       ELSE
*
*	This is the case when we overshot, so backtrack a bit (except
*       if we're resetting results of a final iteration, in which
*       case we backtrack the whole way). First we reset everything to 
*       original values, then compute reduced step.
*
        OVRSHT = OVRSHT + 1
        IF (OVRSHT.EQ.1) THEN
         LLTAB(2) = LOGL
        ELSE
         LLTAB(3) = LOGL
        ENDIF
        DO 210 I=0,P
         BETA(I) = BETA(I) - BETADJ(I)
         BETADJ(I) = BETADJ(I)/LAMBDA(3)
 210    CONTINUE
        IF ((ITER.LE.MAXIT).OR.(MAXIT.EQ.-1)) THEN
         IF (OVRSHT.EQ.1) THEN
          LAMBDA(3) = 0.5D0
         ELSE
*
*     Here's the backtracking. All being well, we use inverse
*     parabolic interpolation, but constrain ourselves to the
*     line segment within which we know a maximum must lie.
* 
          IF (LOGL.GT.LLINVL) THEN
           CALL INVPAR(LAMBDA,LLTAB,1)
           IF (LAMBDA(3).LT.0.0D0) THEN
            LAMBDA(3) = LAMBDA(2)/1.0D1
           ELSEIF (LAMBDA(3).GT.LAMBDA(2)) THEN
            LAMBDA(3) = LAMBDA(2)/2.0D0
           ENDIF
          ELSE
           LAMBDA(3) = LAMBDA(3)/1.0D2
          ENDIF
         ENDIF
         WRITE(*,'(''Reducing step size to '',F7.3,''% of original'')')
     +        1.0D2*LAMBDA(3)
        ELSE
         LAMBDA(3) = 0.0D0
        ENDIF
        DO 211 I=0,P
         BETADJ(I) = LAMBDA(3)*BETADJ(I)
         BETA(I) = BETA(I) + BETADJ(I)
 211    CONTINUE
       ENDIF
*
*       Update nonlinear parameters if there are any, and if we're
*	iterating (doesn't work if MAXIT=0)
*
       IF ( (NP(7).GT.NP(6)).AND.(MAXIT.NE.0) ) THEN
        CALL NLUPDT(THETA,BETA,SETHETA,SEBETA,NP,COVCODE,MXP)
*
*       Also the matrix X. We use the routine COVSET again to recalculate X; 
*       this requires that we set some elements of COVS to their current 
*       values, as RECALC is nonzero and only those elements that have
*       changed will be reset in COVSET - if we don't pass the correct 
*       values across, the interactions will be buggered on exit. Another 
*       tricky feature is that COVSET updates a whole days' rainfall at 
*       once. This is a pain in the bum for this program but very useful 
*       when the routine is used for simulation. Hence the inelegancies 
*       to do with tracking dates etc! We do it by reading in elements of 
*       the RAIN and COVS arrays, modifying as appropriate, then rewriting 
*       them where RAIN is non-missing (i.e. skip over sites with missing 
*       data for the current day). As the site attributes may also have 
*       changed, we recalculate them as well (may not be necessary but it 
*       doesn't take any time).
*
        CALL ATTRXFM(TITLE,SITINF,NSITES,ATTRTXT,NATTR,COVCODE,
     +                   SITXFM,FOUIDX,LEGIDX,THETA,NP,1,MXN,MXP)

        DO 300 I=1,NSITES
         MISSFL(I) = 1
         IF (PRVRST.EQ.1) THEN
          DO 301 J=1,MXLAG
           PREVIOUS(I,J) = -1.0D0
 301      CONTINUE
         ENDIF
 300    CONTINUE

*
*     If we have to recalculate the PREVIOUS array, we'll have to 
*     reopen the gauge input file
* 
        IF (PRVRST.EQ.1) THEN
         Y0 = 0
         M0 = 0
         D0 = 0
         IF (FILOPN.EQ.0) THEN
          CALL FILCHK(FILNM,FILNO,1)
          FILOPN = 1
         ELSE
          REWIND(FILNO)
         ENDIF
        ENDIF

        READ(SCRFN1,REC=1) SITE,YY,MM,DD
        I = 0
        FORCE = 1
 99     IF (I.LT.NOBS) READ(SCRFN1,REC=I+1) 
     +	 		(CASEID(J),J=1,4),Y,(X(J),J=1,P)
        IF ((CASEID(2).EQ.YY).AND.
     +      (CASEID(3).EQ.MM).AND.
     +      (CASEID(4).EQ.DD).AND.(I.LT.NOBS) ) THEN
	 I = I+1
         SITE = CASEID(1)
         MISSFL(SITE) = 0
*
*	NB we are NOT ALLOWED to change year or month effects
*	in COVS between successive calls to COVSET, if  COVSET is 
*       going to recalculate them on this pass because of a change
*       in the nonlinear parametrisation. (Possible source of much 
*       headache, already realised). We *DO*, however, have to 
*       explicitly set things here if COVSET's going to skip
*       over them. See COVSET header (file glm_base.f) for details.
*       Here, we set year effects if there's no nonlinear transformation
*       involved; we *always* set all other effects (currently, there
*       are no nonlinear transformations of month effects so COVSET
*       will skip over them all).
*
	 DO 310 J=1,NP(1)
          COVS(SITE,J) = X(J)
 310     CONTINUE
         DO 311 J=NP(1)+1,NP(2)
          IF(THETA(J,1).GT.1.0D8) COVS(SITE,J) = X(J)
 311     CONTINUE
	 DO 312 J=NP(2)+1,P
          COVS(SITE,J) = X(J)
 312     CONTINUE
         IF (PRVRST.EQ.0) RAIN(SITE) = Y
        ELSE
         IF (PRVRST.EQ.1) THEN
 315      CALL UPDATE(FILNO,RAIN,PREVIOUS,THRESH,THRTYP,NSITES,
     +                             Y0,M0,D0,MXLAG,MXN,SCODES,FILOPN)
          IF ((Y0.NE.YY).OR.(M0.NE.MM).OR.(D0.NE.DD)) GOTO 315
         ENDIF
	 CALL COVSET(COVS,NP,NSITES,COVCODE,TWO,THREE,GLBCOD,
     +        SITINF,YY,MM,DD,PREVIOUS,PWTIDX,MXLAG,GLBVAL,BETA,
     +	      THETA,TRACE,RECALC,FORCE,MISSFL,1,MXN,MXP,0,0)
         FORCE = 0
*
*	Write the modified records to the data file (backwards -
*	we know the current record is I so we'll start there and
*	write to the previous line wherever there's a valid RAIN
*	value
*
         RECNO = I
	 DO 320 SITE=NSITES,1,-1
	  IF (MISSFL(SITE).EQ.0) THEN
	   DO 321 J=1,P
	    X(J) = COVS(SITE,J)
 321       CONTINUE
           Y = RAIN(SITE)
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
* For binary outcomes, responses may need resetting to 0/1.
*
           IF ((MODEL.EQ.1).AND.(Y.GT.0.0D0)) Y = 1.0D0 
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
           WRITE(SCRFN1,REC=RECNO) SITE,YY,MM,DD,Y,(X(J),J=1,P)
           RECNO = RECNO - 1
	  ENDIF
          MISSFL(SITE) = 1
 320     CONTINUE
         IF (I.LT.NOBS) THEN
          READ(SCRFN1,REC=I+1) SITE,YY,MM,DD
         ELSE
          GOTO 330
         ENDIF
        ENDIF
        IF (I.LE.NOBS) GOTO 99       
       ENDIF
*
*     If we're in here we haven't converged, so go back
*
 330   GOTO 100
      ENDIF

***************************************************************************
*	Only reach here when we've converged. Check that we've written
*       all the necessary residual information (if we haven't, it's
*       a programming error).
*
      IF (RESWRT.NE.1) THEN
       WRITE(*,7)
       STOP
      ENDIF
      WRITE(OUTFNO,2) ITER,LOGL,DVNCE,MAXU,IDMAXU,NOBS,
     +                                       P+1,NOBS-(P+1)
*
*     Here's the dispersion parameter, for estimating standard errors.
*     If a spatial dependence structure has been defined, we'll use
*     a sandwich estimator for the covariance matrix of estimates; in 
*     this case the dispersion parameter estimate gets implicitly 
*     incorporated into the matrix, so no need to scale subsequently.
*     Hence set it to 1 unless SPMOD is 0. 
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
      PHI = 1.0D0
      IF (MODEL.EQ.11) THEN
       NUIPAR = DBLE(NOBS - P - 1)/SIGSQ
       IF(.NOT.(NUIPAR.GT.-1.0D11)) NUIPAR = 0.0D0
       WRITE(OUTFNO,4) NUIPAR,MLNUI
       IF (SPMOD.EQ.0) PHI = 1.0D0/DSQRT(NUIPAR)
      ENDIF
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
      CNVRGE = 3
      LLR0 = 2.0D0*(LOGL-LLR0)
      WRITE(*,3) LLR0
      WRITE(OUTFNO,3) LLR0
*
*	Calculate vector of standard errors - each is the square root of
*	the appropriate diagonal element of [X'WX]^-1. As it stands, 
*	XWX contains the Cholesky factor of the matrix X'WX (apart from
*       a factor relating to the dispersion parameter). If a spatial 
*       dependence structure has been defined, adjust the standard errors 
*       and calculate the adjusted increase in log-likelihood from the 
*       supplied starting values - need to calculate the matrix 
*       C=[(MIND)^-1][MADJ] for this, where MIND is the Cholesky factor
*       of the matrix X'WX and MADJ =  [R^-1]^(1/2). Here, R is the `robust'
*       variance estimate, held in XWX after calling SNDWCH. We also
*       compute the Hessian -R^-1, since we'll need it later.
*
      WRITE(*,'(''Computing covariance matrix of estimates ...'')')
      IF (SPMOD.NE.0) CALL UTINV(XWX,MIND,P+1,MXP+1,MXP2+1)
      CALL MATINV(XWX,P+1,0,MXP+1,1)
      IF (SPMOD.NE.0) THEN
       CALL SNDWCH(XWX,XWSWX,UFILNO,P,NDAYS,MXP)
       DO 410 I=0,P
        DO 411 J=0,P
         MADJ(I,J) = XWX(I,J)
 411    CONTINUE
 410   CONTINUE
       CALL MATINV(MADJ,P+1,1,MXP2+1,1)
       DO 412 I=0,P
        DO 413 J=0,P
         HADJ(I,J) = -MADJ(I,J)
 413    CONTINUE
 412   CONTINUE
       CALL CHOLDEC(MADJ,P+1,MXP2+1,IFAIL)
*
*     The matrices to be multiplied are both upper triangular,
*     which saves some time. NB the result is therefore also 
*     upper triangular. NB also: because of the way this has been
*     programmed, the *robust* variance estimate is scaled 
*     correctly but the *naive* one is out by a factor of SQRT(NU).
*     See comment about 30 lines back. Only took about 3 days to 
*     realise this - moral of story is read your comments!
*
       DO 420 I=0,P
        DO 421 J=0,P
         C(I,J) = 0.0D0
         IF (J.GE.I) THEN
          DO 422 K=I,J
           C(I,J) = C(I,J) + (MIND(I,K)*MADJ(K,J))
 422      CONTINUE
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
          IF (MODEL.EQ.11) C(I,J) = C(I,J)/DSQRT(NUIPAR)
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
         ENDIF
 421    CONTINUE
 420   CONTINUE
*
*     Identify where the zeroes were in the starting value - it's
*     assumed that we want to test the hypothesis that these 
*     covariates shouldn't be in the model. NB we only consider
*     the coefficients in the linear predictor here - nonlinear
*     parameters are not considered. Then centre the starting
*     value about the final estimate.
*
       Q = 0
       DO 450 I=0,P
        IF ((I.LE.NP(6)).AND.(DABS(BETA0(I)).LT.1.0D-12)) THEN
         Q = Q + 1
         DELTA(Q) = I
        ENDIF
        BETA0(I) = BETA0(I) - BETA(I)
 450   CONTINUE
 
*
*     Now compute the transformed starting value - store in BETADJ,
*     since it's no longer being used (and it *is* an adjusted 
*     BETA, after all!).
*
       DO 460 I=0,P
        BETADJ(I) = BETA(I)
        DO 461 J=0,P
         BETADJ(I) = BETADJ(I) + (C(I,J)*BETA0(J))
 461    CONTINUE
 460   CONTINUE
*
*     Here's the independence log-likelihood at the transformed
*     value
*
       WRITE(*,'(A50)') 
     +          'Computing dependence-adjusted likelihood ratio ...'
       LLRADJ = 0.0D0
       DO 470 I=1,NOBS
        CALL FITCAL(BETADJ,NP,P,MXP,MODEL,SCRFN1,I,ITER,TRWRN,0.0D0,
     +                          MLNUI,LLINVL,X,Y,SITE,YY,MM,DD,
     +                          LLRADJ,MU,ETA,IFAIL)
 470   CONTINUE
       IF (Q.GT.0) THEN 
*
*     Now calculate the scaling factor to adjust for the fact that 
*     the transformed starting value doesn't necessarily maximise
*     the independence log-likelihood subject to the transformed
*     parameter constraint. First the numerator: here, the matrix
*     DHD consists of elements of HADJ^-1, which is actually 
*     just -XWX at this stage. To exploit positive definiteness
*     of XWX, however, use -HADJ^-1 and correct the minus sign at 
*     the end.
*
        DO 480 I=1,Q
         DO 481 J=1,Q
          DHD(I,J) = XWX(DELTA(I),DELTA(J))
 481     CONTINUE
 480    CONTINUE
        CALL MATINV(DHD,Q,1,MXP2+1,1)
        NUMER = 0.0D0
        DO 482 I=1,Q
         DO 483 J=1,Q
          NUMER = NUMER - (DHD(I,J)*BETA(DELTA(I))*BETA(DELTA(J)))
 483     CONTINUE
 482    CONTINUE
*
*     Now the denominator, which is BETA0' %*% HADJ %*% BETA0 (recall 
*     that BETA0 currently holds the difference between the starting
*     value and the final estimate
*
       DENOM = 0.0D0
       DO 485 I=0,P
        DO 486 J=0,P
         DENOM = DENOM + (HADJ(I,J)*BETA0(I)*BETA0(J))
 486    CONTINUE
 485   CONTINUE
       ELSE
        NUMER = 1.0D0
        DENOM = 1.0D0
       ENDIF
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
       IF (MODEL.EQ.11) THEN
        NUI0 = NUIPAR
        CALL MLNUES(NUI0,SUMLNY,NOBS,LLRADJ)
       ENDIF
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
       LLRADJ = 2.0D0*(LOGL-LLRADJ)*NUMER/DENOM
       WRITE(*,1) LLRADJ
       WRITE(OUTFNO,1) LLRADJ
       IF (Q.GT.0) THEN
        WRITE(*,9) NUMER/DENOM 
        WRITE(OUTFNO,9) NUMER/DENOM 
       ENDIF
      ENDIF
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
      IF ((MODEL.EQ.11).AND.(TRACE.GT.0.0D0)) THEN
       WRITE(*,8)
       WRITE(OUTFNO,8)
      ENDIF 
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
      WRITE(OUTFNO,'(//)')
      WRITE(*,'(//)')
      IF (ITER.EQ.MAXIT) WRITE(OUTFNO,5)
      CLOSE(SCRFN1)
      IF (SPMOD.NE.0) CLOSE(UFILNO)
      DO 500 I=0,P
       SEBETA(I) = PHI*DSQRT(XWX(I,I))
 500  CONTINUE
*
*	Copy standard errors of THETAs to appropriate positions in
*	the SETHETA array
*
      IF (NP(7).GT.NP(6)) THEN
       CALL NLUPDT(THETA,BETA,SETHETA,SEBETA,NP,COVCODE,MXP)
      ENDIF
*
*     If TRACE > 0, pass ML estimate of nuisance parameter back, 
*     as small changes can have a big impact on the initial 
*     log-likelihood for future fits, through the replacement
*     of trace values by their conditional expectations
*
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
      IF ((MODEL.EQ.11).AND.(TRACE.GT.0.0D0)) NUIPAR = MLNUI
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////

 1    FORMAT(5X,'Dependence-adjusted LR statistic - ',T55,F10.4)
 2    FORMAT(//5X,'Results after ',I3,' iterations:'/5X,
     +'Log-likelihood - ',T35,F14.3,/5X,
     +'Deviance - ',T35,F14.3,/5X,
     +'Largest standardised score - ',T39,F10.4,
     +2X,'(parameter ',I3,')',/5X,
     +'Number of observations - ',T35,I14,/5X,
     +'No. of parameters estimated - ',T35,I14,/5X,
     +'Residual degrees of freedom - ',T35,I14)
 3    FORMAT(5X,'Likelihood ratio statistic (initial vs final) - ',
     +T55,F10.4)
 4    FORMAT(5X,'Dispersion parameter (NU) - ',F9.6,1X,
     +          '(ML estimate is ',F9.6,')')
 5    FORMAT(/5X,'****WARNING**** Reached maximum ',
     +'number of iterations.')
 6    FORMAT(5X,'Iteration ',I4,': Log-likelihood=',F14.3)
 7    FORMAT(5X,'****ERROR**** Scratch file #2 not written properly ',
     +'in routine',/,'IWLS. This is a programming error - contact REC.')
 8    FORMAT(/5X,'****WARNING**** Computation of likelihood ratio ',
     +'statistics may be',/5X,'inaccurate, because trace values ',
     +'have been replaced by their',
     +/5X,'approximate conditional expectations.')
 9    FORMAT(8X,'NB incorporates secondary adjustment for',/8X,
     +'hypothesis testing, since initial coefficient',/8X,
     +'vector contained zeroes: scale factor was',T58,F7.5)
 10   FORMAT(/5X,'****ERROR**** in routine IWLS, design matrix seems ',
     +'singular.',/5X,'Problem found on row ',I3,'. Have you ',
     +'accidentally duplicated a covariate',/5X,'that was already ',
     +'defined in another way, or specified a covariate that',/5X,
     +'doesn''t vary in your data set?')
 
      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE NLUPDT(THETA,BETA,SETHETA,SEBETA,NP,COVCODE,MXP)
*
*       Updates unknown nonlinear parameters if we're estimating
*       them, at the end of each iteration (just copies them across
*       from BETA to appropriate bits of THETA). Variables:
*       THETA   - nonlinear parameters
*       BETA    - coefficients of linear predictors
*       SETHETA - Std Errors of THETA
*       SEBETA  - Ditto, BETA
*       NP      - Indexing for parameters
*       COVCODE - indicates which bits of THETA we're updating
*       MXP,MXN - for dimensioning
******************************************************************************
      INTEGER MXP,COVCODE(MXP),NP(10)
      DOUBLE PRECISION BETA(0:MXP),THETA(MXP,3)
      DOUBLE PRECISION SEBETA(0:MXP),SETHETA(MXP,3)
******************************************************************************
*       Additional INTEGERs
*       ^^^^^^^^^^^^^^^^^^^
*       I       - counter
*       IDX?    - used to locate correct element of THETA from COVCODE
******************************************************************************
      INTEGER I,IDX1,IDX2

*
*       THETAs are easy to locate because of the way COVCODE was
*       set up in MDLSET (NB integer division). IDX1 is the 
*       covariate, IDX2 the parameter. We check that parameters
*       aren't going off to infinity, and bail out if they are.
*
      DO 100 I=NP(6)+1,NP(7)
       IDX1 = COVCODE(I)/1000
       IDX2 = COVCODE(I) - (1000*IDX1)
       IF (DABS(BETA(I)).GT.1.0D8) THEN
	WRITE(*,1) IDX2,IDX1
	STOP
       ENDIF
       THETA(IDX1,IDX2) = BETA(I)
       SETHETA(IDX1,IDX2) = SEBETA(I)
 100  CONTINUE

 1    FORMAT(/5X,'****ERROR**** During iteration, the estimate of ',
     +'parameter ',I1,/5X,'in the nonlinear transformation of ',
     +'covariate no. ',I2,/5X,'is approaching infinity. This isn''t',
     +' going to work - I''m bailing out.',/5X,
     +'Try different starting values.')
      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE NEWMDL(INFLNO,OUTFNO,MXP,NP,BETA,THETA,GLBVAL,NU,
     +      RHO,COVCODE,GLBCOD,MXLAG,SITXFM,TWO,THREE,NLSTAT,PWTIDX,
     +      SPMOD,TITLE,MDLTXT,GLBTXT)
******************************************************************************
*       To output a model specification, in standard format, to file
*       OUTFL. 
******************************************************************************
*       INTEGER variables
*       ^^^^^^^^^^^^^^^^^
*       INFLNO  - Channel number of input file
*       OUTFNO  - Channel number of output file
*       MXP     - Maximum allowable number of parameters (used for
*                 dimensioning arrays)
*       NP      - Numbers of parameters in various components of the model.
*                 See main program header for more details
*       COVCODE - Code assiciated with each covariate - used to define
*                 transformations etc.
*       GLBCOD  - Codes defining global quantities
*       SITXFM  - indicates which nonlinear functions of site properties
*                 have been selected. 
*       TWO  - Indicates two-way interactions
*       THREE- Indicates three-way interactions
*	MXLAG	- See main program header
*       NLSTAT  - indicates which nonlinear parameters are known and which 
*                 are to be estimated
*       PWTIDX  - indexing of parameters in weighted averages of previous
*                 days' values
*	SPMOD	- choice of spatial dependence structure
*       COMPNT  - Indicates which component of the model this record's to do 
*                 with (i.e. how far along NP we are)
*       NHEAD   - No. of header lines in file
*       CODE    - Temporary storage for current predictor's codes
*       WTSCHM  - Weighting scheme for averaging previous days' values
******************************************************************************
      INTEGER MXP,NP(10),COVCODE(MXP),NLSTAT(MXP,3),PWTIDX(MXP,0:3)
      INTEGER INFLNO,OUTFNO
      INTEGER GLBCOD(MXP),SITXFM(MXP,2),TWO(MXP,2),THREE(MXP,3)
      INTEGER COMPNT,MXLAG,SPMOD,NHEAD,CODE(3),WTSCHM
      INTEGER I,J
******************************************************************************
*       DOUBLE PRECISION variables
*       ^^^^^^^^^^^^^^^^^^^^^^^^^^
*       BETA    - parameter vector for the model
*	THETA	- nonlinear parameters
*       GLBVAL  - Values of global quantities
*	NU	- dispersion parameter (if < 0 it's ignored)
*	RHO	- parameters in spatial dependence
******************************************************************************
      DOUBLE PRECISION BETA(0:MXP),THETA(MXP,3),GLBVAL(MXP)
      DOUBLE PRECISION NU,RHO(MXP)
******************************************************************************
*       CHARACTER variables
*       ^^^^^^^^^^^^^^^^^^^
*       TITLE   - Title of model
*       HEADLN  - Line of header in input and output file
*       MDLTXT  - descriptive text for model
*       GLBTXT  - descriptive text for global quantities
******************************************************************************
      CHARACTER TITLE*70,HEADLN*80
      CHARACTER*70 MDLTXT(0:MXP),GLBTXT(MXP)

      NHEAD = 46
*
*       Write file header (which is NHEAD lines long). 
*
      REWIND (INFLNO)

      DO 10 I = 1,NHEAD
       READ(INFLNO,'(A80)') HEADLN
       WRITE(OUTFNO,'(A80)') HEADLN
 10   CONTINUE
      CLOSE (INFLNO)
*
*       Obsolete line, reserved for future use
*
      WRITE(OUTFNO,'(20(''%''),''LINE NOT CURRENTLY USED'',20(''%''))')
*   
*       Write model description
*
      WRITE(OUTFNO,'(A70)') TITLE
*
*       And model specification. First constant:
*
      COMPNT = 0
      WRITE(OUTFNO,'(I5,F10.4,15X,A40)') COMPNT,BETA(0),MDLTXT(0)
*
*       Now site effects. There may be 2 codes here - if not, the second
*       one is a zero. it's not a nonlinear xformation of site
*       attributes and COVCODE contains the correct value. If
*       SITXFM(I,2) isn't zero, we're working with something nonlinear,
*       COVCODE *doesn't* contain the correct value and we have
*       to use SITXFM(I,1) in its place. A bit messy ...
*
      COMPNT = 1
      DO 20 I=1,NP(1)
       IF (SITXFM(I,2).EQ.0) THEN
	WRITE(OUTFNO,'(I5,F10.4,I5,10X,A40,1X,I3)') 
     +                  COMPNT,BETA(I),COVCODE(I),MDLTXT(I),I
       ELSE
	WRITE(OUTFNO,'(I5,F10.4,2I5,5X,A40,1X,I3)') 
     +         COMPNT,BETA(I),(SITXFM(I,J),J=1,2),MDLTXT(I),I
       ENDIF
 20   CONTINUE
*
*       Year and month effects have only 1 argument, unless there 
*       are external factors involved
*
      DO 30 COMPNT=2,3
       DO 31 I=NP(COMPNT-1)+1,NP(COMPNT)
        IF (MOD(IABS(COVCODE(I)),1000).LE.50) THEN
	 WRITE(OUTFNO,'(I5,F10.4,I5,10X,A40,1X,I3)') 
     +                   COMPNT,BETA(I),COVCODE(I),MDLTXT(I),I
        ELSE
	 WRITE(OUTFNO,'(I5,F10.4,2I5,5X,A40,1X,I3)') 
     +                   COMPNT,BETA(I),MOD(IABS(COVCODE(I)),1000),
     +                   COVCODE(I)/1000,MDLTXT(I),I
        ENDIF
 31    CONTINUE
 30   CONTINUE
*
*       Day effects are a bit complicated - they're split into
*       `Previous days' and `external':
*
      COMPNT = 4
      CODE(3) = -1
      DO 36 I=NP(3)+1,NP(4)
       CODE(1) = MOD(IABS(COVCODE(I)),1000)
*
*     Here are the previous days - write MXLAG out for the first one.
*
       IF (CODE(1).LE.10) THEN
        CODE(2) = COVCODE(I)/1000
        IF (CODE(3).LT.0) THEN
         CODE(3) = MXLAG
	 WRITE(OUTFNO,'(I5,F10.4,3I5,A40,1X,I3)') 
     +        COMPNT,BETA(I),CODE,MDLTXT(I),I
        ELSE
	 WRITE(OUTFNO,'(I5,F10.4,2I5,5X,A40,1X,I3)') 
     +        COMPNT,BETA(I),CODE(1),CODE(2),MDLTXT(I),I
        ENDIF
*
*     Here are external factors
*
       ELSE
        IF (CODE(1).LE.50) THEN
         WRITE(OUTFNO,'(I5,F10.4,I5,10X,A40,1X,I3)')
     +                 COMPNT,BETA(I),CODE(1),MDLTXT(I),I
        ELSE
         CODE(2) = COVCODE(I)/1000
	 WRITE(OUTFNO,'(I5,F10.4,2I5,5X,A40,1X,I3)') 
     +                   COMPNT,BETA(I),CODE(1),CODE(2),
     +                   MDLTXT(I),I
        ENDIF
       ENDIF
 36   CONTINUE
*
*       For interactions, output appropriate codes; abandon MDLTXT
*
      COMPNT = 5
      DO 40 I=NP(4)+1,NP(5)
       WRITE(OUTFNO,'(I5,F10.4,2I5,5X,A30,I2,A5,I2)')
     +          COMPNT,BETA(I),TWO(I,1),TWO(I,2),
     +          '2-way interaction: covariates ',
     +          TWO(I,1),' and ',TWO(I,2)
 40   CONTINUE
      COMPNT = 6
      DO 50 I=NP(5)+1,NP(6)
       WRITE(OUTFNO,'(I5,F10.4,3I5,A30,I2,A2,I2,A5,I2)') 
     +          COMPNT,BETA(I),THREE(I,1),THREE(I,2),THREE(I,3),
     +          '3-way interaction: covariates ',
     +          THREE(I,1),', ',THREE(I,2),' and ',THREE(I,3)
 50   CONTINUE
*
*       Parameters in nonlinear transformations. Rather than look at 
*       definitions in positions NP(6) to NP(7), it's easier to go 
*       through the THETAs themselves, directly (if you don't believe me,
*       try it and see what happens!). One slight complication:
*       for weighted averages of previous days' values, some
*       elements of THETA have been used as as temporary storage -
*       so we have to check that the current row genuinely requires
*       something to be output.
*
      COMPNT = 7
      DO 60 I=1,NP(4)
       WTSCHM = MOD(COVCODE(I)/10000,10)
       IF ((I.GT.NP(3)).AND.(WTSCHM.GT.1)) THEN
        IF (I.NE.PWTIDX(WTSCHM,0)) GOTO 60
       ENDIF
       DO 61 J=1,3
	IF (THETA(I,J).LT.1.0D8) THEN
	 WRITE(OUTFNO,'(I5,F10.4,3I5,A10,I1,A27,I3)') 
     +          COMPNT,THETA(I,J),I,J,NLSTAT(I,J),'Parameter ',
     +                        J,' in xfrmation of covariate ',I
	ENDIF
 61    CONTINUE
 60   CONTINUE
*
*     Global quantities, if present
*
      COMPNT = 8
      DO 65 I=1,NP(COMPNT)
       IF (GLBCOD(I)/1000.EQ.1) THEN
        WRITE(OUTFNO,'(I5,F10.4,2I5,5X,A39)') COMPNT,GLBVAL(I),
     +                   GLBCOD(I)/1000,MOD(GLBCOD(I),1000),GLBTXT(I)
       ENDIF
 65   CONTINUE
*
*	Dispersion parameter, if required
*
      COMPNT = 9
      IF (NU.GT.0.0D0) THEN
       WRITE(OUTFNO,'(I5,F10.4,15X,A20)') COMPNT,NU,
     +                                  'Dispersion parameter'
      ENDIF
*
*	And spatial structure
*
      COMPNT = 10
      IF (SPMOD.EQ.1) THEN
       WRITE(OUTFNO,'(I5,F10.4,I5,10X,A39)') COMPNT,0.0D0,SPMOD,
     +		'Observed residual correlation structure'
      ELSE       
       DO 70 I=1,NP(10)
        WRITE(OUTFNO,'(I5,F10.4,2I5,5X,A10,I2,A29)') COMPNT,RHO(I),
     +	       SPMOD,I,'Parameter ',I,' in spatial dependence model'
 70    CONTINUE
      ENDIF
      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE EMPCOR(FILNO,NSITES,NOBS,MODEL,RESTYP,DEPTYP,
     +                                           CORMAT,CORRN,MXN)
*
*   Calculates an empirical spatial correlation matrix or matrix of
*   joint occurrence probabilities, from data held in file connected 
*   to unit FILNO. Arguments:
*	
* FILNO  (INTEGER, input) Channel number of file to read
*	 data from. The file is a direct access file: each 
*	 record contains 4 integers (site, year, month and
*        day), then 4 doubles (observed and expected rain,
*	 Pearson residual and something that doesn't concern
*        us here).
* NSITES Number of sites used.
* NOBS	 Total number of records in file FILNO
* MODEL  Type of model. 1=logistic regression, 11=gamma.
* RESTYP (INTEGER, input). Codes for measure to compute 
*        correlations over, as follows:
*	    0 - raw rainfall values
*	    1 - Pearson residuals
*	    2 - Anscombe residuals for gamma observations
* DEPTYP (INTEGER, input). Type of dependence (1 = covariance,
* 	 2 = correlation, 3 = joint occurrence probabilities for rainfall
*        occurrence)
* CORMAT (DOUBLE, output). Correlation matrix.
* CORRN	 (DOUBLE,output). Ns associated with entries of CORMAT.
* MXN	 (INTEGER, input). Physical dimension of CORMAT
*	 (MXN x MXN).
******************************************************************************
*	PARAMETERs
*	^^^^^^^^^^
*	MXN2	Maximum number of sites we'll need. Set it equal
*		to MXN.
******************************************************************************
      INTEGER MXN2
      PARAMETER (MXN2=200)
      INTEGER FILNO,NSITES,NOBS,MODEL,RESTYP,DEPTYP,MXN
      DOUBLE PRECISION CORMAT(MXN,MXN),CORRN(MXN,MXN)
******************************************************************************
*	Additional INTEGERs
*	^^^^^^^^^^^^^^^^^^^
*       I,J,K	- Counters
*       CASEID  - array of identifiers for each case, for purposes of
*                 residual analysis. Col.1 contains site number, 2 
*                 contains year, 3 contains month and 4 day
*	    OLDDAT }- Used to determine when the date changes
*	    NEWDAT }  
*       OLDRES  - Value of RESTYP on previous call
*       OLDDEP  - Value of DEPTYP on previous call
*       OLDFNO  - Value of FILNO on previous call
*       CORFNO  - Number of scratch file used to store the correlations
*                 when we've computed them
***************************************************************************
      INTEGER I,J,K,CASEID(4),OLDDAT,NEWDAT
      INTEGER OLDRES,OLDDEP,OLDFNO,CORFNO
******************************************************************************
*	Additional DOUBLEs
*	^^^^^^^^^^^^^^^^^^
*   Y       Daily rainfall
*   MU      Predicted MEANS
*   ONEDAY  Vector of today's values
*   MEAN    Means for calculating correlations. (i,j)th entry is
*           the mean appropriate for site i when calculating
*           correlation with site j.
*   VAR     Variances, similarly
*   TMP     Temporary storage
*   TOL	    Small number such that variances are treated as zero
*		    if they're less than it.
******************************************************************************
      DOUBLE PRECISION Y,MU
      DOUBLE PRECISION ONEDAY(MXN2)
      DOUBLE PRECISION MEAN(MXN2,MXN2),VAR(MXN2,MXN2)
      DOUBLE PRECISION TMP,TOL
***************************************************************************
      DATA CORFNO,OLDRES,OLDDEP,OLDFNO /4*-1/
      SAVE CORFNO,OLDRES,OLDDEP,OLDFNO
*
*	Check dimensioning:
*
      IF (MXN2.NE.MXN) CALL PRGERR(' EMPCOR ',' fit_both.f',1)
******************************************************************************
*     First scenario: we haven't done this before so we have to calculate 
*     everything
******************************************************************************
      IF ((RESTYP.NE.OLDRES).OR.(DEPTYP.NE.OLDDEP).OR.
     +    (FILNO.NE.OLDFNO)) THEN
*
*     Open scratch file for final matrix
*
       IF (CORFNO.NE.-1) THEN
        CLOSE(CORFNO)
        CORFNO = -1
       ENDIF
       OLDRES = RESTYP
       OLDDEP = DEPTYP
       OLDFNO = FILNO
       CORFNO = 85
       CALL SCROPN('SEQUENTIAL','UNFORMATTED',0,CORFNO)
*
*       Initialise arrays (easier this way than in a DATA statement)
*
       TOL = 1.0D-12
       OLDDAT = 0
       DO 90 I=1,NSITES
        ONEDAY(I) = -1.0D12
        DO 91 J=1,NSITES
         CORMAT(I,J) = 0.0D0
         CORRN(I,J) = 0.0D0
         MEAN(I,J) = 0.0D0
         VAR(I,J) = 0.0D0
 91     CONTINUE
 90    CONTINUE
*
*       Read data & compute the requested measure. Going to NOBS+1
*	    guarantees we include the final day's measurements
*
       DO 100 I=1,NOBS+1
        IF (I.LE.NOBS) THEN
         READ(FILNO,REC=I) (CASEID(J),J=1,4),Y,MU
         IF (RESTYP.EQ.0) THEN
          TMP = Y
         ELSEIF (RESTYP.EQ.1) THEN
          IF (MODEL.EQ.1) THEN
           TMP = (Y-MU)/DSQRT(MU*(1.0D0-MU))
          ELSEIF (MODEL.EQ.11) THEN
           TMP = (Y-MU)/MU
          ELSE
           WRITE(*,1) ' MODEL'
           STOP
          ENDIF
         ELSEIF (RESTYP.EQ.2) THEN
          IF (MODEL.EQ.11) THEN
           TMP = ( (Y/MU)**(1.0D0/3.0D0) )
          ELSE
           WRITE(*,1) ' MODEL'
           STOP
          ENDIF
         ELSE
          WRITE(*,1) 'RESTYP'
          STOP
         ENDIF
        ENDIF
*
*   Now, if the date has changed OR if we're on the last observation, 
*   update yesterday's contributions to the matrix (missing values are 
*   represented by -1.0D12). Always update today's residual vector. This 
*   scheme relies on the fact that the data are in chronological order.
*
        NEWDAT = (10000*CASEID(2))+(100*CASEID(3))+CASEID(4)
        IF ((NEWDAT.NE.OLDDAT).OR.(I.EQ.NOBS+1)) THEN
         OLDDAT=NEWDAT
         DO 200 J=1,NSITES
          IF (ONEDAY(J).GT.-0.9D12) THEN
           DO 205 K=J,NSITES
            IF (ONEDAY(K).GT.-0.9D12) THEN
             CORMAT(J,K) = CORMAT(J,K) + (ONEDAY(J)*ONEDAY(K))
             CORRN(J,K) = CORRN(J,K) + 1.0D0
             MEAN(J,K) = MEAN(J,K) + ONEDAY(J)
             VAR(J,K) = VAR(J,K) + (ONEDAY(J)**2)
             IF (K.NE.J) THEN 
              VAR(K,J) = VAR(K,J) + (ONEDAY(K)**2)
              MEAN(K,J) = MEAN(K,J) + ONEDAY(K)
             ENDIF
            ENDIF
 205       CONTINUE
           ONEDAY(J) = -1.0D12
          ENDIF
 200     CONTINUE
        ENDIF
*
*	And copy the latest entry into today's array
*
        ONEDAY(CASEID(1)) = TMP
 100   CONTINUE
*
*   That's everything collected now, so calculate the 
*   covariance structure and, if necessary, convert to 
*   correlations. If variances are zero, correlations
*   are treated as zero. NB if RESTYP=3 we're calculating
*   joint occurrence probabilities rather than covariances,
*   so in this case there's no need to subtract the product
*   of means.
*       
       DO 220 I=1,NSITES
        DO 221 J=I,NSITES
         IF (CORRN(I,J).GT.0.0D0) THEN
          IF (DEPTYP.NE.3) THEN 
           MEAN(I,J) = MEAN(I,J)/CORRN(I,J)
           IF (J.NE.I) MEAN(J,I) = MEAN(J,I)/CORRN(I,J)
           CORMAT(I,J) = CORMAT(I,J)/CORRN(I,J) - (MEAN(I,J)*MEAN(J,I))
          ELSE
           CORMAT(I,J) = CORMAT(I,J)/CORRN(I,J)
          ENDIF
          IF (DEPTYP.EQ.2) THEN
           VAR(I,J) = VAR(I,J)/CORRN(I,J) - (MEAN(I,J)**2)
           IF (J.NE.I) VAR(J,I) = VAR(J,I)/CORRN(I,J) - (MEAN(J,I)**2)
           IF ((VAR(I,J).GT.TOL).AND.(VAR(J,I).GT.TOL)) THEN
            CORMAT(I,J) = CORMAT(I,J)/DSQRT(VAR(I,J)*VAR(J,I))
           ELSE
            CORMAT(I,J) = 0.0D0
           ENDIF
          ENDIF
         ELSE
          CORMAT(I,J) = -9.999D0
         ENDIF
         CORMAT(J,I) = CORMAT(I,J)
         CORRN(J,I) = CORRN(I,J)
         WRITE(CORFNO) CORRN(I,J),CORMAT(I,J)
 221    CONTINUE
 220   CONTINUE

******************************************************************************
*     Second scenario: reread correlations as computed on a previous pass,
*     from scratch file
******************************************************************************
      ELSE
       REWIND(CORFNO)
       DO 320 I=1,NSITES
        DO 321 J=I,NSITES
         READ(CORFNO) CORRN(I,J),CORMAT(I,J)
         CORMAT(J,I) = CORMAT(I,J)
         CORRN(J,I) = CORRN(I,J)
 321    CONTINUE
 320   CONTINUE
      ENDIF

 1    FORMAT(/5X,'****ERROR**** In routine EMPCOR, an invalid ',
     +'value of ',A6,/5X,'was found. This is a programming ',
     +'error - please inform REC.')
      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE MLNUES(NUHAT,SUMLNY,NOBS,LOGL)
*
*     To calculate ML estimate of dispersion parameter in a Gamma
*     GLM. Arguments are:
*     NUHAT    - Estimated dispersion parameter. INPUT/OUTPUT
*     SUMLNY   - The sum of Ln(Y) values for the dataset. INPUT
*     NOBS     - number of observations. INPUT
*     LOGL     - On entry, the part of the log-likelihood corresponding
*                to the means. On exit, the full log-likelihood.
******************************************************************************
      DOUBLE PRECISION NUHAT,SUMLNY,LOGL
      INTEGER NOBS
******************************************************************************
*     Extra DOUBLEs
*     ^^^^^^^^^^^^^
*     A,B      - limits of current interval (parametrise as log(NU) to 
*                force everything to stay positive)
*     X        - new value
*     F1,F2,FX - objective function values (trying to get them = 0).
*     DIGAMM   - Digamma function
*     GAMMLN   - Log of gamma function
*     RHS      - Right-hand side of equation (bits not involving NUIPAR)
*     TOL      - required precision
******************************************************************************
      DOUBLE PRECISION A,B,X,F1,F2,FX,DIGAMM,GAMMLN,RHS,TOL
******************************************************************************
*     Extra INTEGERs
*     ^^^^^^^^^^^^^^
*     I        - counter
*     IFAIL    - exit code for GAMMLN etc.
*     MAXIT    - maximum no. of iterations
******************************************************************************
      INTEGER I,MAXIT,IFAIL

      MAXIT = 20
      TOL = 1.0D-8
*
*     Do root bracketing. Take as initial guesses half the input value of 
*     NUHAT, and double it; check that root is bracketed before 
*     proceeding
*
      A = DLOG(NUHAT/2.0D0)
      B = DLOG(NUHAT*2.0D0)
      RHS = 1.0D0 + (LOGL/DBLE(NOBS)) + (SUMLNY/DBLE(NOBS))
      
 5    F1 = DIGAMM(DEXP(A),IFAIL) - A - RHS
      IF (F1.GT.0.0D0) THEN
       A = A - 1.0D0
       GOTO 5
      ENDIF
 6    F2 = DIGAMM(DEXP(B),IFAIL) - B - RHS
      IF (F2.LT.0.0D0) THEN
       B = B + 1.0D0
       GOTO 6
      ENDIF
      DO 10 I = 1,MAXIT
       IF (DABS(B-A).LT.TOL) GOTO 20
       X = ((A*F2)-(B*F1))/(F2-F1)
       FX = DIGAMM(DEXP(X),IFAIL) - X - RHS
       B = A
       F2 = F1
       A = X
       F1 = FX
 10   CONTINUE

 20   NUHAT = DEXP((A+B)/2.0D0)
      LOGL = NUHAT*(LOGL + ((DBLE(NOBS)*DLOG(NUHAT)) + SUMLNY) )
      LOGL = LOGL - (DBLE(NOBS)*GAMMLN(NUHAT,IFAIL))
      RETURN

      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE INVPAR(X,FX,TTYPE)
*
*     Performas inverse parabolic interpolation using 3 points
*     (implementation of Press et al 1992, equation 10.2.1).
*     Given 3 values of X, and corresponding function values FX,
*     fits a quadratic to estimate a turning point of FX. On exit,
*     X(3) contains the estimated turning point; FX(1) and FX(2)
*     contain the largest (if TTYPE=1) or smallest (if TTYPE=-1)
*     two of the input FX values, and X(1) and X(2) contain the 
*     corresponding Xs. 
******************************************************************************
      DOUBLE PRECISION X(3),FX(3)
      INTEGER TTYPE
******************************************************************************
*     Extra DOUBLES
*     ~~~~~~~~~~~~~
*     A,B,C,   }    All as in Press et al 1992, equation 10.2.1
*     FA,FB,FC }
*     NUMER         Numerator in formula
*     DENOM         Denominator
*     FOPT          Largest (smallest) input value of FX 
******************************************************************************
      DOUBLE PRECISION A,B,C,FA,FB,FC,NUMER,DENOM,FOPT
******************************************************************************
*     Extra INTEGERS
*     ~~~~~~~~~~~~~~
*     SMLIDX        Index of smallest (largest) element in FX
*     I             Counter
******************************************************************************
      INTEGER SMLIDX,I
      A = X(1)
      B = X(2)
      C = X(3)
       
      FA = FX(1)
      FB = FX(2)
      FC = FX(3)

*
*     Get rid of smallest element from FX and move elements of X
*     accordingly
*
      SMLIDX = 1
      FOPT = FX(1)
      DO 100 I=2,3
       IF ( (DBLE(TTYPE)*FX(I)).LT.FOPT) THEN
        SMLIDX = I
        FOPT = FX(I)
       ENDIF
 100  CONTINUE

      IF (SMLIDX.NE.3) THEN
       X(SMLIDX) = X(3)
       FX(SMLIDX) = FX(3)
      ENDIF
      FX(3) = 0.0D0

*
*     OK, now compute new value of X
*
      NUMER = ( ((B-A)**2)*(FB-FC) ) - ( ((B-C)**2)*(FB-FA) )
      DENOM = ( (B-A)*(FB-FC) ) - ( (B-C)*(FB-FA) )
      X(3) = B - (0.5D0*NUMER/DENOM)
      

      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE FITCAL(BETA,NP,P,MAXP,MODEL,SCRFNO,RECNO,ITER,
     +               TRWRN,TRACE,DISPAR,LLINVL,X,Y,SITE,YY,MM,DD,
     +                                          LOGL,MU,ETA,IFAIL)
*
*     To calculate (a) fitted value (b) linear predictor for a 
*     single observation, and increment the log-likelihood  
*     accordingly.
*
*     Arguments:
*
*     BETA      (DOUBLE, input) Coefficient vector for linear predictor
*     NP        (INTEGER, input) Numbers of parameters in each component
*               of BETA
*     P         (INTEGER,input) Number of parameters being estimated
*     MAXP      (INTEGER, input) Storage allocated to X and P
*     MODEL     (INTEGER, input): model code. See header to IWLS
*               routine for values
*     SCRFNO    (INTEGER, input): number of scratch file holding 
*               covariate info etc.
*     RECNO     (INTEGER, input): the record number to read from SCRFNO
*     ITER      (INTEGER, input): iteration number
*     TRWRN     (INTEGER, input): indicates whether we've issued a 
*               warning message about trace values on the current
*               iteration
*     TRACE     (DOUBLE, input): threshold for `trace' values
*     DISPAR    (DOUBLE, input): dispersion parameter for model
*     LLINVL    (DOUBLE, input): 'Invalid' log-likelihood
*     LOGL      (DOUBLE, input/output): on input, log-likelihood before
*               looking at current case; on output, log-likelihood
*               including current case
*     X         (DOUBLE, output) Vector of covariates, read from file
*     Y         (DOUBLE, output) Response, read from file
*     SITE      (INTEGER, output) Site code, read from file
*     YY        (INTEGER, output) Year, read from file
*     MM        (INTEGER, output) Month, read from file
*     DD        (INTEGER, output) Day, read from file
*     MU        (DOUBLE, output): fitted value for current case
*     ETA       (DOUBLE, output): linear predictor for current case
*
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
*////////////	NB THIS ROUTINE CONTAINS SECTIONS WHICH ARE MODEL-   ///////
*////////////	SPECIFIC. THESE SECTIONS ARE ENCLOSED BY LINES OF    ///////
*////////////   HASHES LIKE THIS, FOR EASY RECOGNITION IF YOU WANT   ///////
*////////////   TO AMEND THEM SUBSEQUENTLY, OR TO ADD EXTRA MODELS.  ///////
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
******************************************************************************
      INTEGER NP(10),P,MAXP,MODEL,SCRFNO,RECNO,TRWRN,IFAIL,ITER
      INTEGER SITE,YY,MM,DD
      DOUBLE PRECISION BETA(0:MAXP),TRACE,DISPAR,LLINVL
      DOUBLE PRECISION X(0:MAXP),Y,LOGL,MU,ETA
******************************************************************************
*     Additional INTEGERs
*     ^^^^^^^^^^^^^^^^^^^
*     J               Counter
******************************************************************************
      INTEGER J
******************************************************************************
*     Additional DOUBLEs
*     ^^^^^^^^^^^^^^^^^^
*     TMP             Temporary storage
******************************************************************************
      DOUBLE PRECISION TMP

      IFAIL = 0
*
*       Compute linear predictor - NB using parameters up to NP(6) - the
*       rest of BETA is just used in updating nonlinear parameters
*
       ETA = BETA(0)
       READ(SCRFNO,REC=RECNO) SITE,YY,MM,DD,Y,(X(J),J=1,P)
       DO 100 J=1,NP(6)
	ETA = ETA + (BETA(J)*X(J))
 100   CONTINUE
*
*	And fitted values ...
*
       TMP = DEXP(ETA)
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
       IF (MODEL.EQ.1) THEN
        MU = TMP/(1.0D0+TMP)
       ELSEIF (MODEL.EQ.11) THEN
        MU = TMP
       ELSE
        WRITE(*,1)
        STOP
       ENDIF
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
*
*	Replace trace values by their approximate conditional
*	expectations as appropriate. We need to modify the value of 
*       Y in the data file as well if this happens.
*
       IF (Y.LT.TRACE) THEN
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
*
*	Gamma model. This is rather a crude approximation, but it 
*	avoids having to do numerical integration, which is important 
*	from the point of view of speed! We also need a NaN trap
*	in case we iterate to somewhere peculiar and end up dividing
*	by zero ... this is accomlished via IF(.NOT.(Y.LE.TRACE)).
*       Use ML estimates of dispersion parameter, so that we're effectively
*       doing an EM algorithm kind of thing.
*
        IF (MODEL.EQ.11) THEN
	 Y = (MU*DISPAR*TRACE)/( (2.0D0*MU) + (DISPAR*TRACE) )
	 IF (.NOT.(Y.LE.TRACE) ) THEN
          IF (TRWRN.LT.10) THEN
	   WRITE(*,'(/A80)') 
     +     '****WARNING**** Some trace values have'//
     +     ' approximate conditional expectation > trace threshold '
           TRWRN = TRWRN + 10
          ENDIF
	  Y = TRACE*0.99
	 ELSEIF (Y.LE.0.0D0) THEN
          IF (MOD(TRWRN,10).EQ.0) THEN
	   WRITE(*,'(/A80)') 
     +	   '****WARNING**** Some trace values have'//
     +     ' approximate conditional expectation < 0 '
           TRWRN = TRWRN + 1
          ENDIF
	  Y = TRACE/1.0D2
	 ENDIF
         WRITE(SCRFNO,REC=RECNO) SITE,YY,MM,DD,Y,(X(J),J=1,P)
        ENDIF
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
       ENDIF
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
*
*	For binary outcomes, check that fitted values aren't 0 or 1 
*	to machine precision. Only stop the prog if this happens with
*	starting values - otherwise we'll backtrack on the iteration
*	later on.
*
       IF (MODEL.EQ.1) THEN
        IF ( (MU.EQ.0.0D0).OR.(MU.EQ.1.0D0) ) THEN
         IF (ITER.EQ.0) THEN
          WRITE(*,2)
          STOP
         ELSE
          LOGL = LLINVL
          IFAIL = 1
         ENDIF
        ENDIF
       ENDIF
*
*	Now update log-likelihood
*
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
       IF (MODEL.EQ.1) THEN
*
*	Logistic ... (McCullagh & Nelder, eqn 4.13)
*
        IF (Y.GT.0.0D0) LOGL = LOGL + ETA
        LOGL = LOGL - DLOG(1.0D0+DEXP(ETA))
       ELSEIF (MODEL.EQ.11) THEN
*
*	Gamma. Omit terms that don't depend on MU, because estimation 
*       of NU at each stage screws things up. So this isn't strictly 
*       the log-likelihood, but it's the bit we think is important 
*       (see McCullagh & Nelder, p.290, for example). NB at the end 
*       we need to remember to multiply through by dispersion parameter,
*       but not here! (save on computation).
*
        LOGL = LOGL - (Y/MU) - DLOG(MU)
       ENDIF
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////

 1    FORMAT(/'*****ERROR***** Invalid model code found in routine ',
     +'FITCAL. This',/,'is a programming error - please contact REC.')
 2    FORMAT(/5X,'****WARNING**** Current parameter estimates yield',
     +' a fitted',/5X,
     +'probability of either 0 or 1 (to machine precision). This ',
     +'isn''t going ',/5X,
     +'to converge - I''m bailing out. Try different starting values.')
      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE RESID(MODEL,NU,NP,NOBS,FILNO,OUTFNO,SCODES,
     +                           SITETXT,NSITES,MXNS)
*
*       To perform a basic residual analysis - Pearson residuals
*       broken down by month, site and year, along with various other
*       model summaries.
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
*////////////   NB THIS ROUTINE CONTAINS QUITE A FEW SECTIONS WHICH  ///////
*////////////   ARE MODEL-SPECIFIC. THESE SECTIONS ARE ENCLOSED BY   ///////
*////////////   LINES OF HASHES LIKE THIS, FOR EASY RECOGNITION IF   ///////
*////////////   YOU SUBSEQUENTLY WANT TO AMEND THEM OR ADD EXTRA     ///////
*////////////   MODELS.                                              ///////
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>    PARAMETERS      >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*       FIRSTYR - First year for which we'll ever have data
*       NYEARS  - No. of years
*       MXNSIT  - Max. no of sites (should be equal to value of MXNSITES
*                 in main program)
*       MXN     - Max. no of observations (for dimensioning arrays to go
*                 into probability plots)
******************************************************************************
	INTEGER MXNSIT,FIRSTYR,NYEARS,MXN
	PARAMETER (MXNSIT=200,FIRSTYR=1870,NYEARS=200,MXN=250000)
******************************************************************************
*               INTEGER variables
*               ^^^^^^^^^^^^^^^^^
*       I,J,K	- Counters
*       MODEL   - Code to indicate which model we're dealing with.
*                 Current values are
*                 1 : Logistic regression for binary responses
*                 11: Gamma model with log link
*	FILNO	- File number of scratch file with residual info in
*	OUTFNO	- File number for output file
*       MXNS    - Max. no of sites (= MXNSIT, but FORTRAN's a bit
*                 crap like that)
*       NP      - No. of parameters estimated in model (=no. of lost
*                 degrees of freedom)
*       NOBS    - no. of observations
*       NSITES  - no. of sites required
*       CASEID  - array of identifiers for each case, for purposes of
*                 residual analysis. Col.1 contains site number, 2 
*                 contains year, 3 contains month and 4 day
*       SFLAGS  - Flags indicating which sites have data present for the
*                 current day.
*       CURDAT  - Current date
*       OLDDAT  - Date for previous record
*       NADJM   - Numbers of observations from each pair of sites in 
*                 each month (used for calculating adjustments for
*                 spatial dependence)
*       NADJY   - Ditto, for each year 
*       YEAR    - Corresponding to values in SFLAGS
*       MONTH   - Ditto
*	CTABLE	- Classification table in residual analysis (logistic model)
*	BOXNO	- Used to identify which part of the forecast distribution
*		  each observation falls in (logistic model)
***************************************************************************
      INTEGER MODEL,NSITES,NP,NOBS,MXNS,FILNO,OUTFNO
      INTEGER I,J,K
      INTEGER CASEID(4),SFLAGS(MXNSIT),CURDAT,OLDDAT      
      INTEGER NADJM(12,MXNSIT,MXNSIT),YEAR,MONTH
      INTEGER NADJY(FIRSTYR:FIRSTYR+NYEARS-1,MXNSIT,MXNSIT)
      INTEGER CTABLE(0:1,0:1),BOXNO
***************************************************************************
*               DOUBLE PRECISION variables
*               ^^^^^^^^^^^^^^^^^^^^^^^^^^
*       Y       Daily value (0 or 1 for logistic)
*       YBAR    Mean of Ys
*       SY      Std Dev of Ys
*       MU      Predicted MEANS (probs for logistic)
*       NU      Fitted dispersion parameter
*       ERRBAR  Mean prediction error
*       RAWRES  Raw residual (observed - expected)
*       MSE     Mean squared error
*       RSQ     R-squared (1 - Error sum of squares/Data sum of squares)
*	PCOBS	Percentage of correct predictions, as observed          } L 
*	PCEXP	Expected percentage of correct predictions under model  } o  M
*	PCOALL	Overall percentage of correct predictions               } g  o
*	PCEALL	Expected overall percentage correct                     } i  d
*	BYPRED	Table of rainfall occurrence frequencies by predicted   } s  e
*		probability. Row 1 is observed proportion in each       } t  l
*		category, Row 2 is expected, and Row 3 is N.            } ic
*       CURPR   Current Pearson residual
*       PRBAR   Mean Pearson residual
*       SPR     Std Dev of pearson residuals
*	SEPR	Standard Error of mean Pearson residual
*	SESCAL	Scaling factor to correct standard errors for spatial
*		dependence
*       PHI     Theoretical std dev of Pearson residuals
*       PRMNTH  Pearson residuals by month } Column 0 is N, 1 is mean,
*       PRSITE  And by site                } 2 is Std Dev, 3 is S.e. (mean)
*       PRYEAR  And by year                } if model is correct
*       ARBARO  Observed mean Anscombe residual (gamma model)
*       SARO    Observed Std dev of Anscombe residuals (gamma model)
*       ARBARE  Expected mean Anscombe residual (gamma model)
*       SARE    Expected Std dev of Anscombe residuals (gamma model)
*	CORMAT	- Spatial correlation matrix for Pearson residuals
*	CORRN	- Ns going to make up entries in CORMAT
*       SEADJM  Adjustments to standard errors for monthly means
*       SEADJY  Ditto, annual means 
*       TMP     Temporary storage
******************************************************************************
      DOUBLE PRECISION Y,MU,NU
      DOUBLE PRECISION YBAR,SY,ERRBAR,MSE,RSQ
      DOUBLE PRECISION RAWRES,CURPR,PRBAR,SPR,SEPR,SESCAL,PHI
      DOUBLE PRECISION PCOBS(0:1),PCEXP(0:1),PCOALL,PCEALL
      DOUBLE PRECISION BYPRED(3,0:9)
      DOUBLE PRECISION PRMNTH(12,0:3),PRSITE(MXNSIT,0:3)
      DOUBLE PRECISION PRYEAR(FIRSTYR:FIRSTYR+NYEARS-1,0:3)
      DOUBLE PRECISION ARBARO,SARO,ARBARE,SARE
      DOUBLE PRECISION CORMAT(MXNSIT,MXNSIT),CORRN(MXNSIT,MXNSIT)
      DOUBLE PRECISION SEADJM(12),SEADJY(FIRSTYR:FIRSTYR+NYEARS-1)
      DOUBLE PRECISION TMP
***************************************************************************
*               CHARACTER variables
*               ^^^^^^^^^^^^^^^^^^^
*       SITETXT - array of site descriptors etc.
*       SCODES  - array of short site codes
***************************************************************************
      CHARACTER*70 SITETXT(MXNSIT)
      CHARACTER SCODES(MXNSIT)*4
***************************************************************************
      DATA YBAR,SY,ERRBAR,MSE,PRBAR,SPR,ARBARO,SARO /8*0.0D0/
      DATA (SFLAGS(I),I=1,MXNSIT) /MXNSIT*0/
***************************************************************************
*	Check dimensioning:
*
      IF (MXNSIT.NE.MXNS) CALL PRGERR(' RESID  ','fit_both.f',1)
*
*       Initialise arrays (easier this way than in a DATA statement)
*
      DO 100 J=0,3
       DO 101 I=1,NSITES
	PRSITE(I,J) = 0.0D0
 101   CONTINUE
       DO 102 I=1,12
	PRMNTH(I,J) = 0.0D0
        SEADJM(I) = 0.0D0
 102   CONTINUE
       DO 103 I=FIRSTYR,FIRSTYR+NYEARS-1
	PRYEAR(I,J) = 0.0D0
        SEADJY(I) = 0.0D0
 103   CONTINUE
 100  CONTINUE

      DO 105 I=1,NSITES
       DO 106 J=I,NSITES
        DO 107 K=1,12
         NADJM(K,I,J) = 0
 107    CONTINUE
        DO 108 K=FIRSTYR,FIRSTYR+NYEARS-1
         NADJY(K,I,J) = 0
 108    CONTINUE
 106   CONTINUE
 105  CONTINUE
      OLDDAT = 0
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
*
*     Logistic model: initialise classification tables & theoretical
*     std dev of Pearson residuals
*
      IF (MODEL.EQ.1) THEN
       DO 110 I=0,1
        PCEXP(I) = 0.0D0
        DO 111 J=0,1
         CTABLE(I,J) = 0.0D0
 111    CONTINUE
 110   CONTINUE
       DO 112 I=1,3
        DO 113 J=0,9
         BYPRED(I,J) = 0.0D0
 113    CONTINUE
 112   CONTINUE
       PHI = 1.0D0
*
*     Gamma model: calculate expected Anscombe residual properties
*
      ELSEIF (MODEL.EQ.11) THEN
       CALL ANSPRM(NU,ARBARE,SARE)
       SARE = DSQRT(SARE)
       PHI = 1.0D0/DSQRT(NU)
      ENDIF
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////

*
*       Go through and calculate all required sums, and sums of squares.
*
      DO 200 I=1,NOBS
       READ(FILNO,REC=I) (CASEID(J),J=1,4),Y,MU
       RAWRES = Y - MU
       MSE = MSE + (RAWRES**2)
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
*
*     Logistic model: produce classification tables, and standard errors. 
*     Rows of CTABLE represent stratification by predicted outcome.
*
       IF (MODEL.EQ.1) THEN
        CURPR = RAWRES/DSQRT(MU*(1.0D0-MU))
        CTABLE(NINT(MU),NINT(Y)) = CTABLE(NINT(MU),NINT(Y)) + 1
        PCEXP(NINT(MU)) = PCEXP(NINT(MU)) + DMAX1(MU,1.0D0-MU)
*
*	Now more complete stratification - split forecasts into deciles
*
        BOXNO = INT(1.0D1*MU)
        BYPRED(1,BOXNO) = BYPRED(1,BOXNO) + Y
        BYPRED(2,BOXNO) = BYPRED(2,BOXNO) + MU
        BYPRED(3,BOXNO) = BYPRED(3,BOXNO) + 1.0D0
*
*     Gamma model: summary of observations, and Anscombe residuals
*
       ELSEIF (MODEL.EQ.11) THEN
        CURPR = RAWRES/MU
        YBAR = YBAR + Y
        SY = SY + (Y**2)
        ERRBAR = ERRBAR + RAWRES
        TMP = (Y/MU)**(1.0D0/3.0D0)
        ARBARO = ARBARO + TMP
        SARO = SARO + ( TMP**2 )
       ENDIF
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
*
*	Now Pearson residuals - firstly, overall scores and secondly,
*	stratified by year, month and site. Also calculate contributions
*       to spatial dependence adjustments
*
       PRBAR = PRBAR + CURPR
       SPR = SPR + (CURPR**2)
*
*	0s are N
*
       PRSITE(CASEID(1),0) = PRSITE(CASEID(1),0) + 1.0D0
       PRYEAR(CASEID(2),0) = PRYEAR(CASEID(2),0) + 1.0D0
       PRMNTH(CASEID(3),0) = PRMNTH(CASEID(3),0) + 1.0D0
*
*       1s are means ...
*
       PRSITE(CASEID(1),1) = PRSITE(CASEID(1),1) + CURPR
       PRYEAR(CASEID(2),1) = PRYEAR(CASEID(2),1) + CURPR
       PRMNTH(CASEID(3),1) = PRMNTH(CASEID(3),1) + CURPR
*
*       2s are observed sds (sum of squares for now)
*
       PRSITE(CASEID(1),2) = PRSITE(CASEID(1),2) + (CURPR**2)
       PRYEAR(CASEID(2),2) = PRYEAR(CASEID(2),2) + (CURPR**2)
       PRMNTH(CASEID(3),2) = PRMNTH(CASEID(3),2) + (CURPR**2)
*
*      Here are the contributions to adjustments for spatial dependence -
*      need to know, for each subset of data, how many times each
*      pair of sites occurs
*
       CURDAT = (10000*CASEID(2)) + (100*CASEID(3)) + CASEID(4)
       SFLAGS(CASEID(1)) = 1
       IF ( ((CURDAT.NE.OLDDAT).AND.(OLDDAT.NE.0)) .OR. 
     +      (I.EQ.NOBS) ) THEN
        YEAR = OLDDAT/10000
        MONTH = MOD(OLDDAT,10000) / 100
        DO 120 J=1,NSITES
         DO 121 K=J,NSITES
          IF ( (SFLAGS(J).EQ.1).AND.(SFLAGS(K).EQ.1) ) THEN
           NADJM(MONTH,J,K) = NADJM(MONTH,J,K) + 1
           NADJY(YEAR,J,K) = NADJY(YEAR,J,K) + 1
          ENDIF
 121     CONTINUE
         SFLAGS(J) = 0
 120    CONTINUE
       ENDIF
       OLDDAT = CURDAT
 200  CONTINUE

*
*	That's everything collected now. Next, compute summary statistics.
*	Start by computing inflation factors for standard errors, to 
*	compensate for spatial dependence. This involves calculating the
*	inter-site correlation matrix. NB FOR CONVENIENCE, DIVIDE
*       DIAGONAL ENTRIES BY 2, SINCE WE DON'T USE THIS MATRIX
*       SUBSEQUENTLY
*
       CALL EMPCOR(FILNO,NSITES,NOBS,MODEL,1,2,CORMAT,CORRN,MXNSIT)
       SESCAL = 0.0D0
       DO 220 I=1,NSITES
        CORMAT(I,I) = 0.5D0 * CORMAT(I,I)
        DO 221 J=I,NSITES
         IF (CORRN(I,J).GT.0.0D0) THEN
          SESCAL = SESCAL + (2.0D0*CORRN(I,J)*CORMAT(I,J))
         ENDIF
         DO 225 K=1,12
          IF (NADJM(K,I,J).GT.0) THEN
           SEADJM(K) = SEADJM(K) + 
     +                     (2.0D0*CORMAT(I,J)*DBLE(NADJM(K,I,J)))
          ENDIF
 225     CONTINUE
         DO 226 K=FIRSTYR,FIRSTYR+NYEARS-1
          IF (NADJY(K,I,J).GT.0) THEN
           SEADJY(K) = SEADJY(K) + 
     +                     (2.0D0*CORMAT(I,J)*DBLE(NADJY(K,I,J)))
          ENDIF
 226     CONTINUE
 221   CONTINUE
 220  CONTINUE
      SESCAL = DSQRT(SESCAL/DBLE(NOBS))
      DO 230 I=1,12
       IF (PRMNTH(I,0).GT.0.0D0) 
     +                   SEADJM(I) = DSQRT(SEADJM(I)/PRMNTH(I,0))
 230  CONTINUE
      DO 231 I=FIRSTYR,FIRSTYR+NYEARS-1
       IF (PRYEAR(I,0).GT.0.0D0) 
     +                   SEADJY(I) = DSQRT(SEADJY(I)/PRYEAR(I,0))
 231  CONTINUE
*
*       Now overall performance (measures of which are model specific). 
*       Note that all Std Err variables currently contain sums of squares. 
*       For model std errs we divide by NOBS - NP
*
      PRBAR = PRBAR/DBLE(NOBS)
      SPR = (SPR - (DBLE(NOBS)*(PRBAR**2)))/DBLE(NOBS-NP)
      SPR = DSQRT(SPR)
      SEPR = SESCAL*SPR/DSQRT(DBLE(NOBS))
      MSE = MSE/DBLE(NOBS-NP)
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
*
*     Logistic model
*
      IF (MODEL.EQ.1) THEN
       PCOALL = 1.0D2*DBLE(CTABLE(0,0)+CTABLE(1,1)) /
     + 	       DBLE( CTABLE(0,0) + CTABLE(0,1)
     +              + CTABLE(1,0) + CTABLE(1,1) )
       PCEALL = 1.D2*(PCEXP(0)+PCEXP(1)) /
     + 	       DBLE( CTABLE(0,0) + CTABLE(0,1)
     +              + CTABLE(1,0) + CTABLE(1,1) )
       DO 205 I=0,1
        IF (CTABLE(I,0)+CTABLE(I,1).NE.0) THEN
         PCOBS(I) = 1.0D2*DBLE(CTABLE(I,I)) / 
     + 				 DBLE(CTABLE(I,0)+CTABLE(I,1))
         PCEXP(I) = 1.0D2*PCEXP(I) / DBLE(CTABLE(I,0)+CTABLE(I,1))
        ELSE
         PCOBS(I) = 0.0D0
         PCEXP(I) = 0.0D0
        ENDIF
 205   CONTINUE
       
       DO 206 I=0,9
        IF (BYPRED(3,I).GT.0.0D0) THEN
         BYPRED(1,I) = BYPRED(1,I)/BYPRED(3,I)
         BYPRED(2,I) = BYPRED(2,I)/BYPRED(3,I)
        ENDIF
 206   CONTINUE
       WRITE(*,11) CTABLE(0,0),CTABLE(0,1),PCOBS(0),PCEXP(0),
     + 	          CTABLE(1,0),CTABLE(1,1),PCOBS(1),PCEXP(1),
     + 		  PCOALL,PCEALL,((BYPRED(I,J),J=0,9),I=1,2),
     + 		  (INT(BYPRED(3,J)),J=0,9),MSE,PRBAR,SPR,SEPR
       WRITE(OUTFNO,11) CTABLE(0,0),CTABLE(0,1),PCOBS(0),PCEXP(0),
     + 	          CTABLE(1,0),CTABLE(1,1),PCOBS(1),PCEXP(1),
     + 		  PCOALL,PCEALL,((BYPRED(I,J),J=0,9),I=1,2),
     + 		  (INT(BYPRED(3,J)),J=0,9),MSE,PRBAR,SPR,SEPR
*
*     Gamma model
*
      ELSEIF (MODEL.EQ.11) THEN
       YBAR = YBAR/DBLE(NOBS)
       SY = (SY - (DBLE(NOBS)*(YBAR**2)))/DBLE(NOBS-1)
       RSQ = 1.0D0 - (MSE/SY)
       SY = DSQRT(SY)
       ERRBAR = ERRBAR/DBLE(NOBS)
       ARBARO = ARBARO/DBLE(NOBS)
       SARO = (SARO - (DBLE(NOBS)*(ARBARO**2)))/DBLE(NOBS-NP)
       SARO = DSQRT(SARO)
       
       WRITE(*,12) YBAR,SY,ERRBAR,DSQRT(MSE),RSQ,PRBAR,SPR,PHI,
     +                                   ARBARO,ARBARE,SARO,SARE
       WRITE(OUTFNO,12) YBAR,SY,ERRBAR,DSQRT(MSE),RSQ,PRBAR,SPR,
     +                               PHI,ARBARO,ARBARE,SARO,SARE
      ENDIF
*///////////////////////////////////////////////////////////////////////////
*///////////////////////////////////////////////////////////////////////////
*
*       Broken down by month. Use divisor N-1 for observed standard 
*       deviations (they're deviations from observed means) 
*
      WRITE(*,13)
      WRITE(OUTFNO,13)
      DO 210 I=1,12
       IF (PRMNTH(I,0).GT.0.0D0) THEN
	PRMNTH(I,1) = PRMNTH(I,1)/PRMNTH(I,0)
        IF (PRMNTH(I,0).GT.1.0D0) THEN
	 PRMNTH(I,2) = (PRMNTH(I,2)-(PRMNTH(I,0)*(PRMNTH(I,1)**2)))
     +                                          /(PRMNTH(I,0)-1.0D0)
	 PRMNTH(I,2) = DSQRT(PRMNTH(I,2))
        ELSE
*
*     If there's only 1 observations, no need to correct for spatial
*     dependence so omit SESCAL!
*
         PRMNTH(I,2) = PHI
        ENDIF
	PRMNTH(I,3) = SEADJM(I)*PHI/(DSQRT(PRMNTH(I,0)))
	WRITE(*,14) I,INT(PRMNTH(I,0)),(PRMNTH(I,J),J=1,3)
	WRITE(OUTFNO,14) I,INT(PRMNTH(I,0)),(PRMNTH(I,J),J=1,3)
       ENDIF
 210  CONTINUE
      WRITE(*,'(T6,49(''-''))') 
      WRITE(OUTFNO,'(T6,49(''-''))') 
*
*       Broken down by site
*
      WRITE(*,15)
      WRITE(OUTFNO,15)
      DO 211 I=1,NSITES
       IF (PRSITE(I,0).GT.0.0D0) THEN
	PRSITE(I,1) = PRSITE(I,1)/PRSITE(I,0)
        IF (PRSITE(I,0).GT.1.0D0) THEN
	 PRSITE(I,2) = (PRSITE(I,2)-(PRSITE(I,0)*(PRSITE(I,1)**2)))
     +                                          /(PRSITE(I,0)-1.0D0)
	 PRSITE(I,2) = DSQRT(PRSITE(I,2))
        ELSE
         PRSITE(I,2) = PHI
        ENDIF
	PRSITE(I,3) = PHI/(DSQRT(PRSITE(I,0)))
	WRITE(*,16) SCODES(I),SITETXT(I),
     +           INT(PRSITE(I,0)),(PRSITE(I,J),J=1,3)
	WRITE(OUTFNO,16) SCODES(I),SITETXT(I),
     +           INT(PRSITE(I,0)),(PRSITE(I,J),J=1,3)
       ENDIF
 211  CONTINUE
      WRITE(*,'(T6,77(''-''))')
      WRITE(OUTFNO,'(T6,73(''-''))') 
*
*       Broken down by year
*
      WRITE(*,17)
      WRITE(OUTFNO,17)
      DO 212 I=FIRSTYR,FIRSTYR+NYEARS-1
       IF ( PRYEAR(I,0).GT.0.0D0 ) THEN
	PRYEAR(I,1) = PRYEAR(I,1)/PRYEAR(I,0)
        IF (PRYEAR(I,0).GT.1.0D0) THEN
	 PRYEAR(I,2) = (PRYEAR(I,2)-(PRYEAR(I,0)*(PRYEAR(I,1)**2)))
     +                                          /(PRYEAR(I,0)-1.0D0)
	 PRYEAR(I,2) = DSQRT(PRYEAR(I,2))
        ELSE
         PRYEAR(I,2) = PHI
        ENDIF
	PRYEAR(I,3) = SEADJY(I)*PHI/(DSQRT(PRYEAR(I,0)))
	WRITE(*,14) I,INT(PRYEAR(I,0)),(PRYEAR(I,J),J=1,3)
	WRITE(OUTFNO,14) I,INT(PRYEAR(I,0)),(PRYEAR(I,J),J=1,3)
       ENDIF
 212  CONTINUE
      WRITE(*,'(T6,49(''-''))') 
      WRITE(OUTFNO,'(T6,49(''-''))') 

      WRITE(*,20)
      WRITE(OUTFNO,20)

      WRITE(*,*)

 11   FORMAT(//,30X,'RESIDUAL ANALYSIS',/30X,17('='),//,
     +T19,'|    Observed    ||     % correct',/,
     +T19,'|   Dry     Wet  || Observed  Expected',/,
     +T6,52('-'),/,
     +T6,'Forecast dry |',1X,I6,2X,I6,1X,'||',3X,F4.1,6X,F4.1,/,
     +T6,'Forecast wet |',1X,I6,2X,I6,1X,'||',3X,F4.1,6X,F4.1,/,
     +T6,52('-'),/,
     +T19,'OVERALL % CORRECT :',3X,F4.1,6X,F4.1,//,
     +T2,'Rainday frequencies vs forecasts:',/,T2,32('-'),/,
     +T2,'Forecast decile |  1  |  2  |  3  ',
     +'|  4  |  5  |  6  |  7  |  8  |  9  | 10  |',/,
     +T2,77('-'),/T2,'Observed propn. |',10(F5.3,'|'),/,
     +T2,'Expected propn. |',10(F5.3,'|'),/,
     +T2,'Number of cases |',10(I5,'|'),/,
     +T2,77('-'),//,
     +T10,'Mean Squared Error (mean Brier score): ',F7.4,/,
     +T10,'Mean Pearson residual: ',F7.4,/,
     +T10,'Standard deviation of Pearson residuals: ',F7.4,/
     +T10,'Standard error of mean Pearson residual: ',F7.4,/)
 12   FORMAT(//,30X,'RESIDUAL ANALYSIS',/30X,17('='),//,
     +5X,'Mean of observations: ',F7.3,/,
     +5X,'Standard deviation of observations: ',F7.3,//,
     +5X,'Mean error (observed - predicted): ',F7.3,/,
     +5X,'Root mean squared error: ',F7.3,/,
     +5X,'Proportion of variance explained by model: ',F6.3,//,
     +5X,'Mean Pearson residual: ',F7.3,/,
     +5X,'Standard deviation of Pearson residuals: ',F7.3,/,
     +5X,'Expected std dev of Pearson residuals: ',F7.3,//
     +5X,'Mean Anscombe residual: ',F7.4,' (expected: ',F7.4,')',/,
     +5X,'Std Dev of Anscombe residuals: ',F7.4,
     +' (expected: ',F7.4,')',/)
 13   FORMAT(/5X,'MODEL PERFORMANCE BY MONTH',/5X,26('-'),//,
     +T13,'|',8X,'||     Pearson residuals',/,
     +T7,'Month | N days ||   Mean    Std Dev   S.E. mean',/,
     +T6,49('-'))
 14   FORMAT(T8,I4,1X,'| ',I6,' ||',1X,F7.4,3X,F7.4,4X,F7.4)
 15   FORMAT(/5X,'MODEL PERFORMANCE BY SITE',/5X,25('-'),//,
     +T35,'|',8X,'||     Pearson residuals',/,
     +T7,' Site  Name',17X,'| N days ||   Mean    Std Dev   S.E. mean',
     +/,T6,73('-'))
 16   FORMAT(T8,A4,2X,A20,1X,'| ',I6,' ||',1X,F7.4,3X,F7.4,4X,F7.4)
 17   FORMAT(/5X,'MODEL PERFORMANCE BY YEAR',/5X,25('-'),//,
     +T13,'|',8X,'||     Pearson residuals',/,
     +T7,' Year | N days ||   Mean    Std Dev   S.E. mean',/,
     +T6,49('-'))
 20   FORMAT(/5X,'NOTE: Standard errors computed for overall, ',
     +'monthly and yearly',/5X,'means above are inflated to ',
     +'compensate for spatial dependence.')

      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE SNDWCH(XWXINV,XWSWX,FILNO,P,NDAYS,MXP)
*
*     To calculate sandwich variance estimator in the presence of 
*     inter-site dependence. Arguments:
*
*     XWXINV   - On entry, the `independence' variance estimator
*                (i.e. [X'WX]^{-1}). On exit, the corrected 
*                estimator.
*     XWSWX    - On entry, X'WX. On exit, the filling in the sandwich.
*     FILNO    - Number of scratch file containing daily scores
*     P        - Number of parameters in model (excluding constant)
*     NDAYS    - Number of days of data
*     MXP      - Dimensioning for XWXINV and XWSWX
******************************************************************************
      INTEGER FILNO,P,NDAYS,MXP
      DOUBLE PRECISION XWXINV(0:MXP,0:MXP),XWSWX(0:MXP,0:MXP)
******************************************************************************
*     Extra INTEGERs
*     ^^^^^^^^^^^^^^
*     MXP2     - Maximum number of model parameters  } dimensioning
*     I,J,K    - counters
******************************************************************************
      INTEGER MXP2
      PARAMETER(MXP2=80)
      INTEGER I,J,K
******************************************************************************
*     Extra DOUBLEs
*     ^^^^^^^^^^^^^
*     U        - Score vector read from file
*     TMP      - temporary storage
******************************************************************************
      DOUBLE PRECISION U(0:MXP2),TMP(0:MXP2,0:MXP2)
*
*     First: check available storage
*
      IF (P.GT.MXP2) CALL PRGERR(' SNDWCH ',' fit_both.f ',2)
*
*     OK, now compute the matrix XWSWX. It's just the sum of squares
*     of rows of FILNO. As it's symmetric, we just compute the upper 
*     triangle.
*
      REWIND(FILNO)
      DO 100 I=0,P
       DO 101 J=0,P
        XWSWX(I,J) = 0.0D0
 101   CONTINUE
 100  CONTINUE

      DO 120 K=1,NDAYS
       READ(FILNO) (U(I),I=0,P)
       DO 130 I=0,P
        DO 131 J=I,P
         XWSWX(I,J) = XWSWX(I,J) + (U(I)*U(J))
 131    CONTINUE
 130   CONTINUE
 120  CONTINUE

*
*     Right: at this point XWSWX is correct in its upper triangle.
*     The lower triangle hasn't been set, but as it's symmetric it
*     doesn't matter. The matrix we require is [XWXINV][XWSWX][XWXINV].
*     It's quickest to do this in 2 steps: first compute [XWXINV][XWSWX]
*     and store in TMP; then compute [TMP][XWXINV]. Since the result
*     is symmetric, we only need to compute one half of it, which is 
*     stored in the lower triangle of XWSWX for the moment. The diagonal
*     goes into U.
*
      DO 160 I=0,P
       DO 165 J=0,P
        TMP(I,J) = 0.0D0
        DO 170 K=0,J-1
         TMP(I,J) = TMP(I,J) + (XWXINV(I,K)*XWSWX(K,J))
 170    CONTINUE
        DO 171 K=J,P
         TMP(I,J) = TMP(I,J) + (XWXINV(I,K)*XWSWX(J,K))
 171    CONTINUE
 165   CONTINUE
 160  CONTINUE

      DO 180 I=0,P
       U(I) = 0.0D0
       DO 185 K=0,P
        U(I) = U(I) + (TMP(I,K)*XWXINV(K,I))
        DO 190 J=0,I-1
         XWSWX(I,J) = XWSWX(I,J) + (TMP(I,K)*XWXINV(K,J))
 190    CONTINUE
 185   CONTINUE
 180  CONTINUE
*
*     Finally, copy correct covariance matrix into XWXINV, and tidy up
*     XWSWX
*
      DO 200 I=0,P
       XWXINV(I,I) = U(I)
       DO 201 J=I+1,P
        XWXINV(I,J) = XWSWX(J,I)
        XWXINV(J,I) = XWXINV(I,J)
        XWSWX(J,I) = XWSWX(I,J)
 201   CONTINUE
 200  CONTINUE

      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE CORFIT(SPMOD,CORMAT,CORRN,SITINF,NATTR,
     +                                  SCODES,NSITES,RHO,MXN,MXP)
***************************************************************************
*       To fit spatial correlation models to a set of empirical correlations.
*       Arguments:
*
*       SPMOD   (integer, input). Code indicating the model to be fitted
*       CORMAT  (double precision, input). A matrix of empirical inter-
*               site correlations
*       CORRN   (double precision, input). Matrix of sample sizes from
*               which the entries in CORMAT have been calculated. 
*       SITINF  (double precision, input). Array containing site information. 
*               See calling program for details
*       NATTR   (integer, input). Number of site attributes that have been 
*               defined
*       SCODES  - array of short site codes
*       NSITES  (integer, input). Number of sites for which empirical 
*               correlations have been calculated.
*       RHO     (double precision, input / output). Array of correlation
*               model parameters. On input this contains an initial 
*               estimate; on output it contains the fitted value
*       MXN     (integer, input). Maximum number of sites (used for
*               dimensioning CORMAT etc.)
*       MXP     (integer, input). Maximum number of model parameters.
*               Used for dimensioning RHO.
*
*       The estimation is done by minimising the weighted sum of 
*       squares
*
*       S = \sum n_{ij} (c(x_{ij},y_{ij},\rho)-r_{ij})^2
*
*       where the sum is over all pairs (i,j) of sites; n_{ij} and r_{ij}
*       are the number of observations and empirical correlation for pair 
*       (i,j); and c(x_{ij},y_{ij},rho) is the fitted correlation for
*       that pair of sites which is possibly dependent on the x- and 
*       y-separations x_{ij} and y_{ij}, and the correlation parameter(s)
*       rho. 
*
*       The general form of the correlation functions in this routine
*       is 
*
*       c(x,y) = alpha + (1-alpha)*cc(x,y)
*
*       where cc(.,.) is a valid spatial correlation function and
*       alpha is in the range (0,1). Where there is no "threshold"
*       parameter alpha, it is taken as zero in the code below. 
*
***************************************************************************
      INTEGER SPMOD,NSITES,NATTR,MXN,MXP
      DOUBLE PRECISION CORMAT(MXN,MXN),CORRN(MXN,MXN)
      DOUBLE PRECISION SITINF(MXN,MXP,0:3)
      DOUBLE PRECISION RHO(MXP)
      CHARACTER SCODES(MXN)*4
***************************************************************************
*       Extra INTEGERs
*       --------------
*       MXN2    Maximum number of sites we'll ever need. Should be 
*               set equal to MXN.
*       NPAR    Number of parameters in dependence model
*       I,J,P,Q Counters
*       SPM2    Number of parameters in spatial correlation model cc
*               (see above)
*       ITER    Current iteration number in search for optimum
*               parameter set
*       MAXIT   Maximum number of iterations
***************************************************************************
      INTEGER MXN2,NPAR,I,J,P,Q,SPM2,ITER,MAXIT
      PARAMETER (MXN2=200)
***************************************************************************
*       Extra DOUBLEs
*       -------------
*       GRAD    Gradient of the least-squares objective function when 
*               doing Newton-Raphson. Can't imagine there will be any
*               dependence models with more than 20 parameters. 
*       HESS    Hessian of the least-squares objective function
*       ADJ     Vector of Newton adjustments to the current parameter
*               vector
*       STEP    Step size for Newton adjustments (used to prevent
*               iteration from going into illegal regions of
*               parameter space
*       LL, UL  Lower and upper limits for parameters
*       ADJMAX  Largest adjustment on current iteration
*       XSEP }  Matrices of inter-site separations in the first
*       YSEP }  two attributes defined in SITINF
*       DIST    Matrix of Euclidean inter-site distances
*       RBAR    Mean of observed correlations
*       DBAR    Mean inter-site distance
*       SUMN    Sum of the entries in CORRN (used to normalise the
*               calculations for stability)
*       R       Theoretical correlation matrix corresponding to 
*               current entries of RHO and ALPHA
*       RR      Values of the correlation function cc (see above)
*       ALPHA   Value of alpha (see above)
*       RRGRAD  Derivatives of the correlation function cc at current 
*               site
*       D2RR    Second derivatives ...
*       TOL     Tolerance used to judge convergence of Newton-Raphson
*       TMP?    Temporary storage
***************************************************************************
      DOUBLE PRECISION GRAD(20),HESS(20,20),ADJ(20),ADJMAX
      DOUBLE PRECISION STEP,LL(20),UL(20)
      DOUBLE PRECISION XSEP(MXN2,MXN2),YSEP(MXN2,MXN2),R(MXN2,MXN2)
      DOUBLE PRECISION RR(MXN2,MXN2),ALPHA
      DOUBLE PRECISION RRGRAD(20),D2RR(20,20)
      DOUBLE PRECISION DIST(MXN2,MXN2),RBAR,DBAR,SUMN,TOL,TMP1,TMP2
***************************************************************************
      DATA (LL(I),I=1,20) /20*-1.0D6/
      DATA (UL(I),I=1,20) /20*1.0D6/      
*
*	First calculate mean inter-site correlation and distance
*       (not needed for all models but it's quick)
*
      RBAR =0.0D0
      DBAR = 0.0D0
      SUMN=0.0D0
      MAXIT = 100
      ITER = 0
      DO 100 I=1,NSITES-1
       DO 101 J=I+1,NSITES
        XSEP(I,J) = 0.0D0
        YSEP(I,J) = 0.0D0
        IF (NATTR.GE.1) XSEP(I,J)=SITINF(I,1,0)-SITINF(J,1,0)
        IF (NATTR.GE.2) YSEP(I,J)=SITINF(I,2,0)-SITINF(J,2,0)
        DIST(I,J) = DSQRT( (XSEP(I,J)*XSEP(I,J)) + 
     +                     (YSEP(I,J)*YSEP(I,J)) )
        IF (CORRN(I,J).GT.0.0D0) THEN
         RBAR = RBAR + (CORRN(I,J)*CORMAT(I,J))
         DBAR = DBAR + (CORRN(I,J)*DIST(I,J))
         SUMN = SUMN + CORRN(I,J)
        ENDIF
 101   CONTINUE
 100  CONTINUE
      RBAR = RBAR / SUMN
      DBAR = DBAR / SUMN
*
*       Constant inter-site correlation
*
      IF (SPMOD.EQ.2) THEN
       RHO(1) = RBAR
*
*       For the remaining correlation functions, use Newton-Raphson
*       to minimise the weighted sum of squares for the nonlinear
*       part of the model. N-R doesn't work if there's a threshold
*       though, because in this case the equations have a root at
*       ALPHA=1. For threshold models therefore, do an inner N-R
*       iteration for the parameters of the fundamental correlation 
*       function; and then calculate the value of ALPHA 
*       corresponding to this.
* 
      ELSE IF ( (SPMOD.GE.3).AND.(SPMOD.LE.6) ) THEN
       TOL = 1.0D-5
       ALPHA = 0.0D0
       SPM2 = SPMOD
       IF ((SPMOD.EQ.4).OR.(SPMOD.EQ.6)) THEN
        SPM2 = SPMOD-1
       ENDIF
*
*       Initial values for the fundamental correlation functions,
*       if they haven't been set by the user. Exponential is initialised 
*       by equating mean observed correlation with fitted value at 
*       average inter-site distance. Powered exponential the same. 
*
       IF (SPM2.EQ.3) THEN
        NPAR = 1
        IF (DABS(RHO(1)).LT.1.0D6) RHO(1) = -DLOG(RBAR) / DBAR
        LL(1) = 0.0D0
       ELSEIF (SPM2.EQ.5) THEN
        NPAR = 2
        IF (DABS(RHO(1)).LT.1.0D6) RHO(1) = -DLOG(RBAR) / DBAR
        IF (DABS(RHO(2)).LT.1.0D6) RHO(2) = 1.0D0
        LL(1) = 0.0D0
        LL(2) = 0.0D0
       ENDIF
*
*       For model with non-zero asymptote, initialise the asymptote
*       using analytical solution. For models *without* this 
*       asymptote, calling CORSET twice is unnecessary but not
*       expensive; to avoid this would make the code rather cumbersome
*
 130   ITER=ITER+1
       CALL CORSET(' fit_both.f ',0,SPM2,RHO,RR,NSITES,
     +			        SITINF,NATTR,SCODES,MXN,MXP)
       IF ( (SPMOD.EQ.4).OR.(SPMOD.EQ.6) ) THEN
        TMP1 = 0.0D0
        TMP2 = 0.0D0
        DO 140 I=1,NSITES-1
         DO 141 J=I+1,NSITES
          TMP1 = TMP1 + (CORRN(I,J)*(CORMAT(I,J)-RR(I,J))*
     +                                              (1.0D0-RR(I,J)))
          TMP2 = TMP2 + (CORRN(I,J)*(1.0D0-RR(I,J))*(1.0D0-RR(I,J)))
 141     CONTINUE
 140    CONTINUE
        ALPHA = TMP1 / TMP2
        RHO(NPAR+1) = ALPHA
       ENDIF
       CALL CORSET(' fit_both.f ',0,SPMOD,RHO,R,NSITES,
     +			        SITINF,NATTR,SCODES,MXN,MXP)
*
*       Set up the gradient and hessian vector
*
       DO 150 P = 1,NPAR
        GRAD(P) = 0.0D0
        DO 151 Q=1,NPAR
         HESS(P,Q) = 0.0D0
 151    CONTINUE
 150   CONTINUE

       DO 160 I=1,NSITES-1
        DO 161 J=I+1,NSITES
         IF (SPM2.EQ.3) THEN
          RRGRAD(1) = -DIST(I,J)*RR(I,J)
          D2RR(1,1) = DIST(I,J)*DIST(I,J)*RR(I,J)
         ELSE IF (SPM2.EQ.5) THEN
          TMP1 = DIST(I,J)**RHO(2)
          TMP2 = TMP1*RR(I,J)
          RRGRAD(1) = -TMP2
          RRGRAD(2) = -TMP2*RHO(1)*DLOG(DIST(I,J))
          D2RR(1,1) = TMP2*TMP1
          TMP1 = (RHO(1)*TMP1) - 1.0D0
          D2RR(1,2) = TMP2 * TMP1 * DLOG(DIST(I,J))
          D2RR(2,1) = D2RR(1,2)
          D2RR(2,2) = TMP2 * RHO(1) * TMP1 * (DLOG(DIST(I,J))**2)
         ENDIF
         DO 165 P=1,NPAR
          GRAD(P) = GRAD(P) - (CORRN(I,J) * ( CORMAT(I,J)-R(I,J) ) *
     +                                                 RRGRAD(P) )
          DO 166 Q=P,NPAR
           HESS(P,Q) = HESS(P,Q) - ( CORRN(I,J) * (
     +                       ( (CORMAT(I,J)-R(I,J))*D2RR(P,Q) )
     +                       - (1.0D0-ALPHA) * RRGRAD(P)*RRGRAD(Q) ) )
 166      CONTINUE
 165     CONTINUE
 161    CONTINUE
 160   CONTINUE
*
*       Scaling by 1-ALPHA works in all cases because ALPHA is 0 for
*       models where it's not needed
*
       DO 170 P = 1,NPAR
        GRAD(P) = (1.0D0-ALPHA) * GRAD(P) / SUMN
        DO 171 Q=P,NPAR
         HESS(P,Q) = (1.0D0-ALPHA) * HESS(P,Q) / SUMN
         IF (Q.NE.P) HESS(Q,P) = HESS(P,Q)
 171    CONTINUE
 170   CONTINUE
*
*       Invert Hessian and compute adjustments. Use STEP to prevent
*       parameters from straying out of legal boundaries
* 
       CALL MATINV(HESS,NPAR,1,20,0)
       STEP = 1.0D0
       ADJMAX = 0.0D0
       DO 180 P=1,NPAR
        ADJ(P) = 0.0D0
        DO 181 Q=1,NPAR
         ADJ(P) = ADJ(P) + ( HESS(P,Q) * GRAD(Q) ) 
         IF (DABS(ADJ(P)).GT.ADJMAX) ADJMAX = DABS(ADJ(P))
 181    CONTINUE
        STEP = DMIN1(STEP,0.5D0*DABS( (RHO(P)-LL(P)) / ADJ(P) ),
     +                     0.5D0*DABS( (UL(P)-RHO(P)) / ADJ(P) ))
 180   CONTINUE
       DO 185 P=1,NPAR
        RHO(P) = RHO(P) - (STEP*ADJ(P))
 185   CONTINUE
       IF (ADJMAX.GT.TOL) THEN
        IF (ITER.LT.MAXIT) GOTO 130
        WRITE(*,2) ADJMAX
       ENDIF
      ELSE
       WRITE(*,1)
       STOP
      ENDIF

 1    FORMAT(/'****ERROR**** In routine CORFIT, an invalid value ',
     +'of variable',
     +/'SPMOD has been found. This is a programming error - please ',
     +'contact REC.')
 2    FORMAT(/'****WARNING**** Iteration limit reached while ',
     +'searching for optimal',/'spatial correlation parameters: ',
     +'largest standardised objective function ',/'gradient is ',
     +F7.5,'. You should check the fit graphically; if inadequate ',
     +/'try different starting values.',/)

      END
