      PROGRAM FITGAM
******************************************************************************
*       PROGRAM AUTHOR:         RICHARD E. CHANDLER
*                               (email: richard@stats.ucl.ac.uk)
*       AFFILIATION:            DEPARTMENT OF STATISTICAL SCIENCE
*                                UNIVERSITY COLLEGE LONDON
*                                 GOWER STREET
*                                  LONDON WC1E 6BT
*                                   UK
*       VERSION: 1f      	DATE: JULY 2001
*        UCL 2001-2010
******************************************************************************
*       To fit a generalized linear model to rain days. The model is 
*       for Gamma distributed observations with constant coefficient 
*       of variation.
*
*       There are NOBS cases. The program allows fitting of up 
*       to MXPARAMS covariates
*
*       Censoring (ie recording all daily rainfall values below .1mm
*       as trace) is allowed for: in the predictors, by having indicator
*       variables for trace (these are orthogonal to main previous day
*       variables), and in the response by replacing traces by conditional
*       expectations under current model parameterisation
*
*       The following classes of explanatory variable are allowed for:
*       -       Previous days' rainfall behaviour (up to 10 previous days)
*       -       Seasonal effects, using sinusoids fitted by month
*       -       Long-term trends (climate change)
*       -       Site effects e.g. elevation
*       In addition, 2- and 3-way interactions between components may be
*       specified; also, estimation of parameters in various nonlinear 
*	transformations of predictors.
*
*       The program should be compiled along with the following other source
*       files:
*       
*	glm_base.f	Routines common to all GLM rainfall programs
*	fit_both.f	Routines used in both GLM fitting programs
*       funcdefs.f      Different scenarios for climate change, seasonal
*                       patterns, dependence on previous days etc.
*	anscombe.f	Stuff for calculating theoretical properties
*			of Anscombe residuals
*       rec_math.f      Various mathematical routines and functions.
*
*	REVISION HISTORY: See manual
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>    PARAMETERS      >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*	BYTELN		No. of characters required in a binary file record to
*			store a single DOUBLE PRECISION number. Value is
*			installation-dependent, but should be as small as
*			possible. Here are some values for installations I've
*			tried:
*			Silicon Graphics (Irix 6.3):	2	(!!!)
*			Sun (SunOS 5.5.1):		8
*			IBM (Aix 2.3):			8
*			PC (Linux 2.0.35):		8
*			PC (MSDOS/Salford F77):		8
*       MXNSITES        Maximum no. of sites the program will cope with. There
*                       is a corresponding statement in almost every 
*                       subroutine.
*       MXPARAMS        Maximum number of parameters in each model. 
*
*	NB IF THESE VALUES ARE CHANGED, THE CORRESPONDING ENTRY IN
*       OTHER SOURCE FILES WILL ALSO NEED CHANGING. 
*
******************************************************************************
	INTEGER BYTELN,MXNSITES,MXPARAMS
	PARAMETER (BYTELN=2,MXNSITES=200,MXPARAMS=80)
******************************************************************************
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>      VARIABLES    >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
***************************************************************************
*               DOUBLE PRECISION variables
*               ^^^^^^^^^^^^^^^^^^^^^^^^^^
*       RAIN    Vector of rainfall amounts for 1 day
*       Y       Vector of all daily rainfalls
*       X       Covariates for the current case
*       COVS    Values of covariates at each site for 1 day
*       BETA    Current estimate of parameter vector (element
*               0 is the constant term)
*       SEBETA  Standard errors of parameters
*       THETA   Array of parameters in nonlinear functions of basic
*               predictors. Each function has up to 2 parameters. 
*               The elements of THETA are arranged to line up with the
*               covariates to which they correspond. For parameters we're
*               estimating rather than assuming to be known, THETAs
*               duplicate the last few elements of BETA - basically
*               because its easier to do the bookkeeping if they're in
*               BETA as well, as they can then be updated as part of the
*               IWLS procedure.
*       SETHETA - standard errors of THETA
*	RHO	parameters defining spatial correlation matrix. Dimensioned to
*		MXPARAMS purely for convenience.
*       LOGL    Log-likelihood
*       ETA     Linear predictor for current case
*       MU      Predicted mean for current case
*       NU      The dispersion parameter (common scale parameter 
*               of the gamma dbns)
*       PI      Pi. Used for calculating seasonal components.
*	GLBVAL  Values of global quantities such as `trace' thresholds.
*       THRESH  Threshold for dealing with small positive values.
*       TRACE   Ditto, but when we're treating them as `trace' values
*       PREVIOUS- Array of previous days' rainfall amounts: Jth column
*                 corresponds to J days previously
*       SITINF- Information regarding each site (Eastings, Northings etc.)
*                 Column 0 holds the actual value, 1 and 2 hold any
*                 derivatives wrt any parameters of nonlinear transformations
******************************************************************************
      DOUBLE PRECISION RAIN(MXNSITES),PREVIOUS(MXNSITES,10)
      DOUBLE PRECISION SITINF(MXNSITES,MXPARAMS,0:3)
      DOUBLE PRECISION Y,X(0:MXPARAMS)
      DOUBLE PRECISION COVS(MXNSITES,MXPARAMS)
      DOUBLE PRECISION BETA(0:MXPARAMS),SEBETA(0:MXPARAMS),LOGL
      DOUBLE PRECISION THETA(MXPARAMS,3),SETHETA(MXPARAMS,3)
      DOUBLE PRECISION ETA,MU,NU
      DOUBLE PRECISION RHO(MXPARAMS)
      DOUBLE PRECISION GLBVAL(MXPARAMS),THRESH,TRACE
******************************************************************************
*               INTEGER variables
*               ^^^^^^^^^^^^^^^^^
*       I,J     Counters
*       P       No. of parameters in model
*       MAXIT   Max. no of iterations permitted
*       SITXFM  Selects nonlinear transformations of site attributes.
*               First column chooses the attribute, second the xfrmation
*       TRENDSEL        Selects long-term trend function.
*       TRENDINTR       Selects which components long-term
*                       trend will interact with. When they're
*                       selected it contains the numbers of the
*                       components
*       TOGGLE          To toggle elements of TRENDINTR
*       PARMIDX         To partition the covariates into main, trace &
*                       trend segments
*       NOBS    - no. of observations
*       NSITES  - no. of sites required
*       NATTR   - no. of site attributes (e.g. elevation, Eastings etc.)
*       NLSTAT  - status of nonlinear parameters. A value of 0 indicates
*                 they're to be treated as known, 1 that we have to 
*                 estimate them.
*	SITE	- number of the current site
*       YY,MM,DD - year, month and day (used for calculation of
*                  previous days' rainfalls)
*       DAYS_IN_MONTH   - no. of days in current month (integer function)
*       NP      - Numbers of parameters in model. Elements are:
*                 1 - No. of main effects associated with site
*                 2 - NP(1) + No. of main effects associated with year
*                 3 - NP(2) + No. of main effects associated with month
*                 4 - NP(3) + No. of main effects associated with previous
*                               days' rainfall (not trace values)
*                 5 - NP(4) + No. of 2-way interactions
*                 6 - NP(5) + No. of 3-way interactions
*                 7 - NP(6) + No. of nonlinear parameters
*                 8 - No. of `global' parameters such as trace thresholds
*		  9 - No. of dispersion parameters
*		 10 - No. of spatial correlation parameters
*       COVCODE - Code numbers of covariates 
*       TWO     - Indices of two-way interactions in model
*       THREE   - guess ...
*       GLBCOD  - Codes defining `global' quantities in GLBVAL
*	FOUIDX	- Index array for Fourier representation of site effects. 
*		  See file funcdefs.f for full details
*	LEGIDX	- Index array for Legendre polynomial representation of
*		  site effects.
*       PWTIDX  - Indexing for weighted averages of previous days' values
*                 ((I,0)th entry is the number of the covariate to which the
*                 parametrisation for the Ith weighting scheme is attached;
*                 (I,J)th is the model number of the Jth parameter in the
*                 scheme).
*       WTSCHM  - Indexing for weighted average of previous days' values
*                 (Ith element is the weighting scheme used in Ith
*                 covariate).
*	MXLAG	- Number of immediately previous days' rainfall required
*		  to be present at a site in order for it to be included
*		  in the fitting. This allows comparison of models with
*		  different numbers of lags.
*	SPMOD	- choice of residual spatial correlation structure
*       THRTYP  - Type of thresholding to be carried out on small
*                 positive values. 0: no thresholding; 1: `soft'
*                 thresholding (zero everything below threshold, 
*                 subtract threshold from everything above); 2:
*                 `hard' thresholding (just zero everything below
*                 threshold)
*       IEXTD } - Indicate whether `external' predictors at daily,
*       IEXTM }   monthly or yearly timescales have been requested
*       IEXTY }
*       FILOPN  - indicator for whether the input file is open or not
*                 contains year, 3 contains month and 4 contains day.
*       ICHECK  - Indicator for whether model parametrisation has been 
*                 checked.
*	REGCODES - In principal, defines subareas for each site. Not used
*		   here except by way of compatibility with routine SITEDEF.
*	MISSFL	- flags for sites which have missing covariate data
*	SCRFN1,	- Channel numbers of scratch files. 1 holds covariate
*	SCRFN2	  information, 2 holds stuff for residual analyses.
*	OUTFNO	- Channel number for output file
***************************************************************************
      INTEGER NSITES,NATTR,NOBS
      INTEGER SITE,YY,MM,DD
      INTEGER NP(10),COVCODE(MXPARAMS),NLSTAT(MXPARAMS,3),SPMOD,MXLAG
      INTEGER GLBCOD(MXPARAMS),IEXTD,IEXTM,IEXTY
      INTEGER SITXFM(MXPARAMS,2),TWO(MXPARAMS,2),THREE(MXPARAMS,3)
      INTEGER FOUIDX(MXPARAMS),LEGIDX(MXPARAMS)
      INTEGER PWTIDX(MXPARAMS,0:3),WTSCHM(MXPARAMS),THRTYP
      INTEGER I,J,P,MAXIT,ICHECK
      INTEGER FILOPN,OUTFNO
      INTEGER REGCODES(MXNSITES),MISSFL(MXNSITES)
      INTEGER SCRFN1,SCRFN2
******************************************************************************
***************************************************************************
***************************************************************************
*               CHARACTER variables
*               ^^^^^^^^^^^^^^^^^^^
*       XLABELS Variable labels for columns of design matrix
*       NLTXT   - Descriptive text for nonlinear parameters
*       TRENDLBL Label for long-term trend function used
*       TRPTXT  - Describes nonlinear parameters in trend functions
*       OPTION  - Used for 'Y/N' input
*       CHAN30  - name of input file containing data from gauges
*       CHAN31  - name of input file containing external annual predictors
*       CHAN32  - name of input file containing external monthly predictors
*       CHAN33  - name of input file containing external daily predictors
*       CHAN36  - name of input file containing site definitions
*       CHAN42  - name of input file containing specfication for intensity
*                 model
*       CHAN60  - name of output file containing model description and
*                 fitting results
*	CHAN65	- name of output data file for residual analysis
*	CHAN66	- name of output data file for spatial correlation
*       CHAN70  - name of file containing model specification for fitted model
*       TITLE   - Description of intensity model
*       SITETXT - array of site descriptors etc.
*       SCODES  - array of short site codes
*       ATTRTXT - Text for site attributes
*       SXPTXT  - Text for parameters in nonlinear site transformations
*       MOTXT   - Text for monthly model components
*       TRNDTXT - Text for trend components
*       DYTXT -   Text for daily components
*       PWTTXT  - Text for parameters in nonlinear weighting schemes for
*                 averaging previous days' values
*	SPATXT	- Text for spatial correlation models
*	SPPTXT	- Text for parameters in spatial correlation 
*		  model (dimension to MXPARAMS for convenience)
*       CORTXT  - array describing quantities to which inter-site 
*                 correlations correspond
*       GLBTXT  - Text for global quantities
*       OPTION  - Used for interactive Y/N input
***************************************************************************
      CHARACTER XLABELS(0:MXPARAMS)*70,NLTXT(MXPARAMS,3)*70
      CHARACTER*12 CHAN30,CHAN31,CHAN32,CHAN33,CHAN36,CHAN42,
     +CHAN60,CHAN65,CHAN66,CHAN70
      CHARACTER*70 TITLE,SITETXT(MXNSITES)
      CHARACTER SCODES(MXNSITES)*4
      CHARACTER*70 ATTRTXT(MXPARAMS)
      CHARACTER*70 MOTXT(MXPARAMS)
      CHARACTER*70 TRNDTXT(MXPARAMS),DYTXT(MXPARAMS)
      CHARACTER*70 PWTTXT(MXPARAMS,3),TRPTXT(MXPARAMS,3)
      CHARACTER*70 SXPTXT(MXPARAMS,3),GLBTXT(MXPARAMS)
      CHARACTER*70 SPATXT(0:MXPARAMS),SPPTXT(MXPARAMS,4)
      CHARACTER*70 CORTXT(MXPARAMS)
      CHARACTER OPTION*1
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>    COMMON BLOCKS   >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      COMMON /DESCRIBE/ SXPTXT,MOTXT,TRNDTXT,TRPTXT,DYTXT,
     +			SPATXT,SPPTXT,CORTXT,PWTTXT,GLBTXT
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>    FILE NUMBERS    >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*       The following FORTRAN file numbers are allocated in this program:
*       UNIT    DESCRIPTION
*       ----    -----------
*       30      Input file containing data from gauges
*       31      Input file containing external annual predictors
*       32      Input file containing external monthly predictors
*       33      Input file containing external daily predictors
*       36      Input file containing definition of sites
*       42      Input file containing model definition
*       60      Output file, containing model description and parameter
*               estimates
*       65      Output file, containing data for residual analysis
*	66	Output file containing spatial correlation structure of
*		residuals
*       70      Output file containing model specification for fitted model
*       99      Temporary scratch file used for doing character/numeric
*               conversion
******************************************************************************
******************************************************************************
******************************************************************************
*
*       Initializing ...
*
      DATA ((PREVIOUS(I,J),J=1,10),I=1,MXNSITES) /
     +  MXNSITES*-1.0D0,MXNSITES*-1.0D0,MXNSITES*-1.0D0,
     +  MXNSITES*-1.0D0,MXNSITES*-1.0D0,MXNSITES*-1.0D0,
     +  MXNSITES*-1.0D0,MXNSITES*-1.0D0,MXNSITES*-1.0D0,
     +  MXNSITES*-1.0D0/
      DATA (RAIN(I),I=1,MXNSITES) /MXNSITES*-1.0D0/
      DATA (NP(I),I=1,10),MXLAG /10*0,0/
      DATA ((THETA(I,J),I=1,MXPARAMS),J=1,3) 
     +/MXPARAMS*1.0D9,MXPARAMS*1.0D9,MXPARAMS*1.0D9/
      DATA (GLBVAL(I),I=1,MXPARAMS) /MXPARAMS*1.0D9/
      DATA (GLBCOD(I),I=1,MXPARAMS) /MXPARAMS*-1/
      DATA ((NLSTAT(I,J),I=1,MXPARAMS),J=1,3)
     +/MXPARAMS*-1,MXPARAMS*-1,MXPARAMS*-1/
      DATA ((SITXFM(I,J),I=1,MXPARAMS),J=1,2) 
     +/MXPARAMS*0,MXPARAMS*0/
      DATA (FOUIDX(I),I=1,MXPARAMS) /MXPARAMS*0/
      DATA (LEGIDX(I),I=1,MXPARAMS) /MXPARAMS*0/
      DATA ((PWTIDX(I,J),J=0,3),I=1,MXPARAMS) /
     +MXPARAMS*0,MXPARAMS*0,MXPARAMS*0,MXPARAMS*0/
      DATA (WTSCHM(I),I=1,MXPARAMS) /MXPARAMS*0/
      DATA SPMOD,(RHO(I),I=1,MXPARAMS),NU /0,MXPARAMS*1.0D9,1.0D0/
      DATA OUTFNO /60/
      DATA CHAN30,CHAN36,CHAN42,CHAN60,CHAN65,CHAN66,CHAN70
     +/'gaugvals.dat','siteinfo.def','gammamdl.def','gammamdl.res',
     +'res_gamm.dat','cor_gamm.dat','gammamdl.de2'/
      DATA CHAN31,CHAN32,CHAN33 /'yr_preds.dat','mn_preds.dat',
     +'dy_preds.dat'/
      DATA YY,MM,DD,IEXTD,IEXTM,IEXTY,ICHECK /7*0/
**************************************************************************
**********              EXECUTABLES START HERE                  **********
**************************************************************************
*
*       Open data and results files
*
      WRITE(*,*) 'Opening files ...'
      CALL FILCHK(CHAN30,30,1)
      FILOPN = 1
      CALL FILCHK(CHAN36,36,1)
      CALL FILCHK(CHAN42,42,1)
      CALL FILCHK(CHAN60,OUTFNO,2)
      WRITE(60,1)
      CALL FILCHK(CHAN65,65,2)
      CLOSE (65)
      CALL FILCHK(CHAN70,70,2)
*
*       Read site locations
*
      WRITE(*,*) 'Reading site information ...'
      CALL SITEDEF(36,CHAN36,ATTRTXT,NATTR,SITINF,REGCODES,
     +                    SCODES,SITETXT,NSITES,MXNSITES,MXPARAMS)
      WRITE(*,2) NSITES
      WRITE(OUTFNO,2) NSITES
*
*       Read model specification, output to screen and summary file.
*       We output the descriptive text associated with each model 
*       choice for further checking. We also calculate any requested
*       nonlinear transformations of site attributes, and check that
*       data files are available if the user has asked for external
*       predictors.
*
      WRITE(*,*) 'Reading model specification ...'
      CALL MDLSET(42,CHAN42,MXPARAMS,11,0,NATTR,NP,GLBVAL,BETA,THETA,
     +     NU,COVCODE,GLBCOD,MXLAG,SPMOD,IEXTD,IEXTM,IEXTY,SITXFM,
     +     WTSCHM,TWO,THREE,NLSTAT,FOUIDX,LEGIDX,PWTIDX,SITINF,
     +     MXNSITES,NSITES,RHO,TITLE)
      CALL RHOCHK(RHO,SPMOD,CHAN42,MXPARAMS)
      IF ((SPMOD.GE.1).AND.(SPMOD.LT.20)) CALL FILCHK(CHAN66,66,2)
      IF (IEXTY.EQ.1) CALL FILCHK(CHAN31,31,1)
      IF (IEXTM.EQ.1) CALL FILCHK(CHAN32,32,1)
      IF (IEXTD.EQ.1) CALL FILCHK(CHAN33,33,1)
      CALL ATTRXFM(TITLE,SITINF,NSITES,ATTRTXT,NATTR,COVCODE,
     +				SITXFM,FOUIDX,LEGIDX,THETA,NP,
     +				ICHECK,MXNSITES,MXPARAMS)
      CALL TXTSET(NP,COVCODE,GLBCOD,SITXFM,PWTIDX,WTSCHM,ATTRTXT,
     +      SXPTXT,TRNDTXT,TRPTXT,MOTXT,DYTXT,PWTTXT,XLABELS,NLTXT,
     +      GLBTXT,MXPARAMS)
      CALL CHKMDL(BETA,SEBETA,THETA,SETHETA,GLBVAL,NU,RHO,COVCODE,
     +  TWO,THREE,GLBCOD,NP,NLSTAT,PWTIDX,SPMOD,TITLE,XLABELS,NLTXT,
     +  SPATXT,SPPTXT,CORTXT,ATTRTXT,GLBTXT,1,11,MXPARAMS)
      CALL MXITST(MAXIT)
*
*       P is the number of parameters estimated. If NP(7) is 0,
*       there are no nonlinear parameters in the model and there
*       are NP(6) parameters to be estimated.
*
      IF (NP(7).EQ.0) THEN
       P = NP(6)
      ELSE
       P = NP(7)
      ENDIF
*
*     Define thresholds for small positive values, if required
*
      CALL TAUDEF(GLBVAL,GLBCOD,NP(8),THRESH,TRACE,THRTYP,MXPARAMS)
*
*	Open scratch files for storing huge arrays. Each record in
*	the first needs (P+2) doubles, then a bit extra for year, 
*	month and day; for the second we need 3 doubles, site,year,month 
*	and day. Channel number search starts at 80 in each case.
*
      SCRFN1 = 85
      SCRFN2 = 85
      CALL SCROPN('DIRECT    ','UNFORMATTED',BYTELN*(P+5),SCRFN1)
      CALL SCROPN('DIRECT    ','UNFORMATTED',9*BYTELN,SCRFN2)
*
*       Read data, and set up design matrix. Read the data a day at a time,
*       and then copy the appropriate bits to the X and Y arrays.
*
      WRITE(*,*) 'Reading data ...'
      NOBS=0
      X(0) = 1.0D0
 50   CALL UPDATE(30,RAIN,PREVIOUS,THRESH,THRTYP,NSITES,YY,MM,DD,
     +                                MXLAG,MXNSITES,SCODES,FILOPN)
      CALL COVSET(COVS,NP,NSITES,COVCODE,TWO,THREE,GLBCOD,SITINF,
     +     YY,MM,DD,PREVIOUS,PWTIDX,MXLAG,GLBVAL,BETA,THETA,TRACE,
     +     0,0,MISSFL,ICHECK,MXNSITES,MXPARAMS,0,1)
*
*       Use this value only if all required predictors are present
*       and the rainfall value is non-zero
*
      DO 51 SITE=1,NSITES
       IF ((RAIN(SITE).GT.0.0D0).AND.(MISSFL(SITE).EQ.0)) THEN
	NOBS = NOBS + 1
	Y = RAIN(SITE)
	DO 52 J=1,P
	 X(J) = COVS(SITE,J)
 52     CONTINUE
*
*	Write this case to first scratch file, and flag to indicate no
*       further parameter checking is required. 
*
        WRITE(SCRFN1,REC=NOBS,ERR=920) SITE,YY,MM,DD,Y,(X(J),J=1,P)
        IF (ICHECK.EQ.0) ICHECK = 1
       ENDIF 
 51   CONTINUE
*
*     Successful read - no further need to check model parametrisation
*
      IF (FILOPN.EQ.1) GOTO 50
      IF (NOBS.EQ.0) GOTO 930
*
*       And do estimation.
*
      WRITE(*,*) 'Beginning estimation ...'
      CALL IWLS(11,NP,P,COVCODE,TWO,THREE,NLSTAT,SPMOD,FOUIDX,
     +			LEGIDX,PWTIDX,MXPARAMS,MXNSITES,MAXIT,MXLAG,
     +			SCRFN1,SCRFN2,OUTFNO,NOBS,NSITES,NATTR,SITXFM,
     +			GLBCOD,LOGL,BETA,SEBETA,THETA,SETHETA,NU,RHO,
     +			GLBVAL,SITINF,TITLE,XLABELS,NLTXT,SPATXT,
     +			SPPTXT,CORTXT,ATTRTXT,GLBTXT,SCODES,CHAN30,30)
*
*	Do estimation of spatial dependence structure if required,
*	then do final model output.
*
      IF (SPMOD.NE.0) THEN
       CALL COREST(SCRFN2,NSITES,NOBS,SPMOD,NU,SCODES,RHO,66,
     +		                 SITINF,NATTR,MXPARAMS,MXNSITES)
      ENDIF

      CALL CHKMDL(BETA,SEBETA,THETA,SETHETA,GLBVAL,NU,RHO,COVCODE,
     +     TWO,THREE,GLBCOD,NP,NLSTAT,PWTIDX,SPMOD,TITLE,XLABELS,
     +	   NLTXT,SPATXT,SPPTXT,CORTXT,ATTRTXT,GLBTXT,3,11,MXPARAMS)
*
*       Write a new model definition file for the fitted model
*
      CALL NEWMDL(42,70,MXPARAMS,NP,BETA,THETA,GLBVAL,NU,
     +     RHO,COVCODE,GLBCOD,MXLAG,SITXFM,TWO,THREE,NLSTAT,PWTIDX,
     +     SPMOD,TITLE,XLABELS,GLBTXT)
*
*       Option to do basic residual analysis
*
      WRITE(*,5)
      READ(*,'(A1)') OPTION
      IF ( (OPTION.EQ.'N').OR.(OPTION.EQ.'n') ) STOP

      CALL RESID(11,NU,P+1,NOBS,SCRFN2,OUTFNO,SCODES,SITETXT,
     +                                      NSITES,MXNSITES)
      CLOSE (OUTFNO)
*
*       Option to generate output file for further residual analysis.
*       NB we have already ascertained that any existing file can
*       be overwritten.
*
      WRITE(*,6)
      READ(*,'(A1)') OPTION
      IF ( (OPTION.EQ.'N').OR.(OPTION.EQ.'n') ) STOP

      OPEN(65,FILE=CHAN65)
      WRITE(65,7)
      DO 200 I=1,NOBS
       READ(SCRFN2,REC=I) SITE,YY,MM,DD,Y,MU,ETA
       WRITE(65,'(A4,1X,I4,T14,I2,T17,I3,T21,F8.4,T31,F8.4,T41,F8.4)')
     +   SCODES(SITE),YY,MM,DD,Y,MU,ETA
 200  CONTINUE

      CLOSE (65)
      CLOSE(SCRFN2)

      STOP
*
*       Error trapping
*
 920  IF (NOBS.EQ.1) THEN
       WRITE(*,10)
      ELSE
       WRITE(*,11)
      ENDIF
      STOP

 930  WRITE(*,*) 
     +'****ERROR**** No data found. Check site definition/data files'
      STOP

 1    FORMAT(/20X,'GENERALIZED LINEAR MODEL FITTING RESULTS',
     +/20X,'========================================',//10X,
     +'This file contains results of fitting a generalised linear ',
     +'model to',/5X,
     +'sequences of daily rainfall values. The model assumes gamma',
     +' distributed',/5X,
     +'errors with constant coefficient of variation, and a log ',
     +'link function',/5X,
     +'(i.e. predicted mean rainfall = Exp[linear combination of ',
     +'predictors]).',/5X,
     +'Note that the model is for strictly positive values only - ',
     +'all others',/5X,'will be discarded from the input file.')
 2    FORMAT(10X,'Data from ',I2,' sites will be used in the ',
     +'fitting.')
 5    FORMAT(//5X,'Do you want a basic residual analysis ',
     +'(Y/N, default=Y)?')
 6    FORMAT(//5X,'Generate output file for residual analysis ',
     +'(Y/N, default=Y)?')
 7    FORMAT('SITE YEAR MONTH DAY OBSERVED PREDICTED    ETA')  
 9    FORMAT(/5X,'****WARNING**** Reached maximum ',
     +'number of iterations.')
 10   FORMAT(/10X,59('*'),/10X,
     +'**** ERROR: In file fit_gamm.f, value of BYTELN is too ',
     +'****',/10X,
     +'**** small. Increase it and recompile (see file header ',
     +'****',/10X,
     +'**** for suggested values for your operating system).  ',
     +'****',/10X,59('*'))
 11   FORMAT(/10X,55('*'),/10X,
     +'**** ERROR: while writing to scratch file. You are ',
     +'****',/10X,
     +'**** probably out of disk space.                   ',
     +'****',/10X,55('*'))
      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE COREST(FILNO,NSITES,NOBS,SPMOD,NU,SCODES,
     +                        RHO,OUTFNO,SITINF,NATTR,MXP,MXN)
*
*       To estimate Anscombe residual correlation structure
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>    PARAMETERS      >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*       MXNSIT  - Max. no of sites (should be equal to value of MXNSITES
*                 in main program)
******************************************************************************
	INTEGER MXNSIT
	PARAMETER (MXNSIT=200)
******************************************************************************
*               INTEGER variables
*               ^^^^^^^^^^^^^^^^^
*   I,J	   - Counters
*   FILNO  - File number of scratch file with residual info in
*   MXP	   - Maximum number of parameters
*   MXN	   - Maximum number of sites
*   NOBS   - no. of observations
*   NSITES - no. of sites required
*   SPMOD  - spatial dependence model selected
*   OUTFNO - File number of output file (for writing observed
*            correlation matrix)
*   NATTR  - Number of site attributes that have been defined
***************************************************************************
      INTEGER NSITES,NOBS,FILNO,SPMOD,MXP,MXN,OUTFNO,NATTR
      INTEGER I,J
***************************************************************************
*               DOUBLE PRECISION variables
*               ^^^^^^^^^^^^^^^^^^^^^^^^^^
*   SITINF Array containing site information. See calling program
*          for details
*   NU     Fitted coeff. of variation
*   RHO    Vector of parameter estimates for spatial dependence
*   CORMAT Spatial correlation matrix/matrix of cross-products
*   CORRN  Ns going to make up entries in CORMAT
*   XSEP } Separations between current site pair in the first 
*   YSEP } two attributes contained in SITINF
******************************************************************************
      DOUBLE PRECISION NU,SITINF(MXN,MXP,0:3)
      DOUBLE PRECISION CORMAT(MXNSIT,MXNSIT),CORRN(MXNSIT,MXNSIT)
      DOUBLE PRECISION RHO(MXP),XSEP,YSEP
***************************************************************************
*               CHARACTER variables
*               ^^^^^^^^^^^^^^^^^^^
*       SCODES  - array of short site codes
***************************************************************************
      CHARACTER SCODES(MXN)*4
***************************************************************************
*	Check dimensioning:
*
      IF (MXNSIT.NE.MXN) CALL PRGERR(' COREST ',' fit_gamm.f ',1)
*
*       Calculate estimated empirical correlation matrix
*
      WRITE(*,2)
      CALL EMPCOR(FILNO,NSITES,NOBS,11,2,2,CORMAT,CORRN,MXNSIT)
*
*	And produce estimates of correlation structure. For all
*       fitted models, empirical correlations are written to file
*       (missing values coded as -9.999) since this provides the
*       opportunity to assess the fit visually afterwards. 
*
      IF ((SPMOD.GT.0).AND.(SPMOD.LT.20)) THEN
       WRITE(OUTFNO,18)
       XSEP = 0.0D0
       YSEP = 0.0D0
       DO 100 I=1,NSITES-1
        DO 101 J=I+1,NSITES
         IF (NATTR.GE.1) XSEP=SITINF(I,1,0)-SITINF(J,1,0)
         IF (NATTR.GE.2) YSEP=SITINF(I,2,0)-SITINF(J,2,0)
         IF (CORRN(I,J).LE.0.0D0) CORMAT(I,J) = -9.999D0
         WRITE(OUTFNO,19) SCODES(I),SCODES(J),CORMAT(I,J),
     +	                              INT(CORRN(I,J)),XSEP,YSEP
 101    CONTINUE
 100   CONTINUE
       IF (SPMOD.GT.1) CALL CORFIT(SPMOD,CORMAT,CORRN,SITINF,NATTR,
     +                                   SCODES,NSITES,RHO,MXN,MXP)
      ELSE
       WRITE(*,1)
       STOP
      ENDIF

 1    FORMAT(/'****ERROR**** In routine COREST, an invalid value ',
     +'of variable',
     +/'SPMOD has been found. This is a programming error - please ',
     +'contact REC.')
 2    FORMAT(/5X,'Iteration complete - now computing spatial ',
     +'dependence structure ...')
 18   FORMAT('Site1  Site2  Corr      N   Xsep    Ysep ') 
 19   FORMAT(A4,4X,A4,4X,F7.4,1X,I5,2(1X,F7.3))

      END
     
