      PROGRAM LOGISTIC
******************************************************************************
*       PROGRAM AUTHOR:         RICHARD E. CHANDLER
*                               (email: richard@stats.ucl.ac.uk)
*       AFFILIATION:            DEPARTMENT OF STATISTICAL SCIENCE
*                                UNIVERSITY COLLEGE LONDON
*                                 GOWER STREET
*                                  LONDON WC1E 6BT
*                                   UK
*       VERSION: 1f      	DATE: JULY 2001
*        UCL 2001
*****************************************************************************
*       To fit a logistic regression model for the occurrence of rain.
*
*       There are NOBS cases. The program allows fitting of up 
*       to MXPARAMS covariates
*
*       Censoring (ie recording all daily rainfall values below .1mm
*       as trace) is allowed for in the predictors, by having indicator
*       variables for trace (these are orthogonal to main previous day
*       variables)
*
*       The following classes of explanatory variable are allowed for:
*       -       Previous days' rainfall behaviour (up to 10 previous days)
*       -       Seasonal effects, using sinusoids fitted by month
*       -       Long-term trends (climate change)
*       -       Site effects e.g. elevation
*       In addition, 2- and 3-way interactions between components may be
*       specified; also, estimation of parameters in various nonlinear 
*	transformations of predictors.
*
*       The program should be compiled along with the following other source
*       files:
*
*	glm_base.f	Routines common to all GLM rainfall programs
*	fit_both.f	Routines used in both GLM fitting programs
*       funcdefs.f      Different scenarios for climate change, seasonal
*                       patterns, dependence on previous days etc.
*       rec_math.f      Various mathematical routines and functions.
*
*	REVISION HISTORY: See manual
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>    PARAMETERS      >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*	BYTELN		No. of characters required in a binary file record to
*			store a single DOUBLE PRECISION number. Value is
*			installation-dependent, but should be as small as
*			possible. Here are some values for installations I've
*			tried:
*			Silicon Graphics (Irix 6.3):	2	(!!!)
*			Sun (SunOS 5.5.1):		8
*			IBM (Aix 2.3):			8
*			PC (Linux 2.0.35):		8
*			PC (MSDOS/Salford F77):		8
*       MXNSITES        Maximum no. of sites the program will cope with. There
*                       is a corresponding statement in almost every 
*                       subroutine.
*       MXPARAMS        Maximum number of parameters in each model. 
*
*	NB IF THESE VALUES ARE CHANGED, THE CORRESPONDING ENTRY IN
*       OTHER SOURCE FILES WILL ALSO NEED CHANGING. 
******************************************************************************
	INTEGER BYTELN,MXNSITES,MXPARAMS
	PARAMETER (BYTELN=2,MXNSITES=200,MXPARAMS=80)
******************************************************************************
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>      VARIABLES     >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
***************************************************************************
*               DOUBLE PRECISION variables
*               ^^^^^^^^^^^^^^^^^^^^^^^^^^
*       RAIN    Vector of rainfall amounts for 1 day
*       Y       Daily rainfall (0 or 1 for logistic)
*       X       Covariates for current case
*	COVS	Values of covariates at each site for 1 day
*       BETA    Current estimate of parameter vector (element
*               0 is the constant term)
*	SEBETA	Standard errors of parameters
*	THETA	Array of parameters in nonlinear functions of basic
*		predictors. Each function has up to 2 parameters. 
*		The elements of THETA are arranged to line up with the
*		covariates to which they correspond. For parameters we're
*		estimating rather than assuming to be known, THETAs
*		duplicate the last few elements of BETA - basically
*		because its easier to do the bookkeeping if they're in
*		BETA as well, as they can then be updated as part of the
*		IWLS procedure.
*	SETHETA	- standard errors of THETA
*	RHO	parameters defining spatial correlation matrix. Dimensioned to
*		MXPARAMS purely for convenience
*       LOGL    Log-likelihood
*       ETA     Linear predictor for the current case
*       PROBS   Predicted probability for current case
*       PI      Pi. Used for calculating seasonal components.
*       GLBVAL  Values of `global' quantities such as trace thresholds
*       PREVIOUS- Array of previous days' rainfall amounts: Jth column
*                 corresponds to J days previously
*       THRESH  Threshold used for dealing with small positive values
*       TRACE   Ditto, when we're trating them as `trace' values
*       SITINF- Information regarding each site (Eastings, Northings etc.)
*		  Column 0 holds the actual value, 1 and 2 hold any
*		  derivatives wrt any parameters of nonlinear transformations
*       TMP     - Temporary storage
******************************************************************************
      DOUBLE PRECISION RAIN(MXNSITES),PREVIOUS(MXNSITES,10)
      DOUBLE PRECISION SITINF(MXNSITES,MXPARAMS,0:3)
      DOUBLE PRECISION Y,X(0:MXPARAMS)
      DOUBLE PRECISION COVS(MXNSITES,MXPARAMS)
      DOUBLE PRECISION BETA(0:MXPARAMS),SEBETA(0:MXPARAMS),LOGL
      DOUBLE PRECISION THETA(MXPARAMS,3),SETHETA(MXPARAMS,3)
      DOUBLE PRECISION GLBVAL(MXPARAMS),RHO(MXPARAMS),THRESH,TRACE
      DOUBLE PRECISION ETA,PROBS
      DOUBLE PRECISION TMP
******************************************************************************
*               INTEGER variables
*               ^^^^^^^^^^^^^^^^^
*       I,J     Counters
*       P       No. of parameters in model
*	MAXIT	Max. no of iterations permitted
*	SITXFM	Selects nonlinear transformations of site attributes.
*		First column chooses the attribute, second the xfrmation
*       TRENDSEL        Selects long-term trend function.
*       TRENDINTR       Selects which components long-term
*                       trend will interact with. When they're
*                       selected it contains the numbers of the
*                       components
*       TOGGLE          To toggle elements of TRENDINTR
*       PARMIDX         To partition the covariates into main, trace &
*                       trend segments
*	NOBS	- no. of observations
*       NSITES  - no. of sites required
*       NATTR 	- no. of site attributes (e.g. elevation, Eastings etc.)
*	NLSTAT	- status of nonlinear parameters. A value of 0 indicates
*		  they're to be treated as known, 1 that we have to 
*		  estimate them.
*	SITE	- number of current site
*	YY,MM,DD - year, month and day (used for calculation of
*		   previous days' rainfalls)
*       DAYS_IN_MONTH   - no. of days in current month (integer function)
*       NP      - Numbers of parameters in model. Elements are:
*                 1 - No. of main effects associated with site
*                 2 - NP(1) + No. of main effects associated with year
*                 3 - NP(2) + No. of main effects associated with month
*                 4 - NP(3) + No. of main effects associated with previous
*                               days' rainfall (not trace values)
*                 5 - NP(4) + No. of 2-way interactions
*                 6 - NP(5) + No. of 3-way interactions
*		  7 - NP(6) + No. of nonlinear parameters
*                 8 - Number of `global' parameters such as trace thresholds
*		  9 - Number of dispersion parameters
*		 10 - No. of spatial correlation parameters
*       COVCODE  - Code numbers of covariates 
*       TWO      - Indices of two-way interactions in model
*       THREE    - guess ...
*       GLBCOD  - Codes defining `global' quantities in GLBVAL
*	FOUIDX	- Index array for Fourier representation of site effects. 
*		  See file funcdefs.f for full details
*	LEGIDX	- Index array for Legendre polynomial representation of
*		  site effects.
*       PWTIDX  - Indexing for weighted averages of previous days' values
*                 ((I,0)th entry is the number of the covariate to which the
*                 parametrisation for the Ith weighting scheme is attached;
*                 (I,J)th is the model number of the Jth parameter in the
*                 scheme).
*       WTSCHM  - Indexing for weighted average of previous days' values
*                 (Ith element is the weighting scheme used in Ith
*                 covariate).
*	MXLAG	- Number of immediately previous days' rainfall required
*		  to be present at a site in order for it to be included
*		  in the fitting. This allows comparison of models with
*		  different numbers of lags.
*	SPMOD   - choice of structure for residual spatial correlation
*       THRTYP  - Type of thresholding to be carried out on small
*                 positive values. 0: no thresholding; 1: `soft'
*                 thresholding (zero everything below threshold, 
*                 subtract threshold from everything above); 2:
*                 `hard' thresholding (just zero everything below
*                 threshold)
*       IEXTD } - Indicate whether `external' predictors at daily,
*       IEXTM }   monthly or yearly timescales have been requested
*       IEXTY }
*	FILOPN	 - indicator for whether the input file is open or not
*       ICHECK   - Indicator for whether model parametrisation has been 
*                  checked
*	REGCODES - In principle, defines subareas for each site. Not used
*		   here except by way of compatibility with routine SITEDEF.
*	MISSFL	- flags for sites which have missing covariate data
*	SCRFN1,	- Channel numbers of scratch files. 1 holds covariate
*	SCRFN2	  information, 2 holds stuff for residual analyses.
*	OUTFNO	- Channel number of output file
***************************************************************************
      INTEGER NSITES,NATTR,NOBS
      INTEGER SITE,YY,MM,DD
      INTEGER NP(10),COVCODE(MXPARAMS),NLSTAT(MXPARAMS,3),SPMOD,MXLAG
      INTEGER GLBCOD(MXPARAMS),IEXTD,IEXTM,IEXTY
      INTEGER SITXFM(MXPARAMS,2),TWO(MXPARAMS,2),THREE(MXPARAMS,3)
      INTEGER FOUIDX(MXPARAMS),LEGIDX(MXPARAMS)
      INTEGER PWTIDX(MXPARAMS,0:3),WTSCHM(MXPARAMS),THRTYP
      INTEGER I,J,P,MAXIT,FILOPN,ICHECK
      INTEGER REGCODES(MXNSITES),MISSFL(MXNSITES)
      INTEGER SCRFN1,SCRFN2,OUTFNO
******************************************************************************
***************************************************************************
***************************************************************************
*               CHARACTER variables
*               ^^^^^^^^^^^^^^^^^^^
*       XLABELS - Variable labels for columns of design matrix
*	NLTXT	- Descriptive text for nonlinear parameters
*       TRENDLBL- Label for long-term trend function used
*	TRPTXT	- Describes nonlinear parameters in trend functions
*       OPTION  - Used for 'Y/N' input
*       CHAN30  - name of input file containing data from gauges
*       CHAN31  - name of input file containing external annual predictors
*       CHAN32  - name of input file containing external monthly predictors
*       CHAN33  - name of input file containing external daily predictors
*       CHAN36  - name of input file containing site definitions
*       CHAN42  - name of input file containing specfication for intensity
*                 model
*       CHAN60  - name of output file containing model description and
*		  fitting results
*	CHAN65	- name of output data file for residual analysis
*	CHAN67	- name of output data file for latent spatial correlations
*       TITLE   - Description of model
*       SITETXT - array of site descriptors etc.
*       SCODES  - array of short site codes
*       ATTRTXT - Text for site attributes
*	SXPTXT	- Text for parameters in nonlinear site transformations
*       MOTXT   - Text for monthly model components
*       TRNDTXT - Text for trend components
*	TRPTXT	- Text for nonlinear parameters in trend components
*       DYTXT   - Text for previous day components
*       PWTTXT  - Text for parameters in nonlinear weighting schemes for
*                 averaging previous days' values
*	SPATXT	- Text for spatial correlation models
*	SPPTXT	- Text for parameters in spatial correlation 
*		  model (dimension to MXPARAMS for convenience)
*       CORTXT  - array describing quantities to which inter-site 
*                 correlations correspond
*       GLBTXT  - Text for global quantities
*	OPTION	- Used for interactive Y/N input
***************************************************************************
      CHARACTER XLABELS(0:MXPARAMS)*70,NLTXT(MXPARAMS,3)*70
      CHARACTER*12 CHAN30,CHAN31,CHAN32,CHAN33,CHAN36,CHAN42
      CHARACTER*12 CHAN60,CHAN65,CHAN67,CHAN70
      CHARACTER*70 TITLE,SITETXT(MXNSITES)
      CHARACTER SCODES(MXNSITES)*4
      CHARACTER*70 ATTRTXT(MXPARAMS)
      CHARACTER*70 MOTXT(MXPARAMS)
      CHARACTER*70 TRNDTXT(MXPARAMS),DYTXT(MXPARAMS)
      CHARACTER*70 PWTTXT(MXPARAMS,3),TRPTXT(MXPARAMS,3)
      CHARACTER*70 SXPTXT(MXPARAMS,3),GLBTXT(MXPARAMS)
      CHARACTER*70 SPATXT(0:MXPARAMS),SPPTXT(MXPARAMS,4)
      CHARACTER*70 CORTXT(MXPARAMS)
      CHARACTER OPTION*1
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>    COMMON BLOCKS   >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      COMMON /DESCRIBE/ SXPTXT,MOTXT,TRNDTXT,TRPTXT,DYTXT,
     +			SPATXT,SPPTXT,CORTXT,PWTTXT,GLBTXT
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>    FILE NUMBERS    >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*       The following FORTRAN file numbers are allocated in this program:
*       UNIT    DESCRIPTION
*       ----    -----------
*       30      Input file containing data from gauges
*       31      Input file containing external annual predictors
*       32      Input file containing external monthly predictors
*       33      Input file containing external daily predictors
*       36      Input file containing definition of sites
*       42      Input file containing model definition
*       60      Output file, containing model description and parameter
*		estimates
*	65	Output file, containing data for residual analysis
*	67	Output file containing latent spatial correlation structure
*	70	Output file containing model specification for fitted model
*	99	Temporary scratch file used for doing character/numeric
*		conversion
******************************************************************************
******************************************************************************
******************************************************************************
*
*       Initializing ...
*
      DATA ((PREVIOUS(I,J),J=1,10),I=1,MXNSITES) /
     +  MXNSITES*-1.0D0,MXNSITES*-1.0D0,MXNSITES*-1.0D0,
     +  MXNSITES*-1.0D0,MXNSITES*-1.0D0,MXNSITES*-1.0D0,
     +  MXNSITES*-1.0D0,MXNSITES*-1.0D0,MXNSITES*-1.0D0,
     +  MXNSITES*-1.0D0/
      DATA (RAIN(I),I=1,MXNSITES) /MXNSITES*-1.0D0/
      DATA (NP(I),I=1,10),MXLAG /10*0,0/
      DATA ((THETA(I,J),I=1,MXPARAMS),J=1,3) 
     +/MXPARAMS*1.0D9,MXPARAMS*1.0D9,MXPARAMS*1.0D9/
      DATA (GLBVAL(I),I=1,MXPARAMS) /MXPARAMS*1.0D9/
      DATA (GLBCOD(I),I=1,MXPARAMS) /MXPARAMS*-1/
      DATA ((NLSTAT(I,J),I=1,MXPARAMS),J=1,3)
     +/MXPARAMS*-1,MXPARAMS*-1,MXPARAMS*-1/
      DATA ((SITXFM(I,J),I=1,MXPARAMS),J=1,2) 
     +/MXPARAMS*0,MXPARAMS*0/
      DATA (FOUIDX(I),I=1,MXPARAMS) /MXPARAMS*0/
      DATA (LEGIDX(I),I=1,MXPARAMS) /MXPARAMS*0/
      DATA ((PWTIDX(I,J),J=0,3),I=1,MXPARAMS) /
     +MXPARAMS*0,MXPARAMS*0,MXPARAMS*0,MXPARAMS*0/
      DATA (WTSCHM(I),I=1,MXPARAMS) /MXPARAMS*0/
      DATA SPMOD,(RHO(I),I=1,MXPARAMS)/0,MXPARAMS*1.0D9/
      DATA OUTFNO /60/
      DATA CHAN30,CHAN36,CHAN42,CHAN60,CHAN65,CHAN67,CHAN70 
     +/'gaugvals.dat','siteinfo.def','logistic.def','logistic.res',
     +'res_logi.dat','cor_logi.dat','logistic.de2'/
      DATA CHAN31,CHAN32,CHAN33 /'yr_preds.dat','mn_preds.dat',
     +'dy_preds.dat'/
      DATA YY,MM,DD,IEXTD,IEXTM,IEXTY,ICHECK /7*0/
**************************************************************************
**********		EXECUTABLES START HERE			**********
**************************************************************************
*
*       Open data and results files
*
      WRITE(*,*) 'Opening files ...'
      CALL FILCHK(CHAN30,30,1)
      FILOPN = 1
      CALL FILCHK(CHAN36,36,1)
      CALL FILCHK(CHAN42,42,1)
      CALL FILCHK(CHAN60,OUTFNO,2)
      WRITE(OUTFNO,1)
      CALL FILCHK(CHAN65,65,2)
      CLOSE (65)
      CALL FILCHK(CHAN70,70,2)
*
*       Read site locations
*
      WRITE(*,*) 'Reading site information ...'
      CALL SITEDEF(36,CHAN36,ATTRTXT,NATTR,SITINF,REGCODES,
     +                    SCODES,SITETXT,NSITES,MXNSITES,MXPARAMS)
      WRITE(*,2) NSITES
      WRITE(OUTFNO,2) NSITES
*
*       Read model specification, output to screen and summary file.
*       We output the descriptive text associated with each model 
*	choice for further checking. We also calculate any requested
*	nonlinear transformations of site attributes, and check that
*       data files are available if the user has asked for external
*       predictors.
*
      WRITE(*,*) 'Reading model specification ...'
      CALL MDLSET(42,CHAN42,MXPARAMS,1,0,NATTR,NP,GLBVAL,BETA,THETA,
     +	TMP,COVCODE,GLBCOD,MXLAG,SPMOD,IEXTD,IEXTM,IEXTY,SITXFM,
     +  WTSCHM,TWO,THREE,NLSTAT,FOUIDX,LEGIDX,PWTIDX,SITINF,MXNSITES,
     +  NSITES,RHO,TITLE)
      CALL ATTRXFM(TITLE,SITINF,NSITES,ATTRTXT,NATTR,COVCODE,
     +				SITXFM,FOUIDX,LEGIDX,THETA,NP,
     +				ICHECK,MXNSITES,MXPARAMS)
      IF (IEXTY.EQ.1) CALL FILCHK(CHAN31,31,1)
      IF (IEXTM.EQ.1) CALL FILCHK(CHAN32,32,1)
      IF (IEXTD.EQ.1) CALL FILCHK(CHAN33,33,1)
      CALL RHOCHK(RHO,SPMOD,CHAN42,MXPARAMS)
      IF ((SPMOD.GE.1).AND.(SPMOD.LT.20)) CALL FILCHK(CHAN67,67,2)
      CALL TXTSET(NP,COVCODE,GLBCOD,SITXFM,PWTIDX,WTSCHM,ATTRTXT,
     +       SXPTXT,TRNDTXT,TRPTXT,MOTXT,DYTXT,PWTTXT,XLABELS,NLTXT,
     +       GLBTXT,MXPARAMS)
      CALL CHKMDL(BETA,SEBETA,THETA,SETHETA,GLBVAL,-1.0D0,RHO,COVCODE,
     +  TWO,THREE,GLBCOD,NP,NLSTAT,PWTIDX,SPMOD,TITLE,XLABELS,NLTXT,
     +  SPATXT,SPPTXT,CORTXT,ATTRTXT,GLBTXT,1,1,MXPARAMS)
      CALL MXITST(MAXIT)
*
*	P is the number of parameters estimated. If NP(7) is 0,
*	there are no nonlinear parameters in the model and there
*	are NP(6) parameters to be estimated.
*
      IF (NP(7).EQ.0) THEN
       P = NP(6)
      ELSE
       P = NP(7)
      ENDIF
*
*     Define thresholds for small positive values, if required
*
      CALL TAUDEF(GLBVAL,GLBCOD,NP(8),THRESH,TRACE,THRTYP,MXPARAMS)
*
*	Open scratch files for storing huge arrays. Each record in
*	the first needs (P+2) doubles, then a bit extra for year, 
*	month and day; for the second we need 3 doubles, site,year,month 
*	and day. Channel number search starts at 85 in each case.
*
      SCRFN1 = 85
      SCRFN2 = 85
      CALL SCROPN('DIRECT    ','UNFORMATTED',BYTELN*(P+5),SCRFN1)
      CALL SCROPN('DIRECT    ','UNFORMATTED',BYTELN*9,SCRFN2)
*
*       Read data, and set up design matrix. We read the data a day at a time,
*	and then copy the appropriate bits to the X and Y arrays.
*
      WRITE(*,*) 'Reading data ...'
      NOBS=0
      X(0) = 1.0D0
 50   CALL UPDATE(30,RAIN,PREVIOUS,THRESH,THRTYP,NSITES,YY,MM,DD,
     +                          	MXLAG,MXNSITES,SCODES,FILOPN)
      CALL COVSET(COVS,NP,NSITES,COVCODE,TWO,THREE,GLBCOD,SITINF,
     +     YY,MM,DD,PREVIOUS,PWTIDX,MXLAG,GLBVAL,BETA,THETA,TRACE,
     +     0,0,MISSFL,ICHECK,MXNSITES,MXPARAMS,0,1)
*
*	Use this value only if all required predictors are present
*
      DO 51 SITE=1,NSITES
       IF ((RAIN(SITE).GE.0.0D0).AND.(MISSFL(SITE).EQ.0) ) THEN
        NOBS = NOBS + 1
        IF (RAIN(SITE).GT.0.0D0) THEN
         Y = 1.0D0
        ELSE
         Y = 0.0D0
        ENDIF
        DO 52 J=1,P
         X(J) = COVS(SITE,J)
 52     CONTINUE
*
*	Write this case to first scratch file, and flag to indicate
*       no further parameter checking is required
*
        WRITE(SCRFN1,REC=NOBS,ERR=920) SITE,YY,MM,DD,Y,(X(J),J=1,P)
        IF (ICHECK.EQ.0) ICHECK = 1
       ENDIF 
 51   CONTINUE
      IF (FILOPN.EQ.1) GOTO 50
      IF (NOBS.EQ.0) GOTO 930
*
*       And start estimating.
*
      WRITE(*,*) 'Beginning estimation ...'
      CALL IWLS(1,NP,P,COVCODE,TWO,THREE,NLSTAT,SPMOD,FOUIDX,
     +		LEGIDX,PWTIDX,MXPARAMS,MXNSITES,MAXIT,MXLAG,
     +		SCRFN1,SCRFN2,OUTFNO,NOBS,NSITES,NATTR,SITXFM,
     +		GLBCOD,LOGL,BETA,SEBETA,THETA,SETHETA,TMP,RHO,
     +		GLBVAL,SITINF,TITLE,XLABELS,NLTXT,SPATXT,
     +          SPPTXT,CORTXT,ATTRTXT,GLBTXT,SCODES,CHAN30,30)
*
*	Do estimation of spatial dependence structure if required,
*	then do final model output.
*
      IF (SPMOD.NE.0) THEN
       IF ((SPMOD.EQ.21).OR.(SPMOD.EQ.22)) THEN 
        CALL SPEST(SCRFN2,NSITES,NOBS,SPMOD,SCODES,RHO,NP,SPPTXT,
     +					MXPARAMS,MXNSITES,BYTELN)
       ELSEIF (SPMOD.GE.1) THEN
        CALL CORLAT(SCRFN2,NSITES,NOBS,SPMOD,SCODES,
     +                     RHO,67,SITINF,NATTR,MXPARAMS,MXNSITES)
       ENDIF
      ENDIF

      CALL CHKMDL(BETA,SEBETA,THETA,SETHETA,GLBVAL,-1.0D0,RHO,COVCODE,
     +  TWO,THREE,GLBCOD,NP,NLSTAT,PWTIDX,SPMOD,TITLE,XLABELS,NLTXT,
     +  SPATXT,SPPTXT,CORTXT,ATTRTXT,GLBTXT,3,1,MXPARAMS)
*
*	Write a new model definition file for the fitted model
*
      CALL NEWMDL(42,70,MXPARAMS,NP,BETA,THETA,GLBVAL,-1.0D0,
     +	RHO,COVCODE,GLBCOD,MXLAG,SITXFM,TWO,THREE,NLSTAT,PWTIDX,SPMOD,
     +  TITLE,XLABELS,GLBTXT)
*
*	Option to do basic residual analysis
*
      WRITE(*,5)
      READ(*,'(A1)') OPTION
      IF ( (OPTION.EQ.'N').OR.(OPTION.EQ.'n') ) STOP

      CALL RESID(1,1.0D0,P+1,NOBS,SCRFN2,OUTFNO,SCODES,SITETXT,
     +					NSITES,MXNSITES)

      CLOSE (OUTFNO)
*
*	Option to generate output file for further residual analysis.
*	NB we have already ascertained that any existing file can
*	be overwritten.
*
      WRITE(*,6)
      READ(*,'(A1)') OPTION
      IF ( (OPTION.EQ.'N').OR.(OPTION.EQ.'n') ) STOP

      OPEN(65,FILE=CHAN65)
      WRITE(65,7)
      DO 500 I=1,NOBS
       READ(SCRFN2,REC=I) SITE,YY,MM,DD,Y,PROBS,ETA
       WRITE(65,'(A4,1X,I4,T14,I2,T17,I3,T21,F8.4,T31,F8.4,T40,F8.4)')
     +   SCODES(SITE),YY,MM,DD,Y,PROBS,ETA
 500  CONTINUE

      CLOSE (65)
      CLOSE(SCRFN2)

      STOP
*
*	Error trapping
*
 920  IF (NOBS.EQ.1) THEN
       WRITE(*,10)
      ELSE
       WRITE(*,11)
      ENDIF
      STOP

 930  WRITE(*,*) 
     +'****ERROR**** No data found. Check site definition/data files'
      STOP

 1    FORMAT(/20X,'GENERALIZED LINEAR MODEL FITTING RESULTS',
     +/20X,'========================================',/)
 2    FORMAT(10X,'Data from ',I2,' sites will be used in the ',
     +'fitting.')
 4    FORMAT(//5X,'Results after ',I3,' iterations:'/5X,
     +'Log-likelihood - ',F14.3,/5X,
     +'Number of observations - ',I6,/5X,
     +'No. of parameters estimated - ',I2,//)
 5    FORMAT(//5X,'Do you want a basic residual analysis ',
     +'(Y/N, default=Y)?')
 6    FORMAT(//5X,'Generate file for further residual analysis ',
     +'(Y/N, default=Y)?')
 7    FORMAT('SITE YEAR MONTH DAY OBSERVED PREDICTED    ETA')  
 10   FORMAT(/10X,59('*'),/10X,
     +'**** ERROR: In file fit_logi.f, value of BYTELN is too ',
     +'****',/10X,
     +'**** small. Increase it and recompile (see file header ',
     +'****',/10X,
     +'**** for suggested values for your operating system).  ',
     +'****',/10X,59('*'))
 11   FORMAT(/10X,55('*'),/10X,
     +'**** ERROR: while writing to scratch file. You are ',
     +'****',/10X,
     +'**** probably out of disk space.                   ',
     +'****',/10X,55('*'))
 17   FORMAT(/5X,'****WARNING**** Current parameter estimates yield',
     +' a fitted',/5X,
     +'probability of either 0 or 1 (to machine precision). This ',
     +'isn''t going ',/5X,
     +'to converge - I''m bailing out. Try different starting values.')
      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE SPEST(FILNO,NSITES,NOBS,SPMOD,SCODES,RHO,NP,SPPTXT,
     +						MXP,MXN,BYTELN)
*
*   To estimate spatial dependence structure so as to match the 
*   variance of the histogram of proportions of wet sites.
*
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>    PARAMETERS      >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*   MXNSIT  - Max. no of sites (should be equal to value of MXNSITES
*             in main program)
******************************************************************************
	INTEGER MXNSIT
	PARAMETER (MXNSIT=200)
******************************************************************************
*               INTEGER variables
*               ^^^^^^^^^^^^^^^^^
*   I,J - Counters
*   FILNO - File number of scratch file with residual info in
*   MXP - Maximum number of parameters
*   MXN - Maximum number of sites
*   NP - Contains number of parameters being estimated
*   NOBS - no. of observations
*   NSITES - no. of sites required
*   NDAYS - No of days
*   SPMOD - spatial dependence model selected
*   CASEID - array of identifiers for each case, for purposes of
*            residual analysis. Col.1 contains site number, 2 
*            contains year, 3 contains month and 4 day
*   OLDDAT }- Used to determine when the date changes
*   NEWDAT }  
*   SCRFNO	- Channel number of scratch file to store daily stats
*   BYTELN	- No. of bytes needed to store a double precision number.
*        	  Carried straight over from calling prog.
*   BRACK	- Indicator in iteration to find out whether we've
*		  yet bracketed the root of an equation.
*   ITER	- Iteration number
***************************************************************************
      INTEGER NSITES,NOBS,FILNO,SPMOD,MXP,MXN,NP(10)
      INTEGER CASEID(4),OLDDAT,NEWDAT,NDAYS
      INTEGER SCRFNO,BYTELN
      INTEGER BRACK,ITER
      INTEGER I,J
***************************************************************************
*               DOUBLE PRECISION variables
*               ^^^^^^^^^^^^^^^^^^^^^^^^^^
*   Y       Daily rainfall
*   PROBS   Predicted probabilities
*   MSY	    Depends on the model:
*              Model 21 - Mean square of observations
*              Model 22 - Sum of squared `Pearson residuals'
*   MSP	    Model-dependent:
*              Model 21 - Mean square of probs
*              Model 22 - Sum(No. of sites - 1).Just using it because
*              it's convenient storage that happens to exist.
*   YBAR    Mean of observations each day
*   PBAR    Mean of predictions each day
*   NT	    No of observations each day
*   RHO	    Vector of parameter estimates for spatial dependence
*   ONEDAY  Today's amounts. Col 1 is observed, 2 is predicted
*   VAR0    Theoretical variance of proportion of wet days under
*           independence
*   SPVAR   Expected value of mean square proportion of wet sites
*   LB,UB   Lower and upper bounds for parameter
*   FL,FB   Objective function evaluated at LB and UB
*   FCUR    Objective function evaluated at current point.
*   TOL	    Accuracy required.
*   VPWET   Theoretical variance of proportion of wet sites under
*           current model parametrisation.
*   TMP     Temporary storage
******************************************************************************
      DOUBLE PRECISION Y,PROBS
      DOUBLE PRECISION MSY,MSP,YBAR,PBAR,NT
      DOUBLE PRECISION ONEDAY(MXNSIT,2)
      DOUBLE PRECISION RHO(MXP),VAR0,SPVAR
      DOUBLE PRECISION LB,UB,FL,FU,FCUR,TOL
      DOUBLE PRECISION TMP
***************************************************************************
*               CHARACTER variables
*               ^^^^^^^^^^^^^^^^^^^
*       SCODES  - array of short site codes
*	SPPTXT	- Descriptive text for parameters
***************************************************************************
      CHARACTER SCODES(MXN)*4
      CHARACTER*70 SPPTXT(MXP,4)
***************************************************************************
*	Check dimensioning:
*
      IF (MXNSIT.NE.MXN) CALL PRGERR(' SPEST  ',' fit_logi.f ',1)
*
*	Find an available file handle and open a scratch file on it 
*	(start searching at number 85)
*
      SCRFNO = 85
      CALL SCROPN('DIRECT    ','UNFORMATTED',BYTELN*2,SCRFNO)
*
*       Initialise arrays (easier this way than in a DATA statement)
*
      TOL = 1.0D-8
      BRACK = 0
      ITER = 0
      OLDDAT = 0
      MSY = 0.0D0
      MSP = 0.0D0
      VAR0 = 0.0D0
      FU = 0.0D0
      NDAYS = 0
      DO 90 I=1,NSITES
       ONEDAY(I,1) = -1.0D12
       ONEDAY(I,2) = -1.0D12
 90   CONTINUE
*
*   Read data & calculate the mean squares of both observations
*   and predictions. Also calculate necessary theoretical 
*   quantities, including starting points for iteration where
*   necessary (assume independence). Going to NOBS+1 guarantees we 
*   include the final day's measurements. All sums of squares
*   for model 21 are weighted according to the number of contributing
*   observations.
*
      WRITE(*,2)
      DO 100 I=1,NOBS+1
       IF (I.LE.NOBS) THEN
        READ(FILNO,REC=I) (CASEID(J),J=1,4),Y,PROBS
       ENDIF
*
*	Now, if the date has changed OR if we're on the last observation, 
*	update yesterday's contributions. This scheme
*	relies on the fact that the data are in chronological order. As
*	we go, we write each day's mean prediction and number of 
*	active sites to a scratch file. NB all sums of squares are
*	weighted by number of contributing observations.
*
       NEWDAT = (10000*CASEID(2))+(100*CASEID(3))+CASEID(4)
       IF (((NEWDAT.NE.OLDDAT).AND.(I.NE.1)).OR.(I.EQ.NOBS+1)) THEN
        OLDDAT=NEWDAT
        YBAR = 0.0D0
        PBAR = 0.0D0
        NT = 0.0D0
        TMP = 0.0D0
        DO 200 J=1,NSITES
         IF (ONEDAY(J,1).GT.-0.9D12) THEN
          YBAR = YBAR + ONEDAY(J,1)
          PBAR = PBAR + ONEDAY(J,2)
          TMP = TMP + ( ONEDAY(J,2)*(1.0D0-ONEDAY(J,2)) )
          NT = NT + 1.0D0
          ONEDAY(J,1) = -1.0D12
          ONEDAY(J,2) = -1.0D12
         ENDIF
 200    CONTINUE
*
*     Some model-dependent stuff ...
*
        IF (SPMOD.EQ.21) THEN
         YBAR = YBAR/NT
         PBAR = PBAR/NT
         MSY = MSY + (NT*(YBAR**2))
         MSP = MSP + (NT*(PBAR**2))
*
*	MODEL 21: Divide by NT instead of NT**2 to allow for weighting
*
         VAR0 = VAR0 + (TMP/NT)
         NDAYS = NDAYS + 1
         WRITE(SCRFNO,REC=NDAYS) PBAR,NT
        ELSEIF (SPMOD.EQ.22) THEN
*
*     MODEL 22: not much to do.
*
         PBAR = PBAR/NT 
         TMP = PBAR * (1.0D0 - PBAR)
         MSP = MSP + NT - 1.0D0
         MSY = MSY + (((YBAR-(NT*PBAR))**2)/(NT*TMP)) - 1.0D0
        ELSE
         WRITE(*,1)
         CLOSE (SCRFNO)
         STOP
        ENDIF
       ENDIF
*
*	And copy the latest entry into today's array
*
       ONEDAY(CASEID(1),1) = Y
       ONEDAY(CASEID(1),2) = PROBS
 100  CONTINUE

      IF (SPMOD.EQ.21) THEN
*
*	MODEL 21: choose parameters to fit to mean and variance 
*	of proportion of wet sites each day. Use the secant method to
*	initally bracket the root of the equation (Observed variance) -
*	(Theoretical variance) = 0, then regula falsi/bisection thereafter.
*	NB if user has requested RHO(1)=0 we reset it to 1.
*
       MSY = MSY/DBLE(NOBS)
       MSP = MSP/DBLE(NOBS)
       VAR0 = VAR0/DBLE(NOBS)
       LB = 0.0D0
       IF (RHO(1).LT.1.0D-12) RHO(1) = 1.0D0
       UB = RHO(1)
       FL = MSY - (VAR0 + MSP)
*
*	For each day, read mean predictions, individual predictions,
*	compute Bs and hence theoretical variances. We now use the
*	first column of ONEDAY to store the probabilities, and
*	the second column to store the Bs
*
 299   ITER = ITER + 1
       FCUR = MSY - MSP - SPVAR(FILNO,SCRFNO,NDAYS,NSITES,RHO,SPMOD)
*
*	That's the value of the difference in variances at
*	the current objective function value. At first we check
*	to make sure the root is actually bracketed (use the 
*	fact that we know the theoretical variance increases
*	with A so that FCUR decreases). Once the root is bracketed
*	we use regula falsi to get to the solution.
*
       IF (FCUR.GT.0.0D0) THEN
        IF (BRACK.EQ.0) THEN
         TMP = RHO(1)+( FCUR*(RHO(1)-LB)/(FL-FCUR) )
         UB = TMP
         LB = RHO(1)
         RHO(1) = UB
         FL = FCUR
        ELSE
         LB = RHO(1)
         FL = FCUR
        ENDIF
       ELSE
        BRACK = 1
        UB = RHO(1)
        FU = FCUR
       ENDIF
       IF (BRACK.EQ.1) THEN
        RHO(1) = ( (FL*UB)-(FU*LB) )/(FL-FU)
       ENDIF
       IF (DABS(FCUR).GT.TOL) GOTO 299
      ELSEIF (SPMOD.EQ.22) THEN
*
*     MODEL 22: All is very straightforward - analytical expression
*
       RHO(1) = (MSP/MSY) - 1.0D0
      ELSE
       WRITE(*,1)
       CLOSE (SCRFNO)
       STOP
      ENDIF
      CLOSE (SCRFNO)

 1    FORMAT(/'****ERROR**** In routine SPEST, an invalid value ',
     +'of variable',
     +/'SPMOD has been found. This is a programming error - please ',
     +'contact REC.')
 2    FORMAT(/5X,'Iteration complete - now computing spatial ',
     +'dependence structure ...')

      END
******************************************************************************
******************************************************************************
******************************************************************************
      FUNCTION SPVAR(FILNO1,FILNO2,NDAYS,NSITES,RHO,SPMOD)
*
*	To evaluate the variance of daily proportion of wet sites
*	under spatial dependence models. FILNO1 contains individual 
*	observations and predictions, FILNO2 contains numbers of sites 
*	for each day, and mean predictions. In fact, what is returned
*	is the expected value of 
*	[ \sum_{t=1}^{T} n_{t} * (YBAR_{t})^{2} ]/ \sum_{t=1}^{T} n_{t}
******************************************************************************
      INTEGER MXN2
      PARAMETER (MXN2=200)
******************************************************************************
*               INTEGER variables
*               ^^^^^^^^^^^^^^^^^
*	FILNO1	- File number of random access file containing observations
*		  and predictions
*	FILNO2	- File number of random access file containing daily
*		  summary info
*	NDAYS	- Number of days in database
*	NTODAT	- Number of cases read to date
*	SPMOD	- Spatial structure chosen
*	I,J,K	- Counters
*	CASEID	- Contains date & site identifiers for each case
******************************************************************************
      INTEGER FILNO1,FILNO2,NDAYS,NSITES,SPMOD,CASEID(4)
      INTEGER NTODAT,I,J,K
******************************************************************************
*               DOUBLE PRECISION variables
*               ^^^^^^^^^^^^^^^^^^^^^^^^^^
*	RHO	- Array of spatial dependence parameters
*	SPVAR	- Final function value
*	A	- Parameter to estimate in dependence model 21
*	B	- Value of B corresponding to A in model 21
*	PBAR	- Probability that a day is `wet' in model 21
*	Y	- Current observation
*	PROBS	- Current prediction
*	CURVAR	- Contribution to overall variance from current day
*	NT	- Number of sites active today
*	ONEDAY	- Vector of rainfall occurrences for today
*	TMP	- temporary storage
******************************************************************************
      DOUBLE PRECISION RHO(4),SPVAR,PBAR
      DOUBLE PRECISION A,B,Y,PROBS,CURVAR,NT
      DOUBLE PRECISION ONEDAY(MXN2,2)
      DOUBLE PRECISION TMP
******************************************************************************
*	Check dimensioning:
*
      IF (NSITES.GT.MXN2) CALL PRGERR(' SPVAR  ',' fit_logi.f',2)
*
*	Initialise:
*
      SPVAR = 0.0D0
*
*	Dependence model 21 (independence given binary weather state)
*
      IF (SPMOD.EQ.21) THEN
       NTODAT = 0
       A = DEXP(RHO(1))
       DO 300 I=1,NDAYS
        DO 302 J=1,NSITES
         ONEDAY(J,1) = -1.0D12
 302    CONTINUE
        READ(FILNO2,REC=I) PBAR,NT
        DO 305 J=1,NINT(NT)
         READ(FILNO1,REC=NTODAT+J) (CASEID(K),K=1,4),Y,PROBS
         ONEDAY(CASEID(1),1) = PROBS
         CALL BCALC(PROBS,PBAR,A,B)
         ONEDAY(CASEID(1),2) = B
 305    CONTINUE
        NTODAT = NTODAT+NINT(NT)
*
*	So now I know the Ps and the Bs, and can calculate
*	the covariances for each site pair, hence the 
*	variance. Covariances go in TMP since we don't need them
*	otherwise. Variances are calculated from marginal probabilities,
*	since covariance formula is inapplicable for i=j.
*
        CURVAR = 0.0D0
        DO 350 J=1,NSITES
         IF (ONEDAY(J,1).GT.-0.9D12) THEN
          TMP = ONEDAY(J,1)*(1.0D0-ONEDAY(J,1))
          CURVAR = CURVAR + TMP
          DO 351 K=J+1,NSITES
           IF (ONEDAY(K,1).GT.-0.9D12 ) THEN
            TMP = (1.0D0-PBAR)*ONEDAY(J,1)*ONEDAY(K,1)*
     +           (1.0D0-ONEDAY(J,1))*(1.0D0-ONEDAY(K,1))*
     +           (1.0D0-ONEDAY(J,2))*(1.0D0-ONEDAY(K,2))
            TMP = TMP/( PBAR*
     +           (1.0D0-(ONEDAY(J,1)*(1.0D0-ONEDAY(J,2))))*
     +           (1.0D0-(ONEDAY(K,1)*(1.0D0-ONEDAY(K,2)))) )
            CURVAR = CURVAR + (2.0D0*TMP)
           ENDIF
 351      CONTINUE
         ENDIF
 350    CONTINUE
*
*	Weighting by # of observations means divide by NT, not NT**2
*
        CURVAR = CURVAR/NT
        SPVAR = SPVAR + CURVAR
 300   CONTINUE
       SPVAR = SPVAR/DBLE(NTODAT)
      ELSE
       WRITE(*,1)
       CLOSE (FILNO1)
       CLOSE (FILNO2)
       STOP
      ENDIF

 1    FORMAT(/'****ERROR**** In routine SPVAR, an invalid value ',
     +'of variable',
     +/'SPMOD has been found. This is a programming error - please ',
     +'contact REC.')
      END
******************************************************************************
******************************************************************************
******************************************************************************
      SUBROUTINE CORLAT(FILNO,NSITES,NOBS,SPMOD,SCODES,
     +                        RHO,OUTFNO,SITINF,NATTR,MXP,MXN)
*
*       To estimate latent Gaussian inter-site correlations
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>    PARAMETERS      >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
*       MXNSIT  - Max. no of sites (should be equal to value of MXNSITES
*                 in main program)
*       MXNOBS  - Maximum number of days for which two sites have
*                 simultaneous observations. 
******************************************************************************
	INTEGER MXNSIT, MXNOBS
	PARAMETER (MXNSIT=200,MXNOBS=50000)
******************************************************************************
*               INTEGER variables
*               ^^^^^^^^^^^^^^^^^
*   FILNO   - File number of scratch file with residual info in
*   MXP	    - Maximum number of parameters
*   MXN	    - Maximum number of sites
*   NOBS    - no. of observations
*   NSITES  - no. of sites required
*   SPMOD   - spatial dependence model selected
*   OUTFNO  - File number of output file (for writing observed
*             correlation matrix)
*   NATTR   - Number of site attributes that have been defined
***************************************************************************
      INTEGER NSITES,NOBS,FILNO,SPMOD,MXP,MXN,OUTFNO,NATTR
***************************************************************************
*               Additional INTEGERs
*               ^^^^^^^^^^^^^^^^^^^
*   I,J,K,L - Counters
*   CASEID  - array of identifiers for each case, for purposes of
*             residual analysis. Col.1 contains site number, 2 
*             contains year, 3 contains month and 4 day
*   NSOFAR  - Number of observations read so far for a pair of sites
*   OLDDAT  - Date of previous observation read from file
*   NEWDAT  - Date of current observation
*   ITER    - Iteration number in regula falsi search
*   IFAIL   - Error flag 
***************************************************************************
      INTEGER CASEID(4)
      INTEGER I,J,K,L,NSOFAR,OLDDAT,NEWDAT,ITER,IFAIL
***************************************************************************
*               DOUBLE PRECISION variables
*               ^^^^^^^^^^^^^^^^^^^^^^^^^^
*   SITINF  Array containing site information. See calling program
*           for details
*   RHO	    Vector of parameter estimates for spatial dependence
*   JNTPRB  Matrix of overall joint occurrence probabilities at pairs of
*           sites
*   CORMAT  Estimated spatial correlation matrix of latent Gaussian field
*   CDFBVN  Joint upper tail probabilities for a standard bivariate 
*           normal distribution
*   QNORM   Quantile function for a univariate normal distribution
*   LB, UB  Lower and upper bounds for latent correlation
*   FL, FU  Values of objective function at current lower and upper bounds
*   FCUR    Value of objective function at current estimate
*   TOL     Tolerance for judging convergence of correlation search
*   JNTN    Ns going to make up entries in JNTPRB
*   Y	    Binary rainfall occurrence indicator read from file
*   P	    Modelled probability of occurrence, again read from file
*   TAU     Array of Z-scores corresponding to marginal 
*           occurrence probabilities for each pair of sites
*   XSEP }  Separation between current site pair in the first 
*   YSEP }  two attributes contained in SITINF
******************************************************************************
      DOUBLE PRECISION SITINF(MXN,MXP,0:3)
      DOUBLE PRECISION JNTPRB(MXNSIT,MXNSIT),CORMAT(MXNSIT,MXNSIT)
      DOUBLE PRECISION JNTN(MXNSIT,MXNSIT),Y,P,TAU(MXNOBS,2)
      DOUBLE PRECISION CDFBVN,QNORM,FCUR,FL,FU,LB,UB,TOL
      DOUBLE PRECISION RHO(MXP),XSEP,YSEP
***************************************************************************
*               CHARACTER variables
*               ^^^^^^^^^^^^^^^^^^^
*       SCODES  - array of short site codes
***************************************************************************
      CHARACTER SCODES(MXN)*4
***************************************************************************
*	Check dimensioning:
*
      IF (MXNSIT.NE.MXN) CALL PRGERR(' CORLAT ',' fit_logi.f ',1)
*
*       Calculate matrix of observed joint probabilites
*
      WRITE(*,3)
      CALL EMPCOR(FILNO,NSITES,NOBS,1,0,3,JNTPRB,JNTN,MXNSIT)
*
*       Now, for each pair of sites, find the latent Gaussian 
*       correlation corresponding to the observed joint occurrence
*       probability. This is done using regula falsi / bisection. 
*       It's a bit tedious - have to read all the probabilities
*       for each pair of sites every time. 
*
      DO 100 I=1,NSITES-1
       CORMAT(I,I) = 1.0D0
       DO 101 J=I+1,NSITES
        WRITE(*,4) SCODES(I),SCODES(J)
        IF (JNTN(I,J).GT.DBLE(MXNOBS)) 
     +		             CALL PRGERR(' CORLAT ',' fit_logi.f ',2)
        IF (JNTN(I,J).GT.0.0D0) THEN
*
*       Read marginal probabilities for current site pair. Going to NOBS+1
*       guarantees we include the final day's measurements
*
         NSOFAR = 0
	 OLDDAT = 0
	 TAU(1,1) = -1.0D12
	 TAU(1,2) = -1.0D12
         DO 120 K=1,NOBS+1
          IF (K.LE.NOBS) THEN
           READ(FILNO,REC=K) (CASEID(L),L=1,4),Y,P
           NEWDAT = (10000*CASEID(2))+(100*CASEID(3))+CASEID(4)
           IF ((NEWDAT.NE.OLDDAT).OR.(I.EQ.NOBS+1)) THEN
	    IF ((TAU(NSOFAR+1,1).GT.-1.0D12).AND.
     +          (TAU(NSOFAR+1,2).GT.-1.0D12)) THEN 
             NSOFAR = NSOFAR + 1
             IF (INT(JNTN(I,J)).EQ.NSOFAR) GOTO 199
            ENDIF
            OLDDAT=NEWDAT
            TAU(NSOFAR+1,1) = -1.0D12
            TAU(NSOFAR+1,2) = -1.0D12
           ENDIF
           IF (CASEID(1).EQ.I) 
     +                   TAU(NSOFAR+1,1) = -QNORM(P,0.0D0,1.0D0,IFAIL)
           IF (CASEID(1).EQ.J) 
     +                   TAU(NSOFAR+1,2) = -QNORM(P,0.0D0,1.0D0,IFAIL)
          ENDIF	
 120     CONTINUE
*
*     At this point we have an array containing the thresholds corresponding
*     to marginal probabilities at each pair of sites in the first NSOFAR
*     rows. We can now find the latent Gaussian correlation corresponding 
*     to the overall probability of joint occurrence. Start by bracketing
*     the correlation in either the range (0,1) or (-1,0) depending on
*     whether the observed joint probability is larger or smaller than 
*     that expected under independence. 
*
 199     LB = -1.0D0
         UB = 1.0D0
         FU = 0.0D0
         FL = 0.0D0
         DO 200 K=1,NSOFAR
	  FU = FU + CDFBVN(TAU(K,1),TAU(K,2),UB,1.0D-5)
	  FL = FL + CDFBVN(TAU(K,1),TAU(K,2),LB,1.0D-5)
 200     CONTINUE
         FU = JNTPRB(I,J) - (FU / DBLE(NSOFAR))
         FL = JNTPRB(I,J) - (FL / DBLE(NSOFAR))
*
*       It is conceivably possible (and has occurred!) that the
*       observed joint probability falls outside the interval 
*       dictated by the modelled margins. In this case issue a
*       warning, use whichever extreme correlation is most 
*       appropriate, and carry on. 
*
         IF (FL*FU.GT.0.0D0) THEN
          IF (FL.GT.0.0D0) THEN
           CORMAT(I,J) = UB
          ELSE
           CORMAT(I,J) = LB
          ENDIF
          WRITE(*,5) CORMAT(I,J)
          GOTO 101
         ENDIF
*
*       Regula falsi starts here. NB solving for the correlations
*       is computationally expensive, and in any case there isn't 
*       much point trying to get it more accurate than the 
*       precision of the joint probability estimate itself. 
* 
         ITER = 0
         TOL = DSQRT(JNTPRB(I,J)*(1.0D0-JNTPRB(I,J))/DBLE(NSOFAR))
         TOL = DMAX1(1.0D-5,TOL / 1.0D1)
 210     ITER = ITER + 1
         CORMAT(I,J) = ( (FL*UB)-(FU*LB) )/(FL-FU)
         FCUR = 0.0D0
         DO 211 K=1,NSOFAR
	  FCUR = FCUR + CDFBVN(TAU(K,1),TAU(K,2),
     +                         CORMAT(I,J),TOL/1.0D2)
 211     CONTINUE
         FCUR = JNTPRB(I,J) - (FCUR / DBLE(NSOFAR))
         IF (FL*FCUR.LT.0.0D0) THEN
          UB = CORMAT(I,J)
          FU = FCUR
         ELSEIF (FU*FCUR.LT.0.0D0) THEN
          LB = CORMAT(I,J)
          FL = FCUR
         ELSE 
          WRITE(*,2)
          STOP
         ENDIF

         IF ((DABS(FCUR).GT.TOL).AND.(DABS(UB-LB).GT.TOL)) GOTO 210
        ELSE
*
*       Missing correlations - code as -9.999
*
         CORMAT(I,J) = -9.999D0
        ENDIF
 101   CONTINUE
 100  CONTINUE
      CORMAT(NSITES,NSITES) = 1.0D0
*
*	And produce estimates of correlation structure. For all
*       fitted models, empirical correlations are written to file
*       (missing values coded as -9.999) since this provides the
*       opportunity to assess the fit visually afterwards. 
*
      IF ((SPMOD.GT.0).AND.(SPMOD.LT.20)) THEN
       WRITE(OUTFNO,18)
       XSEP = 0.0D0
       YSEP = 0.0D0
       DO 500 I=1,NSITES-1
        DO 501 J=I+1,NSITES
         IF (NATTR.GE.1) XSEP=SITINF(I,1,0)-SITINF(J,1,0)
         IF (NATTR.GE.2) YSEP=SITINF(I,2,0)-SITINF(J,2,0)
         WRITE(OUTFNO,19) SCODES(I),SCODES(J),CORMAT(I,J),
     +                                    INT(JNTN(I,J)),XSEP,YSEP
 501    CONTINUE
 500   CONTINUE
       IF (SPMOD.GT.1) CALL CORFIT(SPMOD,CORMAT,JNTN,SITINF,NATTR,
     +                                   SCODES,NSITES,RHO,MXN,MXP)
      ELSE
       WRITE(*,1)
       STOP
      ENDIF

 1    FORMAT(/'****ERROR**** In routine CORLAT, an invalid value ',
     +'of variable',
     +/'SPMOD has been found. This is a programming error - please ',
     +'contact REC.')
 2    FORMAT(/'*****ERROR***** Can''t bracket correlation in routine ',
     +'CORLAT. This',/,'is a programming error - please contact REC.')
 3    FORMAT(/5X,'Iteration complete - now computing spatial ',
     +'dependence structure',/5X,'(this could take some time) ...')
 4    FORMAT(10X,'--- Sites ',A4,' and ',A4,' ---')
 5    FORMAT(/'****WARNING**** Observed joint occurrence frequency ',
     +'is incompatible',/,'with modelled marginal probabilities. ',
     +'Setting latent correlation to ',F4.1) 
 18   FORMAT('Site1  Site2  Corr      N   Xsep    Ysep ') 
 19   FORMAT(A4,4X,A4,4X,F7.4,1X,I5,2(1X,F7.3))

      END
